open HolKernel Parse boolLib bossLib arithmeticTheory pred_setTheory
     listTheory sequenceTheory state_transformerTheory
     probabilityTheory formalizeUseful extra_numTheory combinTheory
     pairTheory realTheory realLib extra_boolTheory jrhUtils
     extra_pred_setTheory prob_algebraTheory probTheory realSimps
     extra_pred_setTools measureTheory numTheory
     simpLib prob_extraTheory probTools prob_extraTools pred_setLib
     subtypeTheory seqTheory std_unifTheory cdfTheory
     limTheory transcTheory itmTheory;

val _ = new_theory "cont_dist";

val uniform_rv_def  = Define
        `uniform_rv a b s= (b - a) * (std_unif_cont s) + a`;

val exp_rv_def  = Define
        `exp_rv l s= (~1/l) * ln (1 - (std_unif_cont s))`;

val rayleigh_rv_def  = Define
        `rayleigh_rv sig s= sig * sqrt (~2 * ln (1 - (std_unif_cont s)))`;


val triangular_rv_def  = Define
        `triangular_rv a s= a * (1 - sqrt (1 - (std_unif_cont s)))`;

(*=======================
Uniform Distribution
========================*);

val INV_CDF_FN_UNIF = store_thm
    ("INV_CDF_FN_UNIF",
    (--`!a b. (a < b) ==>
	(INV_CDF_FN (\x. (b - a) * x + a) 
		(\x. (if x <= a then 0 else (if x < b then (x - a) / (b - a) else 1))))`--),
	RW_TAC std_ss [] THEN
	KNOW_TAC `0 < b - a` THEN1
		RW_TAC real_ss [REAL_LT_SUB_LADD] THEN
	RW_TAC real_ss [INV_CDF_FN_def] THENL [
		Cases_on `x <= a` THEN1
			PROVE_TAC [] THEN
		FULL_SIMP_TAC std_ss [] THEN
		Cases_on `x < b` THEN1 (
			FULL_SIMP_TAC std_ss [] THEN
			POP_ASSUM K_TAC THEN
			POP_ASSUM MP_TAC THEN
			POP_ASSUM MP_TAC THEN
			RW_TAC real_ss [REAL_EQ_LDIV_EQ, REAL_EQ_SUB_RADD,REAL_EQ_IMP_LE]) THEN

		FULL_SIMP_TAC real_ss [],

		Cases_on `x <= a` THEN1
			FULL_SIMP_TAC real_ss [] THEN
		FULL_SIMP_TAC std_ss [] THEN
		Cases_on `x < b` THEN1 (
			FULL_SIMP_TAC std_ss [] THEN
			POP_ASSUM K_TAC THEN
                	POP_ASSUM K_TAC THEN
			POP_ASSUM MP_TAC THEN
			RW_TAC real_ss [REAL_EQ_LDIV_EQ, REAL_EQ_SUB_RADD]) THEN
		
		FULL_SIMP_TAC real_ss [REAL_NOT_LT],

		Cases_on `x <= a` THEN1
			FULL_SIMP_TAC real_ss [] THEN
		FULL_SIMP_TAC std_ss [] THEN
		Cases_on `x < b` THEN1
			FULL_SIMP_TAC real_ss [REAL_DIV_LMUL, REAL_POS_NZ] THEN
		FULL_SIMP_TAC real_ss [],

		KNOW_TAC `(b - a) * x' + a <= a = F` THEN1 (
			FULL_SIMP_TAC real_ss [GSYM REAL_LE_SUB_LADD, GSYM REAL_LE_RDIV_EQ] THEN
			RW_TAC std_ss [REAL_NOT_LE]) THEN
		FULL_SIMP_TAC std_ss [] THEN
		KNOW_TAC `(b - a) * x' + a < b` THEN1 (
			RW_TAC std_ss [REAL_LT_ADD_SUB] THEN
			ONCE_REWRITE_TAC [REAL_MUL_SYM] THEN
			RW_TAC real_ss [GSYM REAL_LT_RDIV_EQ, REAL_DIV_REFL, REAL_POS_NZ]) THEN
		FULL_SIMP_TAC std_ss [REAL_EQ_LDIV_EQ, REAL_MUL_SYM]]); 


val UNIF_CDF_MONO = store_thm
    ("UNIF_CDF_MONO",
    (--`!a  b. (a < b) ==> (!c d. (c < d) ==>
        ((\x. (if x <= a then 0 else (if x < b then (x - a) / (b - a) else 1))) c <=
        (\x. (if x <= a then 0 else (if x < b then (x - a) / (b - a) else 1))) d))`--),
	REPEAT STRIP_TAC THEN
	KNOW_TAC `0 < b - a` THEN1
        	RW_TAC real_ss [REAL_LT_SUB_LADD] THEN
	RW_TAC real_ss [REAL_LE_RDIV_EQ] THENL [

        	FULL_SIMP_TAC real_ss [REAL_LE_SUB_LADD, REAL_NOT_LE, REAL_LT_IMP_LE],

        	FULL_SIMP_TAC real_ss [REAL_LE_LDIV_EQ, REAL_LE_SUB_RADD, REAL_NOT_LE] THEN
        	MATCH_MP_TAC REAL_LT_IMP_LE THEN
        	MATCH_MP_TAC REAL_LTE_TRANS THEN
        	EXISTS_TAC (--`d:real`--) THEN
        	PROVE_TAC [],

        	RW_TAC std_ss [REAL_DIV_RMUL, REAL_LE_SUB_RADD,
                	REAL_LT_IMP_LE, REAL_SUB_ADD, REAL_POS_NZ],

        	RW_TAC real_ss [REAL_LE_LDIV_EQ, REAL_LE_SUB_RADD,
                	REAL_LT_IMP_LE, REAL_SUB_ADD, REAL_POS_NZ],

        	UNDISCH_TAC (--`~((c:real) <= (a:real))`--) THEN
        	RW_TAC std_ss [] THEN
        	MATCH_MP_TAC REAL_LT_IMP_LE THEN
        	MATCH_MP_TAC REAL_LTE_TRANS THEN
        	EXISTS_TAC (--`d:real`--) THEN
        	PROVE_TAC [],

        	FULL_SIMP_TAC std_ss [REAL_NOT_LE, REAL_NOT_LT] THEN
        	MATCH_MP_TAC REAL_LT_IMP_LE THEN
        	MATCH_MP_TAC REAL_LET_TRANS THEN
        	EXISTS_TAC (--`c:real`--) THEN
        	PROVE_TAC []]);



val UNIF_CONT_HELPER = store_thm
    ("UNIF_CONT_HELPER",
    (--`!x a b. (a < b) ==>
        ((\x. (x - a) / (b - a)) contl x)`--),
        RW_TAC std_ss [contl, LIM, REAL_SUB_RZERO] THEN
        EXISTS_TAC (--`(e:real) * (b - a)`--) THEN
        KNOW_TAC `0 < b - a` THEN1
                RW_TAC real_ss [REAL_LT_SUB_LADD] THEN
        RW_TAC real_ss [REAL_LT_MUL] THEN
        KNOW_TAC `((x + h - a) / (b - a) - (x - a) / (b - a)) =
                h / (b - a)` THEN1 (
                KNOW_TAC `(x + h - a) / (b - a) = h / (b - a) + (x - a) / (b - a
)` THEN1 (
                        RW_TAC real_ss [REAL_DIV_ADD, REAL_EQ_RMUL, REAL_POS_NZ,
 real_div,
                                REAL_EQ_SUB_RADD] THEN
                        REAL_ARITH_TAC) THEN
                RW_TAC std_ss [REAL_ADD_SUB_ALT]) THEN
        KNOW_TAC `abs (b - a) = b - a` THEN1
                RW_TAC std_ss [ABS_REFL, REAL_LT_IMP_LE] THEN
        RW_TAC std_ss [ABS_DIV, REAL_POS_NZ, REAL_LT_LDIV_EQ]);

val UNIF_CDF_CONT = store_thm
   ("UNIF_CDF_CONT",
   (--`!x a b. (a < b) ==>
        ((\x. if x <= a then 0 else if x < b then (x - a) / (b - a) else 1)) contl x`--),
	REPEAT STRIP_TAC THEN
	MP_TAC UNIF_CONT_HELPER THEN
	DISCH_TAC THEN
	KNOW_TAC `0 < b - a` THEN1
		RW_TAC real_ss [REAL_LT_SUB_LADD] THEN
	KNOW_TAC `abs (b - a) = b - a` THEN1
		RW_TAC std_ss [ABS_REFL, REAL_LT_IMP_LE] THEN
	RW_TAC real_ss [contl, LIM] THENL [
		NTAC 4 (POP_ASSUM MP_TAC) THEN
		POP_ASSUM (MP_TAC o Q.SPECL [`a`, `a`, `b`]) THEN
		RW_TAC real_ss [contl, LIM] THEN
		NTAC 4 (POP_ASSUM MP_TAC) THEN
		POP_ASSUM (MP_TAC o Q.SPEC `e`) THEN		
		RW_TAC real_ss [REAL_NEG_0] THEN
		FULL_SIMP_TAC std_ss [ABS_DIV, REAL_POS_NZ, REAL_LT_LDIV_EQ] THEN	
		EXISTS_TAC (--`d:real`--) THEN
		RW_TAC real_ss [ABS_0, ABS_DIV, REAL_POS_NZ, REAL_LT_LDIV_EQ] THENL [
			KNOW_TAC `0 < h` THEN1 (
				KNOW_TAC `0 < h = x < x + h` THEN1
					RW_TAC std_ss [REAL_LT_ADDR] THEN
				ASM_REWRITE_TAC [] THEN
				MATCH_MP_TAC REAL_LET_TRANS THEN
				EXISTS_TAC (--`a:real`--) THEN
				PROVE_TAC [REAL_NOT_LE]) THEN		
			KNOW_TAC `0 < x + h - a` THEN1
				RW_TAC real_ss [REAL_LT_SUB_LADD, GSYM REAL_NOT_LE] THEN
			KNOW_TAC `abs h < e * (b - a)` THEN1
				PROVE_TAC [] THEN
			KNOW_TAC `(abs h = h) /\ (abs (x + h - a) = (x + h - a))` THEN1
				PROVE_TAC [ABS_REFL, REAL_LT_IMP_LE] THEN
			FULL_SIMP_TAC std_ss [] THEN
			MATCH_MP_TAC REAL_LET_TRANS THEN
			EXISTS_TAC (--`h:real`--) THEN
			RW_TAC std_ss [REAL_LE_SUB_RADD] THEN
			RW_TAC real_ss [GSYM REAL_LE_SUB_LADD], 

			KNOW_TAC `0 < h` THEN1 (
				KNOW_TAC `0 < h = x < x + h` THEN1
					RW_TAC std_ss [REAL_LT_ADDR] THEN
				ASM_REWRITE_TAC [] THEN
				MATCH_MP_TAC REAL_LET_TRANS THEN
				EXISTS_TAC (--`a:real`--) THEN
				PROVE_TAC [REAL_NOT_LE]) THEN	
			KNOW_TAC `abs h < e * (b - a)` THEN1
				PROVE_TAC [] THEN
			KNOW_TAC `(abs h = h)` THEN1
				PROVE_TAC [ABS_REFL, REAL_LT_IMP_LE] THEN
			FULL_SIMP_TAC std_ss [REAL_NOT_LE, REAL_NOT_LT] THEN
			MATCH_MP_TAC REAL_LET_TRANS THEN
			EXISTS_TAC (--`(h:real)/(b - a)`--) THEN
			RW_TAC real_ss [REAL_LT_LDIV_EQ, REAL_LE_RDIV_EQ, REAL_LE_SUB_RADD] THEN
			MATCH_MP_TAC REAL_LE_TRANS THEN
			EXISTS_TAC (--`(x:real) + h`--) THEN
			ASM_REWRITE_TAC [GSYM REAL_LE_SUB_LADD] THEN
			ONCE_REWRITE_TAC [REAL_ADD_SYM] THEN
			RW_TAC real_ss []], 


		NTAC 5 (POP_ASSUM MP_TAC) THEN
		POP_ASSUM (MP_TAC o Q.SPECL [`x`, `a`, `b`]) THEN
		RW_TAC real_ss [contl, LIM] THEN
		NTAC 5 (POP_ASSUM MP_TAC) THEN
		POP_ASSUM (MP_TAC o Q.SPEC `e`) THEN		
		RW_TAC real_ss [REAL_NEG_0] THEN
		FULL_SIMP_TAC std_ss [REAL_NOT_LE] THEN	
		EXISTS_TAC (--`d:real`--) THEN
		RW_TAC real_ss [ABS_0] THENL [
			KNOW_TAC `0 < abs (a - x)` THEN1
				RW_TAC real_ss [GSYM ABS_NZ, REAL_NE_LT_GT, REAL_LT_SUB_RADD] THEN
			KNOW_TAC `abs (a - x) < d` THEN1 (
				RW_TAC std_ss [GSYM ABS_BETWEEN] THENL [
					MATCH_MP_TAC REAL_LTE_TRANS THEN
					EXISTS_TAC (--`(x:real) + h`--) THEN
					RW_TAC std_ss [REAL_LT_SUB_RADD, GSYM REAL_ADD_ASSOC, REAL_LT_ADDR] THEN
					ONCE_REWRITE_TAC [REAL_ADD_SYM] THEN
					RW_TAC std_ss [GSYM REAL_LT_SUB_RADD, REAL_SUB_LZERO] THEN
					MATCH_MP_TAC REAL_LET_TRANS THEN
					EXISTS_TAC (--`abs h`--) THEN
					PROVE_TAC [GSYM ABS_NEG, ABS_LE],
				
					MATCH_MP_TAC REAL_LT_TRANS THEN
					EXISTS_TAC (--`x:real`--) THEN
					RW_TAC std_ss [REAL_LT_ADDR]]) THEN
			KNOW_TAC `0 = (x + (a - x) - a) / (b - a)` THEN1
				RW_TAC real_ss [GSYM  REAL_SUB_ASSOC2] THEN
			PROVE_TAC [],

			KNOW_TAC `0 < abs (b - x)` THEN1
				RW_TAC real_ss [GSYM ABS_NZ, REAL_NE_LT_GT,REAL_LT_SUB_LADD] THEN
			KNOW_TAC `abs (b - x) < d` THEN1 (
				RW_TAC std_ss [GSYM ABS_BETWEEN, REAL_LT_SUB_RADD] THENL [
					MATCH_MP_TAC REAL_LT_TRANS THEN
					EXISTS_TAC (--`b:real`--) THEN
					RW_TAC std_ss [REAL_LT_ADDR],

					MATCH_MP_TAC REAL_LET_TRANS THEN
					EXISTS_TAC (--`(x:real) + h`--) THEN
					FULL_SIMP_TAC std_ss [REAL_NOT_LT, REAL_LT_LADD] THEN
					MATCH_MP_TAC REAL_LET_TRANS THEN
					EXISTS_TAC (--`abs h`--) THEN
					RW_TAC std_ss [ABS_LE]]) THEN
		
			KNOW_TAC `1 = (x + (b - x) - a) / (b - a)` THEN1
				RW_TAC real_ss [GSYM  REAL_SUB_ASSOC2, REAL_DIV_REFL, REAL_POS_NZ] THEN
			PROVE_TAC []],

		NTAC 5 (POP_ASSUM MP_TAC) THEN
		POP_ASSUM (MP_TAC o Q.SPECL [`1`, `a`, `b`]) THEN
		RW_TAC real_ss [contl, LIM] THEN
		NTAC 5 (POP_ASSUM MP_TAC) THEN
		POP_ASSUM (MP_TAC o Q.SPEC `e`) THEN		
		RW_TAC real_ss [REAL_NEG_0] THEN
		FULL_SIMP_TAC std_ss [REAL_NOT_LE] THEN	
		EXISTS_TAC (--`d:real`--) THEN
		RW_TAC real_ss [ABS_0] THENL [
			KNOW_TAC `0 < abs (b - a)` THEN1
				RW_TAC std_ss [] THEN
			KNOW_TAC `abs (b - a) < d` THEN1 (
				RW_TAC std_ss [GSYM ABS_BETWEEN, REAL_LT_SUB_RADD] THENL [
					MATCH_MP_TAC REAL_LT_TRANS THEN
					EXISTS_TAC (--`b:real`--) THEN
					RW_TAC std_ss [REAL_LT_ADDR],

								
					ONCE_REWRITE_TAC [REAL_ADD_SYM] THEN
					RW_TAC std_ss [GSYM REAL_LT_SUB_RADD] THEN
					MATCH_MP_TAC REAL_LET_TRANS THEN
					EXISTS_TAC (--`abs h`--) THEN
					FULL_SIMP_TAC std_ss [REAL_NOT_LT, REAL_LE_SUB_RADD] THEN
					MATCH_MP_TAC REAL_LE_TRANS THEN
					EXISTS_TAC (--`x:real`--) THEN
					RW_TAC std_ss [] THEN
					ONCE_REWRITE_TAC [REAL_ADD_SYM] THEN
					RW_TAC std_ss [GSYM REAL_LE_SUB_RADD] THEN
					MATCH_MP_TAC REAL_LE_TRANS THEN
					EXISTS_TAC (--`(x:real) + h`--) THEN
					RW_TAC std_ss [REAL_LE_SUB_RADD, REAL_LE_ADDR, GSYM REAL_ADD_ASSOC] THEN
					ONCE_REWRITE_TAC [REAL_ADD_SYM] THEN
					RW_TAC real_ss [GSYM REAL_LE_SUB_RADD, REAL_SUB_LZERO] THEN
 					PROVE_TAC [GSYM ABS_NEG, ABS_LE]]) THEN

			KNOW_TAC `abs (0 - 1) = abs ((1 + (b - a) - a) / (b - a) - (1 - a) / (b - a))` THEN1 (
				KNOW_TAC `(1 + (b - a) - a) / (b - a) = (1 - a) / (b - a) + (b - a)/(b- a)` THEN1 (
					RW_TAC real_ss [REAL_DIV_ADD, REAL_SUB_ASSOC2] THEN
					KNOW_TAC `1 + (b - a - a) = 1 + (b - a) - a` THEN1
						REAL_ARITH_TAC THEN
					PROVE_TAC []) THEN
				ASM_REWRITE_TAC [] THEN
				ONCE_REWRITE_TAC [REAL_ADD_SYM] THEN
				RW_TAC real_ss [REAL_DIV_REFL, REAL_POS_NZ, REAL_SUB_LZERO, ABS_NEG]) THEN	
			PROVE_TAC [],


			KNOW_TAC `0 < abs (x + h - b)` THEN1
				RW_TAC real_ss [GSYM ABS_NZ, REAL_NE_LT_GT, REAL_LT_SUB_RADD] THEN
			KNOW_TAC `abs (x + h - b) < d` THEN1 (
				RW_TAC std_ss [GSYM ABS_BETWEEN, REAL_LT_SUB_RADD] THENL [
					MATCH_MP_TAC REAL_LET_TRANS THEN
					EXISTS_TAC (--`(x:real)`--) THEN
					FULL_SIMP_TAC std_ss [REAL_NOT_LT, REAL_NOT_LE, REAL_LT_ADDR, 
						GSYM REAL_ADD_ASSOC] THEN
					ONCE_REWRITE_TAC [REAL_ADD_SYM] THEN
					RW_TAC std_ss [GSYM REAL_LT_SUB_RADD, REAL_SUB_LZERO] THEN
					MATCH_MP_TAC REAL_LET_TRANS THEN
					EXISTS_TAC (--`abs h`--) THEN
					PROVE_TAC [GSYM ABS_NEG, ABS_LE],


					MATCH_MP_TAC REAL_LT_TRANS THEN
					EXISTS_TAC (--`(b:real)`--) THEN
					RW_TAC std_ss [REAL_LT_ADDR]]) THEN

			KNOW_TAC `abs ((1 + (x + h - b) - a) / (b - a) - (1 - a) / (b - a)) < e` THEN1
				PROVE_TAC [] THEN
			KNOW_TAC `((1 + (x + h - b) - a) / (b - a) - (1 - a) / (b - a)) = 
			((x + h - a) / (b - a) - 1)` THEN1 (
				RW_TAC std_ss [REAL_EQ_SUB_RADD] THEN
				KNOW_TAC `(x + h - a) / (b - a) - 1 + (1 - a) / (b - a) = 
				(x + h - a) / (b - a) + (a - b) / (b - a) + (1 - a) / (b - a)` THEN1 (
					REWRITE_TAC [real_sub] THEN
					RW_TAC std_ss [REAL_EQ_RADD, REAL_EQ_LADD] THEN
					RW_TAC std_ss [GSYM real_sub, REAL_EQ_RDIV_EQ, 
					GSYM REAL_NEG_MINUS1, REAL_NEG_SUB]) THEN		
				ONCE_ASM_REWRITE_TAC [] THEN
				RW_TAC real_ss [REAL_DIV_ADD, GSYM REAL_SUB_ASSOC2] THEN
				KNOW_TAC `(1 - b + (x + h) - a) = (x + h - a - b + 1)` THEN1
					REAL_ARITH_TAC THEN
				PROVE_TAC []) THEN
			PROVE_TAC []]]);

val UNIF_CDF_AT_POSINF = store_thm
  ("UNIF_CDF_AT_POSINF",
  (--`!a b. (a < b) ==> 
	(((\n.(\x. (if x <= a then 0 else 
		(if x < b then (x - a) / (b - a) else 1)))  ((\n. &n) n)) --> 1))`--),
RW_TAC std_ss [SEQ] THEN
EXISTS_TAC (--`ceiling b`--) THEN
RW_TAC real_ss [ABS_0] THENL [
	KNOW_TAC `n < ceiling b` THEN1 (
		REWRITE_TAC [GSYM REAL_LT] THEN
		MATCH_MP_TAC REAL_LTE_TRANS THEN
		EXISTS_TAC (--`b:real`--) THEN	
		RW_TAC std_ss [LB_CEILING] THEN
		MATCH_MP_TAC REAL_LET_TRANS THEN
		EXISTS_TAC (--`a:real`--) THEN	
		PROVE_TAC []) THEN
	FULL_SIMP_TAC std_ss [GREATER_EQ, GSYM REAL_LE, GSYM REAL_LT, GSYM REAL_NOT_LT],

	KNOW_TAC `& (ceiling b) < b`  THEN1 (
		MATCH_MP_TAC REAL_LET_TRANS THEN
		EXISTS_TAC (--`&n:real`--) THEN	
		RW_TAC std_ss [REAL_LE, GSYM GREATER_EQ]) THEN
	PROVE_TAC [LB_CEILING, GSYM REAL_NOT_LT]]);

val UNIF_CDF_AT_NEGINF = store_thm
   ("UNIF_CDF_AT_NEGINF",
   (--`!a b. (a < b) ==> 
	(((\n.(\x. (if x <= a then 0 else 
		(if x < b then (x - a) / (b - a) else 1)))  ((\n. ~&n) n)) --> 0))`--),
	RW_TAC std_ss [SEQ] THEN
	EXISTS_TAC (--`ceiling (abs a)`--) THEN
	RW_TAC real_ss [ABS_0, GREATER_EQ] THENL [
		KNOW_TAC `a < ~ (& (ceiling (abs a)))` THEN1 (
			MATCH_MP_TAC REAL_LTE_TRANS THEN
			EXISTS_TAC (--`~ & n:real`--) THEN	
			RW_TAC std_ss [GSYM REAL_NOT_LE,  REAL_LE_NEG, REAL_LE]) THEN
		KNOW_TAC `~(a < ~& (ceiling (abs a)))` THEN1 (
			RW_TAC std_ss [REAL_NOT_LT] THEN
			Cases_on `0 <= a` THENL [
				MATCH_MP_TAC REAL_LE_TRANS THEN
				EXISTS_TAC (--`0:real`--) THEN
				RW_TAC std_ss [REAL_NEG_LE0, CEILING_ABS_POS],

				MATCH_MP_TAC REAL_LE_TRANS THEN
				EXISTS_TAC (--`~ (abs (a:real))`--) THEN
				RW_TAC std_ss [REAL_LE_NEG, LB_CEILING, abs, 
						REAL_NEG_NEG, REAL_LE_REFL]]),
			
		KNOW_TAC `a < ~ (& (ceiling (abs a)))` THEN1 (
			MATCH_MP_TAC REAL_LTE_TRANS THEN
			EXISTS_TAC (--`~ & n:real`--) THEN	
			RW_TAC std_ss [GSYM REAL_NOT_LE,  REAL_LE_NEG, REAL_LE]) THEN
		KNOW_TAC `~(a < ~& (ceiling (abs a)))` THEN1 (
			RW_TAC std_ss [REAL_NOT_LT] THEN
			Cases_on `0 <= a` THENL [
				MATCH_MP_TAC REAL_LE_TRANS THEN
				EXISTS_TAC (--`0:real`--) THEN
				RW_TAC std_ss [REAL_NEG_LE0, CEILING_ABS_POS],

				MATCH_MP_TAC REAL_LE_TRANS THEN
				EXISTS_TAC (--`~ (abs (a:real))`--) THEN
				RW_TAC std_ss [REAL_LE_NEG, LB_CEILING, abs, 
						REAL_NEG_NEG, REAL_LE_REFL]])]);


val UNIF_CDF_IS_CONT_CDF_FN = store_thm
   ("UNIF_CDF_IS_CONT_CDF_FN",
   (--`! a b. (a < b) ==>
        (IS_CONT_CDF_FN (\x. (if x <= a then 0 else 
		(if x < b then (x - a) / (b - a) else 1))))`--),
        REPEAT STRIP_TAC THEN
        REWRITE_TAC [IS_CONT_CDF_FN_def] THEN
        RW_TAC std_ss [IS_CONT_CDF_FN_def, UNIF_CDF_CONT,  
			UNIF_CDF_AT_NEGINF, UNIF_CDF_AT_POSINF, UNIF_CDF_MONO]);


val CDF_UNIF = store_thm
    ("CDF_UNIF",
    (--`!x a b. (a < b) ==>
        (prob bern {s | uniform_rv a b s <= x} =
        (if x <= a then 0 else (if x < b then (x - a) / (b - a) else 1)))`--),
        REPEAT STRIP_TAC THEN
	REWRITE_TAC [uniform_rv_def] THEN
        KNOW_TAC `!s. (b - a) * std_unif_cont s + a =
                        (\x. (b - a) * x + a) (std_unif_cont s)` THEN1
                RW_TAC std_ss [] THEN
        KNOW_TAC `(if x <= a then 0 else (if x < b then (x - a) / (b - a) else 1)) =
        (\x. (if x <= a then 0 else (if x < b then (x - a) / (b - a) else 1))) x` THEN1
                RW_TAC std_ss [] THEN
        ONCE_ASM_REWRITE_TAC [] THEN
        MATCH_MP_TAC ITM THEN
        RW_TAC std_ss [UNIF_CDF_IS_CONT_CDF_FN, INV_CDF_FN_UNIF]);

val UNIF_RV_LE_IN_EVENTS_BERN = store_thm
    ("UNIF_RV_LE_IN_EVENTS_BERN",
    (--`!x a b. (a < b) ==>
        ({s | uniform_rv a b s <= x} IN events bern)`--),
        REPEAT STRIP_TAC THEN
	REWRITE_TAC [uniform_rv_def] THEN
        KNOW_TAC `!s. (b - a) * std_unif_cont s + a =
                        (\x. (b - a) * x + a) (std_unif_cont s)` THEN1
                RW_TAC std_ss [] THEN
        ONCE_ASM_REWRITE_TAC [] THEN
        MATCH_MP_TAC FN_STD_UNIF_CONT_EVENTS_BERN THEN
        EXISTS_TAC (--`(\x. (if x <= a then 0 else 
		(if x < b then (x - a) / (b - a) else 1)))`--) THEN
        RW_TAC std_ss [UNIF_CDF_IS_CONT_CDF_FN, INV_CDF_FN_UNIF]);

val UNIF_RV_EQ_IN_EVENTS_BERN = store_thm
    ("UNIF_RV_EQ_IN_EVENTS_BERN",
    (--`!x a b. (a < b) ==>
        ({s | uniform_rv a b s = x} IN events bern)`--),
        REPEAT STRIP_TAC THEN
	REWRITE_TAC [uniform_rv_def] THEN
        KNOW_TAC `!s. (b - a) * std_unif_cont s + a =
                        (\x. (b - a) * x + a) (std_unif_cont s)` THEN1
                RW_TAC std_ss [] THEN
        ONCE_ASM_REWRITE_TAC [] THEN
        MATCH_MP_TAC FN_STD_UNIF_CONT_EQ_EVENTS_BERN THEN
        EXISTS_TAC (--`(\x. (if x <= a then 0 else 
		(if x < b then (x - a) / (b - a) else 1)))`--) THEN
        RW_TAC std_ss [UNIF_CDF_IS_CONT_CDF_FN, INV_CDF_FN_UNIF]);

val PMF_UNIF = store_thm
   ("PMF_UNIF",
   (--`!x a b. (a < b) ==>
        (prob bern {s | uniform_rv a b s = x} = 0)`--),
	REPEAT STRIP_TAC THEN
	REWRITE_TAC [uniform_rv_def] THEN
	KNOW_TAC `!s. ((b - a) * (std_unif_cont s) + a) = 
		(\x. ((b - a) * (std_unif_cont x) + a)) s` THEN1
		RW_TAC std_ss [] THEN
	ONCE_ASM_REWRITE_TAC [] THEN
	MATCH_MP_TAC CONT_PROB_BERN_EQ_0 THEN
	POP_ASSUM K_TAC THEN
	RW_TAC std_ss [CDF_def, CDF_in_events_bern_def, UNIF_RV_EQ_IN_EVENTS_BERN,
		UNIF_RV_LE_IN_EVENTS_BERN, CDF_UNIF, UNIF_CDF_CONT, GSYM uniform_rv_def]);





(*=======================
Exponential Distribution
=========================*)
val INV_CDF_FN_EXP = store_thm
    ("INV_CDF_FN_EXP",
    (--`!l. (0 < l) ==>
        (INV_CDF_FN (\x. (~1/l) * ln (1 - x)) (\x. if x <= 0 then 0 else (1 - exp (~l * x))))`--),
	RW_TAC real_ss [INV_CDF_FN_def, LN_1] THENL [
		Cases_on `x <= 0` THEN1
			FULL_SIMP_TAC std_ss [] THEN
		FULL_SIMP_TAC std_ss [] THEN	
		FULL_SIMP_TAC real_ss [REAL_EQ_SUB_RADD, EXP_NEG, REAL_INV_1OVER,
		REAL_EQ_RDIV_EQ, EXP_POS_LT, REAL_LT_MUL, REAL_NOT_LE] THEN
		FULL_SIMP_TAC std_ss [GSYM EXP_0, EXP_INJ] THEN
		KNOW_TAC `0 < l * x` THEN1
			PROVE_TAC [REAL_LT_MUL] THEN
		POP_ASSUM MP_TAC THEN
		RW_TAC std_ss [REAL_NOT_LT, REAL_LE_REFL],

		Cases_on `x <= 0` THEN1
			FULL_SIMP_TAC real_ss [] THEN
		FULL_SIMP_TAC std_ss [] THEN
		FULL_SIMP_TAC real_ss [REAL_EQ_SUB_RADD, EXP_NEG, REAL_INV_1OVER] THEN
		POP_ASSUM MP_TAC THEN
		POP_ASSUM MP_TAC THEN
		ONCE_REWRITE_TAC [REAL_ADD_SYM] THEN
		FULL_SIMP_TAC real_ss [GSYM REAL_EQ_SUB_RADD, REAL_EQ_RDIV_EQ, EXP_POS_LT,
			REAL_LT_MUL, REAL_NOT_LE],
	
		Cases_on `x <= 0` THEN1
			FULL_SIMP_TAC real_ss [] THEN
		FULL_SIMP_TAC real_ss [LN_EXP, real_div, REAL_MUL_ASSOC, REAL_MUL_LINV, REAL_POS_NZ],

		KNOW_TAC `~1 / l * ln (1 - x') <= 0 = F` THEN1 (
			RW_TAC real_ss [REAL_NOT_LE, GSYM REAL_NEG_LMUL, real_div] THEN
			ONCE_REWRITE_TAC [REAL_MUL_SYM] THEN
			REWRITE_TAC [REAL_NEG_LMUL] THEN
			MATCH_MP_TAC REAL_LT_MUL THEN
			KNOW_TAC `0 < 1 - x'` THEN1
				RW_TAC std_ss [REAL_SUB_LT] THEN
			RW_TAC std_ss [REAL_INV_POS, GSYM LN_INV, GSYM LN_1] THEN
			KNOW_TAC `ln 1 < ln (inv (1 - x')) = 1 < inv (1 - x')` THEN1 (
				MATCH_MP_TAC LN_MONO_LT THEN
				RW_TAC real_ss [REAL_INV_POS]) THEN
			ASM_REWRITE_TAC [] THEN
			RW_TAC real_ss [REAL_INV_1OVER, REAL_LT_RDIV_EQ, 
					REAL_LT_SUB_RADD, REAL_LT_ADDR]) THEN
 		FULL_SIMP_TAC real_ss [real_div, REAL_MUL_ASSOC, REAL_MUL_RINV, REAL_POS_NZ] THEN
		KNOW_TAC `exp (ln (1 - x')) = 1 - x'` THEN1
			RW_TAC std_ss [EXP_LN, REAL_SUB_LT] THEN
		ASM_REWRITE_TAC [] THEN
		RW_TAC real_ss []]);

val EXP_CDF_MONO = store_thm
    ("EXP_CDF_MONO",
    (--`!l. (0 < l) ==> (!c d. (c < d) ==>
	((\x. if x <= 0 then 0 else (1 - exp (~l * x))) c <= 
	 (\x. if x <= 0 then 0 else (1 - exp (~l * x))) d))`--),
	RW_TAC std_ss [] THEN
	Cases_on `c <= 0` THENL [
		Cases_on `d <= 0` THEN1
			RW_TAC std_ss [REAL_LE_REFL] THEN
		RW_TAC real_ss [REAL_LE_SUB_LADD, GSYM EXP_0, EXP_MONO_LE, REAL_LE_LNEG] THEN
		PROVE_TAC [REAL_LT_IMP_LE, REAL_LT_MUL, REAL_NOT_LE],
	
		Cases_on `d <= 0` THEN1
			PROVE_TAC [REAL_NOT_LE, REAL_LET_TRANS, REAL_LT_ANTISYM] THEN
		RW_TAC std_ss [] THEN
		FULL_SIMP_TAC std_ss [REAL_NOT_LE] THEN
		RW_TAC std_ss [REAL_LE_SUB_RADD] THEN
		ONCE_REWRITE_TAC [REAL_ADD_SYM] THEN
		RW_TAC real_ss [GSYM REAL_SUB_ASSOC2, GSYM REAL_LE_SUB_RADD, EXP_MONO_LE,
		REAL_LE_NEG, REAL_LE_LMUL, REAL_LT_IMP_LE]]);



val EXP_DIFF_COMPOSITE = store_thm
   ("EXP_DIFF_COMPOSITE",
   (--`!g m x.
        ((g diffl m) x ==> ((\x. exp (g x)) diffl (exp (g x) * m)) x)`--),
	RW_TAC std_ss [DIFF_COMPOSITE]);

val ONE_MINUS_EXP_CONT = store_thm
   ("ONE_MINUS_EXP_CONT",
   (--`!x l. (0 < l) ==>
        (\x. 1 - exp (~l * x)) contl x`--),
	RW_TAC  std_ss [] THEN
	KNOW_TAC `(\x. 1 - exp (~l * x)) =
        (\x.(\x. 1) x - (\x. exp (~l * x)) x)` THEN1
        	RW_TAC std_ss [] THEN
	ONCE_ASM_REWRITE_TAC [] THEN
	MATCH_MP_TAC CONT_SUB THEN
	RW_TAC std_ss [CONT_CONST] THEN
	MATCH_MP_TAC DIFF_CONT THEN
	EXISTS_TAC (--`exp(~l * (x:real)) * ~l`--) THEN
	KNOW_TAC `!l x. exp (~l * x) = exp ((\y. ~l * y) x)` THEN1
        	RW_TAC std_ss [] THEN
	ASM_REWRITE_TAC [] THEN
	MATCH_MP_TAC EXP_DIFF_COMPOSITE THEN
	KNOW_TAC `((\y. ~l * y) diffl ~l) x =
                ((\x. ~l * (\x. x) x) diffl (~l * 1)) x` THEN1
        	RW_TAC real_ss [] THEN
	ASM_REWRITE_TAC [] THEN
	MATCH_MP_TAC DIFF_CMUL THEN
	RW_TAC std_ss [DIFF_X]);

val EXP_CDF_CONT = store_thm
   ("EXP_CDF_CONT",
   (--`!x l. (0 < l) ==>
	((\x. (if x <= 0 then 0 else 1 - exp (~l * x))) contl x)`--),
	REPEAT STRIP_TAC THEN
	MP_TAC ONE_MINUS_EXP_CONT THEN
	DISCH_TAC THEN
	RW_TAC real_ss [contl, LIM] THENL [
		POP_ASSUM MP_TAC THEN
		POP_ASSUM MP_TAC THEN
		POP_ASSUM (MP_TAC o Q.SPECL [`0`, `l`]) THEN
		RW_TAC real_ss [contl, LIM] THEN
		POP_ASSUM MP_TAC THEN
		POP_ASSUM MP_TAC THEN
		POP_ASSUM (MP_TAC o Q.SPEC `e`) THEN		
		RW_TAC real_ss [REAL_NEG_0, EXP_0] THEN
		FULL_SIMP_TAC std_ss [] THEN	
		EXISTS_TAC (--`d:real`--) THEN
		RW_TAC real_ss [ABS_0] THEN
		KNOW_TAC `0 < abs (x + h)` THEN1 (
			MATCH_MP_TAC REAL_LTE_TRANS THEN
			EXISTS_TAC (--`((x:real) + (h:real))`--) THEN
			FULL_SIMP_TAC std_ss [REAL_NOT_LE, REAL_LT_IMP_LE, ABS_LE]) THEN
		KNOW_TAC `abs (x + h) < d` THEN1 (
			MATCH_MP_TAC REAL_LET_TRANS THEN
			EXISTS_TAC (--`(abs (h:real))`--) THEN
			KNOW_TAC `abs (x + h) = x + h` THEN1
				PROVE_TAC [ABS_REFL, REAL_NOT_LE, REAL_LT_IMP_LE] THEN
			KNOW_TAC `abs (h) = h` THEN1 (
				FULL_SIMP_TAC std_ss [ABS_REFL, REAL_NOT_LE] THEN
				MATCH_MP_TAC REAL_LE_TRANS THEN
				EXISTS_TAC (--`(x:real) + (h:real)`--) THEN
				RW_TAC real_ss [GSYM REAL_LE_SUB_LADD]) THEN
			FULL_SIMP_TAC real_ss [GSYM REAL_LE_SUB_LADD]) THEN
		PROVE_TAC [],

	
		POP_ASSUM MP_TAC THEN
        	POP_ASSUM MP_TAC THEN
        	POP_ASSUM (MP_TAC o Q.SPECL [`x`, `l`]) THEN
		RW_TAC real_ss [contl, LIM, REAL_NOT_LE] THEN
        	POP_ASSUM MP_TAC THEN
        	POP_ASSUM MP_TAC THEN
        	POP_ASSUM (MP_TAC o Q.SPEC `e`) THEN
		RW_TAC real_ss [] THEN
		FULL_SIMP_TAC std_ss [] THEN
        	EXISTS_TAC (--`d:real`--) THEN
		RW_TAC real_ss [ABS_0] THEN
		NTAC 5 (POP_ASSUM MP_TAC) THEN
		POP_ASSUM (MP_TAC o Q.SPEC `~x`) THEN
		RW_TAC real_ss [REAL_ADD_RINV, REAL_NEG_0, EXP_0] THEN
		KNOW_TAC `0 < abs (~x)` THEN1 (
			MATCH_MP_TAC REAL_LTE_TRANS THEN
			EXISTS_TAC (--`x:real`--) THEN
			RW_TAC std_ss [ABS_NEG, ABS_LE]) THEN
		KNOW_TAC `abs ~x < d` THEN1 (
			MATCH_MP_TAC REAL_LET_TRANS THEN
			EXISTS_TAC (--`abs (h:real)`--) THEN
			RW_TAC std_ss [ABS_NEG] THEN
			MATCH_MP_TAC REAL_LE_TRANS THEN
			EXISTS_TAC (--`~(h:real)`--) THEN
			KNOW_TAC `abs x = x` THEN1
				PROVE_TAC [ABS_REFL, REAL_LT_IMP_LE] THEN
			KNOW_TAC `abs h = abs (~ h)` THEN1
				PROVE_TAC [ABS_NEG] THEN
			RW_TAC std_ss [ABS_LE, REAL_LE_RNEG]) THEN
		PROVE_TAC []]);

val EXP_CDF_AT_NEGINF = store_thm
   ("EXP_CDF_AT_NEGINF",
   (--`!l. (0 < l) ==> 
	(((\n.(\x. if x <= 0 then 0 else (1 - exp (~l * x))) ((\n. ~&n) n)) --> 0))`--),
	RW_TAC std_ss [SEQ] THEN
	EXISTS_TAC (--`1:num`--) THEN
	RW_TAC real_ss [ABS_0]);

val EXP_CDF_AT_POSINF = store_thm
   ("EXP_CDF_AT_POSINF",
   (--`!l. (0 < l) ==> 
	(((\n.(\x. if x <= 0 then 0 else (1 - exp (~l * x))) ((\n. &n) n)) --> 1))`--),
	REPEAT STRIP_TAC THEN
	ONCE_REWRITE_TAC [SEQ_SUC] THEN
	RW_TAC arith_ss [REAL_LE] THEN
	KNOW_TAC `(\n. 1 - exp (~l * & (SUC n))) = 
		(\n. (\n. 1) n - (\n. exp (~l * & (SUC n))) n)` THEN1
		RW_TAC std_ss [] THEN
	KNOW_TAC `1 = 1 - 0` THEN1
		RW_TAC real_ss [] THEN
	ONCE_ASM_REWRITE_TAC [] THEN
	MATCH_MP_TAC SEQ_SUB THEN
	RW_TAC std_ss [SEQ_CONST] THEN
	ONCE_REWRITE_TAC [REAL_MUL_SYM] THEN
	RW_TAC std_ss [EXP_N, EXP_NEG] THEN
	KNOW_TAC `(\n. inv (exp l) pow SUC n) = 
		(\n. (\n. inv (exp l) pow n) (SUC n))` THEN1
		RW_TAC std_ss [] THEN
	ASM_REWRITE_TAC [] THEN
	ONCE_REWRITE_TAC [GSYM SEQ_SUC] THEN
	MATCH_MP_TAC SEQ_POWER THEN
	KNOW_TAC `abs (inv (exp l)) = (inv (exp l))` THEN1
		RW_TAC std_ss [ABS_REFL, REAL_INV_POS, EXP_POS_LT, REAL_LT_IMP_LE] THEN
	FULL_SIMP_TAC real_ss [REAL_INV_1OVER, EXP_POS_LT, REAL_LT_LDIV_EQ, EXP_LT_1]);



val EXP_CDF_IS_CONT_CDF_FN = store_thm
    ("EXP_CDF_IS_CONT_CDF_FN",
    (--`!l. (0 < l) ==>
        (IS_CONT_CDF_FN (\x. if x <= 0 then 0 else (1 - exp (~l * x))))`--),
        REPEAT STRIP_TAC THEN
        REWRITE_TAC [IS_CONT_CDF_FN_def] THEN
        RW_TAC std_ss [EXP_CDF_CONT, EXP_CDF_MONO, EXP_CDF_AT_POSINF, EXP_CDF_AT_NEGINF]);

val CDF_EXP = store_thm
    ("CDF_EXP",
    (--`!x l. (0 < l) ==>
        (prob bern {s | exp_rv l s <= x} =
                        (if x <= 0 then 0 else (1 - exp (~l * x))))`--),
        REPEAT STRIP_TAC THEN
	REWRITE_TAC [exp_rv_def] THEN
        KNOW_TAC `!s. ((~1/l) * ln (1 - std_unif_cont s)) =
                (\x. (~1/l) * ln (1 - x)) (std_unif_cont s)` THEN1
                RW_TAC std_ss [] THEN
        KNOW_TAC `(if x <= 0 then 0 else (1 - exp (~l * x))) =
        (\x. (if x <= 0 then 0 else (1 - exp (~l * x)))) x` THEN1
                RW_TAC std_ss [] THEN
        ONCE_ASM_REWRITE_TAC [] THEN
        MATCH_MP_TAC ITM THEN
        RW_TAC std_ss [EXP_CDF_IS_CONT_CDF_FN, INV_CDF_FN_EXP]);

val EXP_RV_LE_IN_EVENTS_BERN = store_thm
    ("EXP_RV_LE_IN_EVENTS_BERN",
    (--`!x l. (0 < l) ==>
        ({s | exp_rv l s <= x} IN events bern)`--),
        REPEAT STRIP_TAC THEN
	REWRITE_TAC [exp_rv_def] THEN
        KNOW_TAC `!s. ((~1/l) * ln (1 - std_unif_cont s)) =
                (\x. (~1/l) * ln (1 - x)) (std_unif_cont s)` THEN1
                RW_TAC std_ss [] THEN
        ONCE_ASM_REWRITE_TAC [] THEN
        MATCH_MP_TAC FN_STD_UNIF_CONT_EVENTS_BERN THEN
        EXISTS_TAC (--`(\x. (if x <= 0 then 0 else (1 - exp (~l * x))))`--) THEN
        RW_TAC std_ss [EXP_CDF_IS_CONT_CDF_FN, INV_CDF_FN_EXP]);

val EXP_RV_EQ_IN_EVENTS_BERN = store_thm
   ("EXP_RV_EQ_IN_EVENTS_BERN",
   (--`!x l. (0 < l) ==>
        ({s | exp_rv l s = x} IN events bern)`--),
        REPEAT STRIP_TAC THEN
	REWRITE_TAC [exp_rv_def] THEN
        KNOW_TAC `!s. ((~1/l) * ln (1 - std_unif_cont s)) =
                (\x. (~1/l) * ln (1 - x)) (std_unif_cont s)` THEN1
                RW_TAC std_ss [] THEN
        ONCE_ASM_REWRITE_TAC [] THEN
        MATCH_MP_TAC FN_STD_UNIF_CONT_EQ_EVENTS_BERN THEN
        EXISTS_TAC (--`(\x. (if x <= 0 then 0 else (1 - exp (~l * x))))`--) THEN
        RW_TAC std_ss [EXP_CDF_IS_CONT_CDF_FN, INV_CDF_FN_EXP]);
	

val PMF_EXP = store_thm
    ("PMF_EXP",
    (--`!x l. (0 < l) ==>
        (prob bern {s | exp_rv l s  = x} = 0)`--),
	REPEAT STRIP_TAC THEN
	REWRITE_TAC [exp_rv_def] THEN
	KNOW_TAC `!s. (~1/l) * ln (1 - std_unif_cont s) = 
		(\x. (~1/l) * ln (1 - std_unif_cont x)) s` THEN1
		RW_TAC std_ss [] THEN
	ONCE_ASM_REWRITE_TAC [] THEN
	MATCH_MP_TAC CONT_PROB_BERN_EQ_0 THEN
	POP_ASSUM K_TAC THEN
	RW_TAC std_ss [CDF_def, CDF_in_events_bern_def, EXP_RV_EQ_IN_EVENTS_BERN,
		EXP_RV_LE_IN_EVENTS_BERN, CDF_EXP, EXP_CDF_CONT, GSYM exp_rv_def]);



(*=======================
  Rayleigh Distribution
========================*);

val INV_CDF_FN_RAYLEIGH = store_thm
   ("INV_CDF_FN_RAYLEIG",
   (--`!sig. (0 < sig) ==>
	(INV_CDF_FN (\x. sig * sqrt (~2 * ln (1 - x))) 
		(\x. (if x <= 0 then 0 else (1 - exp (~(x pow 2)/ (2 * (sig pow 2)))))))`--),
	RW_TAC real_ss [INV_CDF_FN_def, LN_1, REAL_NEG_0, SQRT_0] THENL [
		Cases_on `x <= 0` THEN1
			FULL_SIMP_TAC std_ss [] THEN
		FULL_SIMP_TAC real_ss [EXP_NEG, real_div, REAL_EQ_SUB_RADD] THEN
		FULL_SIMP_TAC real_ss [REAL_INV_1OVER, REAL_NOT_LE] THEN
		KNOW_TAC `0 < (x pow 2 * (1 / (2 * sig pow 2)))` THEN1 (
			MATCH_MP_TAC REAL_LT_MUL THEN
			KNOW_TAC `0 < 2 * (sig pow 2)` THEN1 (
				MATCH_MP_TAC REAL_LT_MUL THEN	
				RW_TAC real_ss [REAL_POW_LT]) THEN
			RW_TAC real_ss [real_div, REAL_POW_LT, REAL_INV_POS]) THEN
		KNOW_TAC `1 <  exp (x pow 2 * (1 / (2 * sig pow 2)))` THEN1
			RW_TAC std_ss [EXP_LT_1] THEN
		FULL_SIMP_TAC real_ss [REAL_EQ_RDIV_EQ, EXP_POS_LT] THEN	
		PROVE_TAC [REAL_LT_REFL],	

		Cases_on `x <= 0` THEN1
			FULL_SIMP_TAC real_ss [] THEN
		FULL_SIMP_TAC real_ss [EXP_NEG, real_div, REAL_EQ_SUB_RADD] THEN
		NTAC 2 (POP_ASSUM MP_TAC) THEN
		ONCE_REWRITE_TAC [REAL_ADD_SYM] THEN
		RW_TAC std_ss [GSYM REAL_EQ_SUB_RADD, REAL_NOT_LE, REAL_INV_1OVER] THEN
		FULL_SIMP_TAC real_ss [REAL_EQ_RDIV_EQ, EXP_POS_LT],	
	
		Cases_on `x <= 0` THEN1
			FULL_SIMP_TAC real_ss [] THEN
		FULL_SIMP_TAC real_ss [LN_EXP, real_div, REAL_INV_MUL, REAL_POW_LT, REAL_POS_NZ, REAL_MUL_ASSOC] THEN
		KNOW_TAC `(2 * x pow 2 * inv 2 * inv (sig pow 2)) = (x pow 2 * inv (sig pow 2))` THEN1 (
			ONCE_REWRITE_TAC [REAL_MUL_SYM] THEN
			RW_TAC std_ss [GSYM REAL_EQ_LMUL2, REAL_INV_POS, REAL_POW_LT, REAL_POS_NZ] THEN
			ONCE_REWRITE_TAC [REAL_MUL_SYM] THEN
			ONCE_REWRITE_TAC [REAL_MUL_ASSOC] THEN
			ONCE_REWRITE_TAC [REAL_MUL_SYM] THEN
			RW_TAC real_ss [REAL_MUL_LINV, REAL_POS_NZ]) THEN
		FULL_SIMP_TAC std_ss [REAL_NOT_LE] THEN
		RW_TAC std_ss [SQRT_MUL, REAL_INV_POS, REAL_POW_LT, REAL_LT_IMP_LE] THEN
		RW_TAC real_ss [POW_2_SQRT, REAL_LT_IMP_LE, REAL_INV_1OVER, 
		SQRT_DIV, REAL_POW_LT, SQRT_1, REAL_MUL_ASSOC] THEN
		ONCE_REWRITE_TAC [REAL_MUL_SYM] THEN
		RW_TAC real_ss [real_div, REAL_MUL_LINV, REAL_POS_NZ, REAL_MUL_ASSOC],

		KNOW_TAC `sig * sqrt ~(2 * ln (1 - x')) <= 0 = F` THEN1 (
			RW_TAC real_ss [REAL_NOT_LE] THEN
			MATCH_MP_TAC REAL_LT_MUL THEN
			KNOW_TAC `0 < (1 - x')` THEN1
				RW_TAC std_ss [REAL_SUB_LT] THEN
			RW_TAC std_ss [REAL_NEG_RMUL, GSYM LN_INV] THEN
			KNOW_TAC `1 < inv (1 - x')` THEN1 
				RW_TAC real_ss [REAL_INV_1OVER, REAL_LT_RDIV_EQ, 
					REAL_LT_SUB_RADD, REAL_LT_ADDR] THEN
			RW_TAC real_ss [LN_POS, SQRT_MUL, REAL_LT_IMP_LE] THEN
			MATCH_MP_TAC REAL_LT_MUL THEN
			RW_TAC real_ss [SQRT_POS_LT] THEN
			MATCH_MP_TAC SQRT_POS_LT THEN
			RW_TAC real_ss [GSYM LN_1, REAL_INV_POS, LN_MONO_LT]) THEN
		FULL_SIMP_TAC real_ss [] THEN
		KNOW_TAC `0 < (1 - x')` THEN1
			RW_TAC std_ss [REAL_SUB_LT] THEN
		RW_TAC std_ss [REAL_NEG_RMUL, GSYM LN_INV, POW_MUL] THEN
		KNOW_TAC `1 < inv (1 - x')` THEN1
			RW_TAC real_ss [REAL_INV_1OVER, REAL_LT_RDIV_EQ, 
				REAL_LT_SUB_RADD, REAL_LT_ADDR] THEN
		RW_TAC real_ss [LN_POS, SQRT_MUL, REAL_LT_IMP_LE, POW_MUL, SQRT_POW_2, REAL_MUL_ASSOC] THEN
		ONCE_REWRITE_TAC [REAL_MUL_SYM] THEN
		RW_TAC real_ss [real_div, GSYM REAL_MUL_ASSOC] THEN
		KNOW_TAC `(sig pow 2 * (2 * inv (sig pow 2 * 2))) = 1` THEN1 (
			REWRITE_TAC [REAL_MUL_ASSOC] THEN
			ONCE_REWRITE_TAC [REAL_MUL_SYM] THEN
			RW_TAC real_ss [REAL_MUL_LINV, REAL_POS_NZ,
				REAL_POW_LT, REAL_LT_MUL]) THEN
		RW_TAC real_ss [EXP_NEG] THEN
		KNOW_TAC `(exp (ln (inv (1 - x')))) = inv (1 - x')` THEN1 (
			RW_TAC std_ss [EXP_LN] THEN
			MATCH_MP_TAC REAL_LT_TRANS THEN
			EXISTS_TAC (--`1:real`--) THEN
			RW_TAC real_ss []) THEN
		RW_TAC real_ss [REAL_INV_INV]]);

val RAYLEIGH_CDF_MONO = store_thm
    ("RAYLEIGH_CDF_MONO",
    (--`!sig. (0 < sig) ==> (!c d. (c < d) ==>
	((\x. (if x <= 0 then 0 else (1 - exp (~(x pow 2)/ (2 * (sig pow 2)))))) c <= 
	 (\x. (if x <= 0 then 0 else (1 - exp (~(x pow 2)/ (2 * (sig pow 2)))))) d))`--),
	RW_TAC std_ss [] THEN
	Cases_on `c <= 0` THENL [
		Cases_on `d <= 0` THEN1
			RW_TAC std_ss [REAL_LE_REFL] THEN
		RW_TAC real_ss [REAL_LE_SUB_LADD, GSYM EXP_0, EXP_MONO_LE, REAL_LE_LNEG, 
		REAL_LE_LDIV_EQ, REAL_POW_LT, REAL_LT_MUL] THEN
		PROVE_TAC [REAL_LT_IMP_LE, REAL_NOT_LE, REAL_POW_LT],
	
		Cases_on `d <= 0` THEN1
			PROVE_TAC [REAL_NOT_LE, REAL_LET_TRANS, REAL_LT_ANTISYM] THEN
		RW_TAC std_ss [] THEN
		FULL_SIMP_TAC std_ss [REAL_NOT_LE] THEN
		RW_TAC std_ss [REAL_LE_SUB_RADD] THEN
		ONCE_REWRITE_TAC [REAL_ADD_SYM] THEN
		RW_TAC real_ss [GSYM REAL_SUB_ASSOC2, GSYM REAL_LE_SUB_RADD, EXP_MONO_LE,
		REAL_LE_NEG, real_div, REAL_LE_RMUL, REAL_POW_LT, REAL_LT_MUL, POW_LE, REAL_LT_IMP_LE]]);


val ONE_MINUS_RAYLEIGH_CONT = store_thm 
    ("ONE_MINUS_RAYLEIGH_CONT",
    (--`!x sig. (0 < sig) ==>
        ((\x. 1 - exp (~(x pow 2)/ (2 * (sig pow 2)))) contl x)`--),
	RW_TAC  std_ss [] THEN
	KNOW_TAC `(\x. 1 - exp (~(x pow 2)/ (2 * (sig pow 2)))) =
        (\x. (\x. 1) x - (\x. exp (~(x pow 2)/ (2 * (sig pow 2)))) x)` THEN1
        	RW_TAC std_ss [] THEN
	ONCE_ASM_REWRITE_TAC [] THEN
	MATCH_MP_TAC CONT_SUB THEN
	RW_TAC std_ss [CONT_CONST] THEN
	MATCH_MP_TAC DIFF_CONT THEN
	EXISTS_TAC (--`exp (~(x pow 2)/ (2 * (sig pow 2))) * ~(2 * x)/ (2 * (sig pow 2))`--) THEN
	KNOW_TAC `!sig x. exp (~(x pow 2)/ (2 * (sig pow 2))) = exp ((\y. ~(y pow 2)/ (2 * (sig pow 2))) x)` THEN1
        	RW_TAC std_ss [] THEN
	ASM_REWRITE_TAC [real_div, GSYM REAL_MUL_ASSOC] THEN
	MATCH_MP_TAC EXP_DIFF_COMPOSITE THEN
	ONCE_REWRITE_TAC [REAL_MUL_SYM] THEN
	KNOW_TAC `((\y. inv (2 * sig pow 2) * ~(y pow 2)) diffl
     			(inv (2 * sig pow 2) * ~(2 * x))) x =
		((\x. ~inv (2 * sig pow 2) * (\x. (x pow 2)) x) diffl
			(~inv (2 * sig pow 2) * (2 * x))) x` THEN1
		RW_TAC real_ss [] THEN
	ASM_REWRITE_TAC [] THEN
	MATCH_MP_TAC DIFF_CMUL THEN
	MP_TAC (Q.SPECL [`2`, `x`] DIFF_POW) THEN
	RW_TAC real_ss [POW_1]);

val RAYLEIGH_CDF_CONT = store_thm
   ("RAYLEIGH_CDF_CONT",
   (--`!x sig. (0 < sig) ==>
        ((\x. (if x <= 0 then 0 else (1 - exp (~(x pow 2)/ (2 * (sig pow 2)))))) contl x)`--),
        REPEAT STRIP_TAC THEN
        MP_TAC ONE_MINUS_RAYLEIGH_CONT THEN
        DISCH_TAC THEN
        RW_TAC real_ss [contl, LIM] THENL [
                POP_ASSUM MP_TAC THEN
                POP_ASSUM MP_TAC THEN
                POP_ASSUM (MP_TAC o Q.SPECL [`0`, `sig`]) THEN
                RW_TAC real_ss [contl, LIM] THEN
                POP_ASSUM MP_TAC THEN
                POP_ASSUM MP_TAC THEN
                POP_ASSUM (MP_TAC o Q.SPEC `e`) THEN
                RW_TAC real_ss [REAL_NEG_0, EXP_0] THEN
                FULL_SIMP_TAC std_ss [] THEN
                EXISTS_TAC (--`d:real`--) THEN
                RW_TAC real_ss [ABS_0] THEN
                KNOW_TAC `0 < abs (x + h)` THEN1 (
                        MATCH_MP_TAC REAL_LTE_TRANS THEN
                        EXISTS_TAC (--`((x:real) + (h:real))`--) THEN
                        FULL_SIMP_TAC std_ss [REAL_NOT_LE, REAL_LT_IMP_LE, ABS_LE]) THEN
                KNOW_TAC `abs (x + h) < d` THEN1 (
                        MATCH_MP_TAC REAL_LET_TRANS THEN
                        EXISTS_TAC (--`(abs (h:real))`--) THEN
                        KNOW_TAC `abs (x + h) = x + h` THEN1
                                PROVE_TAC [ABS_REFL, REAL_NOT_LE, REAL_LT_IMP_LE] THEN
                        KNOW_TAC `abs (h) = h` THEN1 (
                                FULL_SIMP_TAC std_ss [ABS_REFL, REAL_NOT_LE] THEN
                                MATCH_MP_TAC REAL_LE_TRANS THEN
                                EXISTS_TAC (--`(x:real) + (h:real)`--) THEN
                                RW_TAC real_ss [GSYM REAL_LE_SUB_LADD]) THEN
                        FULL_SIMP_TAC real_ss [GSYM REAL_LE_SUB_LADD]) THEN
                PROVE_TAC [],


                POP_ASSUM MP_TAC THEN
                POP_ASSUM MP_TAC THEN
                POP_ASSUM (MP_TAC o Q.SPECL [`x`, `sig`]) THEN
                RW_TAC real_ss [contl, LIM, REAL_NOT_LE] THEN
                POP_ASSUM MP_TAC THEN
                POP_ASSUM MP_TAC THEN
                POP_ASSUM (MP_TAC o Q.SPEC `e`) THEN
                RW_TAC real_ss [] THEN
                FULL_SIMP_TAC std_ss [] THEN
                EXISTS_TAC (--`d:real`--) THEN
                RW_TAC real_ss [ABS_0] THEN
                NTAC 5 (POP_ASSUM MP_TAC) THEN
                POP_ASSUM (MP_TAC o Q.SPEC `~x`) THEN
                RW_TAC real_ss [REAL_ADD_RINV, REAL_NEG_0, EXP_0] THEN
                KNOW_TAC `0 < abs (~x)` THEN1 (
                        MATCH_MP_TAC REAL_LTE_TRANS THEN
                        EXISTS_TAC (--`x:real`--) THEN
                        RW_TAC std_ss [ABS_NEG, ABS_LE]) THEN
                KNOW_TAC `abs ~x < d` THEN1 (
                        MATCH_MP_TAC REAL_LET_TRANS THEN
                        EXISTS_TAC (--`abs (h:real)`--) THEN
                        RW_TAC std_ss [ABS_NEG] THEN
                        MATCH_MP_TAC REAL_LE_TRANS THEN
                        EXISTS_TAC (--`~(h:real)`--) THEN
                        KNOW_TAC `abs x = x` THEN1
                                PROVE_TAC [ABS_REFL, REAL_LT_IMP_LE] THEN
                        KNOW_TAC `abs h = abs (~ h)` THEN1
                                PROVE_TAC [ABS_NEG] THEN
                        RW_TAC std_ss [ABS_LE, REAL_LE_RNEG]) THEN
                PROVE_TAC []]);

val RAYLEIGH_CDF_AT_NEGINF = store_thm
  ("RAYLEIGH_CDF_AT_NEGINF",
  (--`!sig. (0 < sig) ==> 
	(((\n.(\x. (if x <= 0 then 0 else 
		(1 - exp (~(x pow 2)/ (2 * (sig pow 2)))))) ((\n. ~&n) n)) --> 0))`--),
	RW_TAC std_ss [SEQ] THEN
	EXISTS_TAC (--`1:num`--) THEN
	RW_TAC real_ss [ABS_0]);

val POWPOW_PLUS1 = store_thm
  ("POWPOW_PLUS1",
  (--`!e. 0 < e ==> !n. 1 + & n * e <= (1 + e) pow (n * n)`--),
	GEN_TAC THEN DISCH_TAC THEN
	Induct THEN1
		RW_TAC real_ss [] THEN
	
		RW_TAC real_ss [ADD1, LEFT_ADD_DISTRIB, RIGHT_ADD_DISTRIB] THEN
		RW_TAC std_ss [POW_ADD, ADD_ASSOC, REAL_MUL_ASSOC, POW_1, REAL_ADD_ASSOC,
			GSYM REAL_ADD, REAL_ADD_RDISTRIB, REAL_ADD_LDISTRIB] THEN
		MATCH_MP_TAC REAL_LE_ADD2 THEN
		RW_TAC real_ss [] THENL [
				MATCH_MP_TAC REAL_LE_TRANS THEN
				EXISTS_TAC (--`(1 + e) pow (n * n)`--) THEN
				ASM_REWRITE_TAC [] THEN
				KNOW_TAC `0 < (1 + e) pow (n * n)` THEN1 (
					MATCH_MP_TAC REAL_POW_LT THEN
					RW_TAC real_ss [REAL_LT_ADD]) THEN
				RW_TAC std_ss [GSYM REAL_LE_LDIV_EQ, REAL_DIV_REFL, REAL_POS_NZ] THEN
				POP_ASSUM K_TAC THEN
				POP_ASSUM K_TAC THEN			
				Induct_on `n` THEN1
					RW_TAC real_ss [] THEN

					RW_TAC real_ss [] THEN	
					MATCH_MP_TAC REAL_LE_TRANS THEN
					EXISTS_TAC (--`(1 + e) pow (2 * n)`--) THEN
					RW_TAC std_ss [ADD1, LEFT_ADD_DISTRIB, POW_ADD] THEN
					ONCE_REWRITE_TAC [REAL_MUL_SYM] THEN
					KNOW_TAC `0 < (1 + e) pow (2 * n)` THEN1 (
						MATCH_MP_TAC REAL_POW_LT THEN
						RW_TAC real_ss [REAL_LT_ADD]) THEN
					RW_TAC std_ss [GSYM REAL_LE_LDIV_EQ, 
					REAL_DIV_REFL, REAL_POS_NZ] THEN				
					MATCH_MP_TAC REAL_LE1_POW2 THEN
					RW_TAC std_ss [REAL_LE_ADDR, REAL_LT_IMP_LE],
		
		RW_TAC real_ss [GSYM REAL_LE_LDIV_EQ, REAL_DIV_REFL, REAL_POS_NZ] THEN	
		POP_ASSUM K_TAC THEN
		MATCH_MP_TAC REAL_LE_TRANS THEN
		EXISTS_TAC (--`1 * 1:real`--) THEN
		CONJ_TAC THEN1
			RW_TAC real_ss [] THEN
	
			MATCH_MP_TAC REAL_LE_MUL2 THEN
			RW_TAC real_ss [] THENL [
			Induct_on `n` THEN1
				RW_TAC real_ss [] THEN
			
				MATCH_MP_TAC REAL_LE_TRANS THEN
				EXISTS_TAC (--`(1 + e) pow (2 * n)`--) THEN
				RW_TAC std_ss [ADD1, LEFT_ADD_DISTRIB, POW_ADD] THEN
				ONCE_REWRITE_TAC [REAL_MUL_SYM] THEN
				KNOW_TAC `0 < (1 + e) pow (2 * n)` THEN1 (
						MATCH_MP_TAC REAL_POW_LT THEN
						RW_TAC real_ss [REAL_LT_ADD]) THEN
				RW_TAC std_ss [GSYM REAL_LE_LDIV_EQ, 
					REAL_DIV_REFL, REAL_POS_NZ] THEN	
				MATCH_MP_TAC REAL_LE1_POW2 THEN
				RW_TAC std_ss [REAL_LE_ADDR, REAL_LT_IMP_LE], 

			Induct_on `n` THEN1
				RW_TAC real_ss [] THEN
		
				MATCH_MP_TAC REAL_LE_TRANS THEN
				EXISTS_TAC (--`(1 + e) pow (n * n)`--) THEN
				RW_TAC std_ss [ADD1, LEFT_ADD_DISTRIB, RIGHT_ADD_DISTRIB, 
					POW_ADD, POW_1, GSYM REAL_MUL_ASSOC] THEN
				ONCE_REWRITE_TAC [REAL_MUL_SYM] THEN
				KNOW_TAC `0 < (1 + e) pow (n * n)` THEN1 (
					MATCH_MP_TAC REAL_POW_LT THEN
					RW_TAC real_ss [REAL_LT_ADD]) THEN
				RW_TAC std_ss [GSYM REAL_LE_LDIV_EQ, REAL_DIV_REFL, REAL_POS_NZ] THEN
				RW_TAC real_ss [REAL_MUL_ASSOC] THEN
				POP_ASSUM K_TAC THEN
				POP_ASSUM K_TAC THEN
				ONCE_REWRITE_TAC [REAL_MUL_SYM] THEN
				MATCH_MP_TAC REAL_LE_TRANS THEN
				EXISTS_TAC (--`1 * 1:real`--) THEN
				CONJ_TAC THEN1
					RW_TAC real_ss [] THEN
					MATCH_MP_TAC REAL_LE_MUL2 THEN
					RW_TAC real_ss [REAL_LE_ADDR, REAL_LT_IMP_LE] THEN
					MATCH_MP_TAC REAL_LE_TRANS THEN
					EXISTS_TAC (--`1 * 1:real`--) THEN
						CONJ_TAC THEN1				
							RW_TAC real_ss [] THEN	
							MATCH_MP_TAC REAL_LE_MUL2 THEN
							FULL_SIMP_TAC real_ss [] THEN
							Induct_on `n` THEN1
								RW_TAC real_ss [POW_0] THEN
								MATCH_MP_TAC REAL_LE_TRANS THEN
								EXISTS_TAC (--`(1 + e) pow n`--) THEN
								RW_TAC real_ss [pow, REAL_RDISTRIB, 
										REAL_LE_ADDR] THEN
								MATCH_MP_TAC REAL_LE_MUL THEN
								RW_TAC real_ss [REAL_LT_IMP_LE] THEN
								MATCH_MP_TAC REAL_LE_TRANS THEN	
								EXISTS_TAC (--`1:real`--) THEN
								RW_TAC real_ss []]]);

val SEQ_POWPOW = store_thm
  ("SEQ_POWPOW",
  (--`!c. 0 < c /\ c < 1 ==> (\n. c pow (n*n)) --> &0`--),
GEN_TAC THEN DISCH_TAC THEN
        KNOW_TAC `!n. c pow (n*n) = inv(inv(c pow (n*n)))` THEN1 (
                GEN_TAC THEN CONV_TAC SYM_CONV THEN MATCH_MP_TAC REAL_INVINV THEN
                MATCH_MP_TAC POW_NZ THEN
                RW_TAC std_ss [REAL_POS_NZ]) THEN
        ONCE_ASM_REWRITE_TAC [] THEN
        CONV_TAC(EXACT_CONV[X_BETA_CONV (--`n:num`--) (--`inv(c pow (n*n))`--)]) THEN
        MATCH_MP_TAC SEQ_INV0 THEN
        BETA_TAC THEN X_GEN_TAC (--`y:real`--) THEN
        KNOW_TAC `~(c = &0)` THEN1
                RW_TAC std_ss [REAL_POS_NZ] THEN
        RW_TAC std_ss [POW_INV] THEN
        REWRITE_TAC [real_gt] THEN
        KNOW_TAC `&0 < inv(c) - &1` THEN1 (
                REWRITE_TAC[REAL_LT_SUB_LADD] THEN REWRITE_TAC[REAL_ADD_LID] THEN
                ONCE_REWRITE_TAC[GSYM REAL_INV1] THEN MATCH_MP_TAC REAL_LT_INV THEN
                ASM_REWRITE_TAC[]) THEN
        MP_TAC(SPEC (--`inv(c) - &1`--) REAL_ARCH) THEN ASM_REWRITE_TAC[] THEN
        DISCH_THEN(X_CHOOSE_TAC (--`N:num`--) o SPEC (--`y:real`--)) THEN
        EXISTS_TAC (--`N:num`--) THEN
        X_GEN_TAC (--`n:num`--) THEN

        REWRITE_TAC[GREATER_EQ] THEN
        DISCH_TAC THEN
        KNOW_TAC `y < (&n * (inv(c) - &1))` THEN1 (
                MATCH_MP_TAC REAL_LTE_TRANS THEN
        EXISTS_TAC (--`&N * (inv(c) - &1)`--) THEN ASM_REWRITE_TAC[] THEN
        FIRST_ASSUM(fn th => GEN_REWR_TAC I [MATCH_MP REAL_LE_RMUL th]) THEN
        ASM_REWRITE_TAC[REAL_LE]) THEN

        MATCH_MP_TAC REAL_LT_TRANS THEN
        EXISTS_TAC (--`&n * (inv(c) - &1)`--) THEN ASM_REWRITE_TAC[] THEN
        MATCH_MP_TAC REAL_LTE_TRANS THEN
        EXISTS_TAC (--`&1 + (&n * (inv(c) - &1))`--) THEN
        REWRITE_TAC[REAL_LT_ADDL, REAL_LT_01] THEN
        MATCH_MP_TAC REAL_LE_TRANS THEN
        EXISTS_TAC (--`(&1 + (inv(c) - &1)) pow (n*n)`--) THEN
        CONJ_TAC THENL [
                MATCH_MP_TAC POWPOW_PLUS1 THEN
                ASM_REWRITE_TAC[], 

                RW_TAC real_ss [REAL_SUB_ASSOC2]]);

val RAYLEIGH_CDF_AT_POSINF = store_thm
    ("RAYLEIGH_CDF_AT_POSINF", 
    (--`!sig. (0 < sig) ==>
        (((\n.(\x. (if x <= 0 then 0 else
                (1 - exp (~(x pow 2)/ (2 * (sig pow 2)))))) ((\n. &n) n)) --> 1))`--),
        REPEAT STRIP_TAC THEN
        ONCE_REWRITE_TAC [SEQ_SUC] THEN
        RW_TAC arith_ss [REAL_LE] THEN
        KNOW_TAC `(\n. 1 - exp (~(& (SUC n) pow 2) / (2 * sig pow 2))) =
                (\n. (\n. 1) n - (\n. exp (~(& (SUC n) pow 2) / (2 * sig pow 2))) n)` THEN1
                RW_TAC std_ss [] THEN
        KNOW_TAC `1 = 1 - 0` THEN1
                RW_TAC real_ss [] THEN
        ONCE_ASM_REWRITE_TAC [] THEN
        MATCH_MP_TAC SEQ_SUB THEN
        RW_TAC std_ss [SEQ_CONST, real_div, POW_2, REAL_NEG_RMUL, GSYM REAL_MUL_ASSOC] THEN
        RW_TAC std_ss [EXP_N, REAL_MUL_LNEG, REAL_NEG_RMUL, EXP_NEG] THEN
        KNOW_TAC `(\n. (inv (exp (inv (2 * (sig * sig)))) pow SUC n) pow SUC n) =
                (\n. (\n. (inv (exp (inv (2 * (sig * sig)))) pow n) pow n) (SUC n)) ` THEN1
                RW_TAC std_ss [] THEN
        ASM_REWRITE_TAC [] THEN
        ONCE_REWRITE_TAC [GSYM SEQ_SUC] THEN
        RW_TAC std_ss [REAL_POW_POW] THEN
        MATCH_MP_TAC SEQ_POWPOW THEN
        RW_TAC real_ss [REAL_INV_POS,
                        REAL_INV_1OVER, EXP_POS_LT, REAL_LT_LDIV_EQ] THEN
        MATCH_MP_TAC EXP_LT_1 THEN
        RW_TAC real_ss [real_div, REAL_LT_MUL, REAL_INV_POS]);

val RAYLEIGH_CDF_IS_CONT_CDF_FN = store_thm
    ("RAYLEIGH_CDF_IS_CONT_CDF_FN",
    (--`!sig. (0 < sig) ==>
        (IS_CONT_CDF_FN (\x. (if x <= 0 then 0 
		else (1 - exp (~(x pow 2)/ (2 * (sig pow 2)))))))`--),
        REPEAT STRIP_TAC THEN
        REWRITE_TAC [IS_CONT_CDF_FN_def] THEN
        RW_TAC std_ss [RAYLEIGH_CDF_CONT, RAYLEIGH_CDF_MONO,
			RAYLEIGH_CDF_AT_POSINF, RAYLEIGH_CDF_AT_NEGINF]);

val CDF_RAYLEIGH = store_thm
    ("CDF_RAYLEIGH",
    (--`!x sig. (0 < sig) ==>
        (prob bern {s | rayleigh_rv sig s <= x} =
                        (if x <= 0 then 0 else (1 - exp (~(x pow 2)/ (2 * (sig pow 2))))))`--),
        REPEAT STRIP_TAC THEN
	REWRITE_TAC [rayleigh_rv_def] THEN
        KNOW_TAC `!s. sig * sqrt (~2 * ln (1 - std_unif_cont s)) =
                (\x. sig * sqrt (~2 * ln (1 - x))) (std_unif_cont s)` THEN1
                RW_TAC std_ss [] THEN
        KNOW_TAC `(if x <= 0 then 0 else (1 - exp (~(x pow 2)/ (2 * (sig pow 2))))) =
        (\x. (if x <= 0 then 0 else (1 - exp (~(x pow 2)/ (2 * (sig pow 2)))))) x` THEN1
                RW_TAC std_ss [] THEN
        ONCE_ASM_REWRITE_TAC [] THEN
        MATCH_MP_TAC ITM THEN
        RW_TAC std_ss [RAYLEIGH_CDF_IS_CONT_CDF_FN, INV_CDF_FN_RAYLEIGH]);

val RAYLEIGH_RV_LE_IN_EVENTS_BERN = store_thm
   ("RAYLEIGH_RV_LE_IN_EVENTS_BERN",
   (--`!x sig. (0 < sig) ==>
        ({s | rayleigh_rv sig s  <= x} IN events bern)`--),
        REPEAT STRIP_TAC THEN
	REWRITE_TAC [rayleigh_rv_def] THEN
        KNOW_TAC `!s. sig * sqrt (~2 * ln (1 - std_unif_cont s)) =
                (\x. sig * sqrt (~2 * ln (1 - x))) (std_unif_cont s)` THEN1
                RW_TAC std_ss [] THEN
        ONCE_ASM_REWRITE_TAC [] THEN
        MATCH_MP_TAC FN_STD_UNIF_CONT_EVENTS_BERN THEN
        EXISTS_TAC (--`(\x. (if x <= 0 then 0 else 
		(1 - exp (~(x pow 2)/ (2 * (sig pow 2))))))`--) THEN
        RW_TAC std_ss [RAYLEIGH_CDF_IS_CONT_CDF_FN, INV_CDF_FN_RAYLEIGH]);

val RAYLEIGH_RV_EQ_IN_EVENTS_BERN = store_thm
   ("RAYLEIGH_RV_EQ_IN_EVENTS_BERN",
   (--`!x sig. (0 < sig) ==>
        ({s | rayleigh_rv sig s  = x} IN events bern)`--),
        REPEAT STRIP_TAC THEN
	REWRITE_TAC [rayleigh_rv_def] THEN
        KNOW_TAC `!s. sig * sqrt (~2 * ln (1 - std_unif_cont s)) =
                (\x. sig * sqrt (~2 * ln (1 - x))) (std_unif_cont s)` THEN1
                RW_TAC std_ss [] THEN
        ONCE_ASM_REWRITE_TAC [] THEN
        MATCH_MP_TAC FN_STD_UNIF_CONT_EQ_EVENTS_BERN THEN
        EXISTS_TAC (--`(\x. (if x <= 0 then 0 else 
			(1 - exp (~(x pow 2)/ (2 * (sig pow 2))))))`--) THEN
        RW_TAC std_ss [RAYLEIGH_CDF_IS_CONT_CDF_FN, INV_CDF_FN_RAYLEIGH]);

val PMF_RAYLEIGH = store_thm
    ("PMF_RAYLEIGH",
    (--`!x sig. (0 < sig) ==>
        (prob bern {s | rayleigh_rv sig s  = x} = 0)`--),
	REPEAT STRIP_TAC THEN
	REWRITE_TAC [rayleigh_rv_def] THEN
	KNOW_TAC `!s. sig * sqrt (~2 * ln (1 - std_unif_cont s)) = 
		(\x. sig * sqrt (~2 * ln (1 - std_unif_cont x))) s` THEN1
		RW_TAC std_ss [] THEN	
	ONCE_ASM_REWRITE_TAC [] THEN
	MATCH_MP_TAC CONT_PROB_BERN_EQ_0 THEN
	POP_ASSUM K_TAC THEN
	RW_TAC std_ss [CDF_def, CDF_in_events_bern_def, RAYLEIGH_RV_EQ_IN_EVENTS_BERN,
		RAYLEIGH_RV_LE_IN_EVENTS_BERN, CDF_RAYLEIGH, RAYLEIGH_CDF_CONT, GSYM rayleigh_rv_def]);




(*=================================
Triangular Distribution with x0 = 0
===================================*);
val INV_CDF_FN_TRIANGULAR = store_thm
   ("INV_CDF_FN_TRIANGULAR",
   (--`!x a.(0 < a) ==>
        (INV_CDF_FN (\x. a * (1 - sqrt (1 - x))))
                 (\x. if (x <= 0) then 0 else
                        (if (x < a) then (2/a * (x - (x pow 2)/(2 * a))) else 1))`--),
        RW_TAC real_ss [INV_CDF_FN_def, SQRT_1, SQRT_0] THENL [
                Cases_on `x <= 0` THEN1
                        PROVE_TAC [] THEN
                FULL_SIMP_TAC std_ss [] THEN
                Cases_on `x < a` THENL [
                        FULL_SIMP_TAC std_ss [REAL_ENTIRE] THEN1 (
                                KNOW_TAC `2 = 0 * a` THEN1
                                        PROVE_TAC [REAL_EQ_LDIV_EQ] THEN
                                FULL_SIMP_TAC real_ss []) THEN

                                KNOW_TAC `0 < 2 * a` THEN1
                                        FULL_SIMP_TAC real_ss [REAL_LT_MUL] THEN
                                FULL_SIMP_TAC real_ss [REAL_SUB_0, REAL_POS_NZ,
                                REAL_EQ_RDIV_EQ, POW_2, REAL_EQ_LMUL, REAL_NOT_LE] THEN
                                KNOW_TAC `x = 2 * a` THEN1
                                        PROVE_TAC [] THEN
                                FULL_SIMP_TAC real_ss [] THEN
                                KNOW_TAC `2 < a/a` THEN1
                                        PROVE_TAC [GSYM REAL_LT_RDIV_EQ] THEN
                                POP_ASSUM MP_TAC THEN
                                RW_TAC real_ss [REAL_DIV_REFL, REAL_POS_NZ, REAL_NOT_LT],

                                FULL_SIMP_TAC real_ss []], 

                Cases_on `x <= 0` THEN1
                        FULL_SIMP_TAC real_ss [] THEN
                FULL_SIMP_TAC std_ss [] THEN
                Cases_on `x < a` THENL [
                        FULL_SIMP_TAC std_ss [] THEN
                        KNOW_TAC `(x - a) * (x - a) = 0` THEN1 (
                                RW_TAC real_ss [REAL_SUB_LDISTRIB, REAL_SUB_RDISTRIB,
					REAL_EQ_SUB_RADD] THEN
				ONCE_REWRITE_TAC [REAL_ADD_SYM] THEN
                                KNOW_TAC ` x * a + (a * x - a * a) = (2 * a) * x - a * a` THEN1
                                        REAL_ARITH_TAC THEN
                                ONCE_ASM_REWRITE_TAC [] THEN
                                REWRITE_TAC [REAL_EQ_SUB_LADD] THEN
                                ONCE_REWRITE_TAC [REAL_ADD_SYM] THEN
                                REWRITE_TAC [GSYM REAL_EQ_SUB_LADD] THEN
                                KNOW_TAC `a * a = 1 * (a * a)` THEN1
                                        RW_TAC real_ss [] THEN
                                KNOW_TAC `0 < a * a` THEN1
                                        RW_TAC std_ss [REAL_LT_MUL] THEN
                                ONCE_ASM_REWRITE_TAC [] THEN
                                RW_TAC std_ss [GSYM REAL_EQ_RDIV_EQ] THEN
                                KNOW_TAC `(2 * a * x - x * x) = 2 * a * (x - x pow 2 / (2 * a))`
THEN1 (
                                        RW_TAC std_ss [REAL_SUB_LDISTRIB, POW_2] THEN
                                        SUFF_TAC `x * x = 2 * a * (x * x / (2 * a))` THEN1
                                                PROVE_TAC [] THEN
                                        KNOW_TAC `0 < 2 * a` THEN1
                                                RW_TAC real_ss [REAL_LT_MUL] THEN
                                        PROVE_TAC [REAL_DIV_LMUL, REAL_POS_NZ]) THEN
                                ONCE_ASM_REWRITE_TAC [] THEN
                                FULL_SIMP_TAC real_ss [real_div] THEN
                                ONCE_REWRITE_TAC [REAL_MUL_SYM] THEN
                                RW_TAC std_ss [REAL_POS_NZ, REAL_INV_MUL, REAL_MUL_ASSOC] THEN
                                KNOW_TAC `inv a * inv a * 2 * a * (x - x pow 2 * inv (2 * a)) =
                                (inv a * a) * (2 * inv a * (x - x pow 2 * inv (2 * a)))` THEN1
                                        REAL_ARITH_TAC THEN
                                FULL_SIMP_TAC real_ss [REAL_POS_NZ, REAL_MUL_LINV]) THEN
                        FULL_SIMP_TAC std_ss [REAL_ENTIRE, REAL_SUB_0, REAL_EQ_IMP_LE], 

                        FULL_SIMP_TAC std_ss [REAL_NOT_LT]], 

                Cases_on `x <= 0` THEN1
                        FULL_SIMP_TAC real_ss [] THEN
                FULL_SIMP_TAC real_ss [] THEN
                Cases_on `x < a` THENL [
                        FULL_SIMP_TAC std_ss [] THEN
                KNOW_TAC `(1 - 2 / a * (x - x pow 2 / (2 * a))) =
                        (1 - x * inv a) * (1 - x * inv a)` THEN1 (
                        RW_TAC real_ss [REAL_SUB_LDISTRIB, REAL_SUB_RDISTRIB, real_div,
                                REAL_INV_MUL, REAL_POS_NZ,  REAL_SUB_ASSOC, POW_2] THEN
                        KNOW_TAC `(2 * inv a * (x * x * (inv 2 * inv a))) =
                                (2 * inv 2) * x * inv a * (x * inv a)` THEN1
                                REAL_ARITH_TAC THEN
                        ONCE_ASM_REWRITE_TAC [] THEN
                        RW_TAC real_ss [REAL_MUL_RINV, REAL_POS_NZ] THEN
                        REAL_ARITH_TAC) THEN
                ONCE_ASM_REWRITE_TAC [] THEN
                KNOW_TAC `0 <= (1 - x * inv a)` THEN1
                        RW_TAC real_ss [REAL_SUB_LE, GSYM real_div, REAL_LE_LDIV_EQ,
				 REAL_LT_IMP_LE] THEN
                RW_TAC real_ss [GSYM POW_2,     POW_2_SQRT, REAL_MUL_ASSOC] THEN
                ONCE_REWRITE_TAC [REAL_MUL_SYM] THEN
                RW_TAC real_ss [REAL_MUL_LINV, REAL_POS_NZ, REAL_MUL_ASSOC], 

                FULL_SIMP_TAC real_ss []], 

        KNOW_TAC `a * (1 - sqrt (1 - x')) <= 0 = F` THEN1 (
                RW_TAC real_ss [REAL_NOT_LE] THEN
                MATCH_MP_TAC REAL_LT_MUL THEN
                RW_TAC real_ss [GSYM REAL_LT_ADD_SUB] THEN
                CCONTR_TAC THEN
                FULL_SIMP_TAC std_ss [REAL_NOT_LT] THEN
                KNOW_TAC `1 * 1 <= sqrt (1 - x') * sqrt (1 - x')` THEN1 (
                        MATCH_MP_TAC REAL_LE_MUL2 THEN
                        RW_TAC real_ss []) THEN
                KNOW_TAC `0 < (1 - x')` THEN1
                        RW_TAC real_ss [REAL_SUB_LT] THEN
                FULL_SIMP_TAC real_ss [GSYM POW_2, SQRT_POW_2, REAL_LT_IMP_LE, REAL_LE_SUB_LADD]
THEN
                POP_ASSUM K_TAC THEN
                POP_ASSUM MP_TAC THEN
                RW_TAC real_ss [REAL_LT_ADDR, REAL_NOT_LE] THEN
                KNOW_TAC `0 < (1 - x')` THEN1
                        RW_TAC real_ss [REAL_SUB_LT] THEN
                RW_TAC real_ss [SQRT_POW_2, REAL_LT_IMP_LE, REAL_LT_SUB_RADD, REAL_LT_ADDR]) THEN

        FULL_SIMP_TAC real_ss [] THEN
        KNOW_TAC `a * (1 - sqrt (1 - x')) < a` THEN1 (
                ONCE_REWRITE_TAC [REAL_MUL_SYM] THEN
                RW_TAC real_ss [GSYM REAL_LT_RDIV_EQ, REAL_DIV_REFL, REAL_POS_NZ,
                                REAL_LT_SUB_RADD, REAL_LT_ADDR] THEN
                MATCH_MP_TAC SQRT_POS_LT THEN
                RW_TAC real_ss [REAL_LT_SUB_LADD]) THEN
        FULL_SIMP_TAC real_ss [real_div, POW_2, REAL_INV_MUL, REAL_POS_NZ,
                        REAL_SUB_LDISTRIB, REAL_MUL_ASSOC, REAL_SUB_RDISTRIB] THEN
        KNOW_TAC `
2 * inv a * a * sqrt (1 - x') * a * sqrt (1 - x') * inv 2 * inv a =
(2 * inv 2) * (a * inv a) * (a * inv a) * sqrt (1 - x') * sqrt (1 - x')` THEN1
                REAL_ARITH_TAC THEN
        KNOW_TAC `2 * inv a * a * a * sqrt (1 - x') * inv 2 * inv a =
(2 * inv 2) * (a * inv a) * (a * inv a) * sqrt (1 - x')` THEN1
                REAL_ARITH_TAC THEN
        KNOW_TAC `2 * inv a * a * sqrt (1 - x') * a * inv 2 * inv a =
(2 * inv 2) * (a * inv a) * (a * inv a) * sqrt (1 - x')` THEN1
                REAL_ARITH_TAC THEN
        KNOW_TAC `2 * inv a * a * a * inv 2 * inv a =
(2 * inv 2) * (a * inv a) * (a * inv a)` THEN1
                REAL_ARITH_TAC THEN
        KNOW_TAC `2 * inv a * a * sqrt (1 - x') =
2 * (a * inv a) * sqrt (1 - x')` THEN1
                REAL_ARITH_TAC THEN
        KNOW_TAC `2 * inv a * a = 2 * (a * inv a)` THEN1
                REAL_ARITH_TAC THEN
        ONCE_ASM_REWRITE_TAC [] THEN
        KNOW_TAC `0 < (1 - x')` THEN1
                        RW_TAC real_ss [REAL_SUB_LT] THEN
        RW_TAC real_ss [REAL_MUL_RINV, REAL_POS_NZ, GSYM POW_2, SQRT_POW_2, REAL_LT_IMP_LE] THEN
        REAL_ARITH_TAC]);

val TRIANGULAR_CDF_MONO = store_thm
   ("TRIANGULAR_CDF_MONO",
   (--`!a . (0 < a) ==> (!c d. (c < d) ==>
        ((\x. if (x <= 0) then 0 else
                        (if (x < a) then (2/a * (x - (x pow 2)/(2 * a))) else 1)) c <=
        (\x. if (x <= 0) then 0 else
                        (if (x < a) then (2/a * (x - (x pow 2)/(2 * a))) else 1)) d ))`--),
        RW_TAC real_ss [] THEN
        Cases_on `c <= 0` THENL [
                RW_TAC real_ss [] THEN
                MATCH_MP_TAC REAL_LE_MUL THEN
                RW_TAC real_ss [REAL_LE_RDIV_EQ, POW_2, REAL_SUB_LE,
			REAL_LE_LDIV_EQ, REAL_LT_MUL] THEN
                FULL_SIMP_TAC std_ss [REAL_LE_LMUL, REAL_NOT_LE] THEN
                MATCH_MP_TAC REAL_LE_TRANS THEN
                EXISTS_TAC (--`a:real`--) THEN
                RW_TAC real_ss [REAL_LT_IMP_LE, GSYM REAL_LE_LDIV_EQ,
			REAL_POS_NZ, REAL_DIV_REFL], 
		Cases_on `d <= 0` THENL [
                        CCONTR_TAC THEN
                        POP_ASSUM K_TAC THEN
                        KNOW_TAC `d < c` THEN1(
                                MATCH_MP_TAC REAL_LET_TRANS THEN
                                EXISTS_TAC (--`0:real`--) THEN
                                PROVE_TAC [REAL_NOT_LE]) THEN
                        POP_ASSUM MP_TAC THEN
                        RW_TAC std_ss [REAL_NOT_LT, REAL_LT_IMP_LE], 
                RW_TAC real_ss [] THENL [
                        MATCH_MP_TAC REAL_LE_LMUL_IMP THEN
                        RW_TAC real_ss [REAL_LE_RDIV_EQ, real_div, 
				REAL_INV_MUL, REAL_POS_NZ] THEN
			KNOW_TAC `!x y. (0 < y) ==> (x - x pow 2 * (inv 2 * inv y) =
                        (inv 2) * (inv y) * (2 * y * x - x pow 2))` THEN1 (
                                RW_TAC real_ss [REAL_SUB_LDISTRIB] THEN
                                KNOW_TAC `inv 2 * inv y * (2 * y * x) =
                                (inv 2 * 2) * (inv y * y) * x` THEN1
                                        REAL_ARITH_TAC THEN
                                ASM_REWRITE_TAC [] THEN
                                RW_TAC real_ss [REAL_MUL_LINV, REAL_POS_NZ] THEN
                                REAL_ARITH_TAC) THEN
                        FULL_SIMP_TAC std_ss [REAL_NOT_LE] THEN
                        MATCH_MP_TAC REAL_LE_LMUL_IMP THEN
                        RW_TAC real_ss [REAL_LE_MUL, REAL_INV_POS, REAL_LT_IMP_LE] THEN
                        KNOW_TAC ` 2 * a * c - c pow 2 <= 2 * a * d - d pow 2 =
                        d pow 2 - 2 * a * d + a pow 2 <= c pow 2 - 2 * a * c + a pow 2` THEN1
                                REAL_ARITH_TAC THEN
                        ONCE_ASM_REWRITE_TAC [] THEN
                        KNOW_TAC `!x y. x pow 2 - 2 * y * x + y pow 2 = (y - x) * (y - x)` THEN1 (
                                RW_TAC std_ss [POW_2, REAL_SUB_LDISTRIB, REAL_SUB_RDISTRIB] THEN
                                REAL_ARITH_TAC) THEN
                        ASM_REWRITE_TAC [] THEN
                        MATCH_MP_TAC REAL_LE_MUL2 THEN
                        REWRITE_TAC [REAL_SUB_LE,  REAL_LE_SUB_LADD, REAL_SUB_ASSOC2] THEN
                        ONCE_REWRITE_TAC [REAL_ADD_SYM] THEN
                        RW_TAC real_ss [REAL_LT_IMP_LE, GSYM REAL_LE_SUB_LADD, REAL_LE_SUB_RADD], 

                        RW_TAC real_ss [real_div, REAL_INV_MUL, REAL_POS_NZ] THEN
                        KNOW_TAC `2 * inv a * (c - c pow 2 * (inv 2 * inv a)) =
                                (2 * a * c - c pow 2) * (inv (a pow 2))` THEN1 (
                                RW_TAC real_ss [REAL_SUB_LDISTRIB, REAL_SUB_RDISTRIB,
                                                REAL_INV_MUL, REAL_POS_NZ, POW_2] THEN
                                KNOW_TAC `(2 * a * c * (inv a * inv a) =
                                        2 * inv a * c * (a * inv a)) /\
                                (2 * inv a * (c * c * (inv 2 * inv a)) =
                                        c * c * (inv a * inv a) * (2 * inv 2))` THEN1
                                        REAL_ARITH_TAC THEN
                                ASM_REWRITE_TAC [] THEN
                                RW_TAC real_ss [REAL_MUL_RINV, REAL_POS_NZ]) THEN
                        ASM_REWRITE_TAC [] THEN
                        RW_TAC real_ss [GSYM REAL_LE_RDIV_EQ, REAL_INV_POS, REAL_POW_LT,
                                real_div, REAL_INVINV, REAL_POS_NZ, REAL_LE_SUB_RADD] THEN
                        ONCE_REWRITE_TAC [GSYM REAL_SUB_LE] THEN
                        KNOW_TAC `!x y. x pow 2 + y pow 2 - 2 * x * y = (x - y) * (x - y)` THEN1 (
                                RW_TAC std_ss [POW_2, REAL_SUB_LDISTRIB, REAL_SUB_RDISTRIB] THEN
                                REAL_ARITH_TAC) THEN
                        ASM_REWRITE_TAC [] THEN
                        PROVE_TAC [REAL_LE_SQUARE], 

                        KNOW_TAC `c < a` THEN 
                                MATCH_MP_TAC REAL_LT_TRANS THEN
                                EXISTS_TAC (--`d:real`--) THEN
                                PROVE_TAC []]]] );

val TRIANGULAR_CDF_AT_POSINF = store_thm
   ("TRIANGULAR_CDF_AT_POSINF",
  (--`!a. (0 < a) ==>
        (((\n.(\x. if (x <= 0) then 0 else
                        (if (x < a) then (2/a * (x - (x pow 2)/(2 * a))) else 1))
				((\n. &n) n)) --> 1))`--),
        RW_TAC std_ss [SEQ] THEN
        EXISTS_TAC (--`ceiling a`--) THEN
        RW_TAC real_ss [GREATER_EQ, ABS_0] THEN1
                PROVE_TAC [LB_CEILING, GSYM REAL_NOT_LT] THEN
                KNOW_TAC `& (ceiling a) < a` THEN1 (
                        MATCH_MP_TAC REAL_LET_TRANS THEN
                        EXISTS_TAC (--`&n:real`--) THEN
                        RW_TAC std_ss [REAL_LE, GSYM GREATER_EQ]) THEN
                PROVE_TAC [LB_CEILING, GSYM REAL_NOT_LT]);


val TRIANGULAR_CDF_AT_NEGINF = store_thm
  ("TRIANGULAR_CDF_AT_NEGINF",
  (--`!a. (0 < a) ==>
        (((\n.(\x. if (x <= 0) then 0 else
                        (if (x < a) then (2/a * (x - (x pow 2)/(2 * a))) else 1))
					 ((\n. ~&n) n))  --> 0))`--),
        RW_TAC std_ss [SEQ] THEN
        EXISTS_TAC (--`1:num`--) THEN
        RW_TAC real_ss [GREATER_EQ, ABS_0]);

val TRIANGULAR_CONT_HELPER = store_thm
  ("TRIANGULAR_CONT_HELPER",
  (--`!x a. (0 < a) ==>
        (\x. (2/a * (x - (x pow 2)/(2 * a)))) contl x`--),
	RW_TAC  std_ss [real_div] THEN
        KNOW_TAC `(\x. 2 * inv a * (x - x pow 2 * inv (2 * a))) =
		(\x.(\x. 2* inv a) x * (\x. (\x. x) x - 
			(\x. (\x. x pow 2) x * (\x. inv (2 * a)) x) x) x)` THEN1
		RW_TAC std_ss [] THEN
        ONCE_ASM_REWRITE_TAC [] THEN
        MATCH_MP_TAC CONT_MUL THEN
        CONJ_TAC THEN1
                RW_TAC std_ss [CONT_CONST] THEN

                MATCH_MP_TAC CONT_SUB THEN
                CONJ_TAC THEN1 (
                        MATCH_MP_TAC DIFF_CONT THEN
                        EXISTS_TAC (--`1:real`--) THEN
                        RW_TAC std_ss [DIFF_X]) THEN

                        MATCH_MP_TAC CONT_MUL THEN
                        CONJ_TAC THEN1 (
                                MATCH_MP_TAC DIFF_CONT THEN
                                EXISTS_TAC (--`&2 * x pow (2 - 1)`--) THEN
                                MP_TAC (Q.SPECL [`2`, `x`] DIFF_POW) THEN
                                RW_TAC std_ss []) THEN

                                RW_TAC std_ss [CONT_CONST]);

val TRIANGULAR_CDF_CONT = store_thm
   ("TRIANGULAR_CDF_CONT",
   (--`!x a. (0 < a) ==>
        (\x. if (x <= 0) then 0 else
                        (if (x < a) then (2/a * (x - (x pow 2)/(2 * a))) else 1)) contl x`--),
        REPEAT STRIP_TAC THEN
        MP_TAC TRIANGULAR_CONT_HELPER THEN
        DISCH_TAC THEN

        RW_TAC real_ss [contl, LIM] THENL [
                NTAC 2 (POP_ASSUM MP_TAC) THEN
                POP_ASSUM (MP_TAC o Q.SPECL [`0`, `a`]) THEN
                RW_TAC real_ss [contl, LIM] THEN
                NTAC 2 (POP_ASSUM MP_TAC) THEN
                POP_ASSUM (MP_TAC o Q.SPEC `e`) THEN
                RW_TAC real_ss [REAL_NEG_0] THEN
                FULL_SIMP_TAC std_ss [ABS_DIV, REAL_POS_NZ, REAL_LT_LDIV_EQ] THEN
                EXISTS_TAC (--`d:real`--) THEN
                RW_TAC real_ss [ABS_0, ABS_DIV, REAL_POS_NZ, REAL_LT_LDIV_EQ] THENL [
                        KNOW_TAC `0 < h` THEN1 (
                                KNOW_TAC `0 < h = x < x + h` THEN1
                                        RW_TAC std_ss [REAL_LT_ADDR] THEN
                                ASM_REWRITE_TAC [] THEN
                                MATCH_MP_TAC REAL_LET_TRANS THEN
                                EXISTS_TAC (--`0:real`--) THEN
                                PROVE_TAC [REAL_NOT_LE]) THEN
                        FULL_SIMP_TAC std_ss [REAL_NOT_LE] THEN
                        KNOW_TAC `(abs h = h) /\ (0 < abs (x + h))` THEN1
                                PROVE_TAC [ABS_REFL, REAL_LT_IMP_LE] THEN
                        KNOW_TAC `abs (x + h) = x + h` THEN1
                                PROVE_TAC [ABS_REFL, REAL_LT_IMP_LE] THEN
                        FULL_SIMP_TAC std_ss [] THEN
                        KNOW_TAC `x + h < d` THEN1 (
                                MATCH_MP_TAC REAL_LET_TRANS THEN
                                EXISTS_TAC (--`h:real`--) THEN
                                RW_TAC real_ss [GSYM REAL_LE_SUB_LADD]) THEN
                        PROVE_TAC [], 

                        FULL_SIMP_TAC std_ss [REAL_NOT_LT, REAL_NOT_LE] THEN
                        KNOW_TAC `0 < h` THEN1 (
                                KNOW_TAC `0 < h = x < x + h` THEN1
                                        RW_TAC std_ss [REAL_LT_ADDR] THEN
                                ASM_REWRITE_TAC [] THEN
                                MATCH_MP_TAC REAL_LET_TRANS THEN
                                EXISTS_TAC (--`0:real`--) THEN
                                PROVE_TAC [REAL_NOT_LE]) THEN
                        KNOW_TAC `(abs a = a) /\ (abs h = h)` THEN1
                                PROVE_TAC [ABS_REFL, REAL_LT_IMP_LE] THEN
                        KNOW_TAC `abs a < d` THEN1 (
                                ASM_REWRITE_TAC [] THEN
                                MATCH_MP_TAC REAL_LET_TRANS THEN
                                EXISTS_TAC (--`x + h:real`--) THEN
                                ASM_REWRITE_TAC [] THEN
                                MATCH_MP_TAC REAL_LET_TRANS THEN
                                EXISTS_TAC (--`h:real`--) THEN
                                FULL_SIMP_TAC real_ss [GSYM REAL_LE_SUB_LADD]) THEN

                        KNOW_TAC `abs (2 / a * (a - a pow 2/ (2 * a))) < e` THEN1
                                PROVE_TAC [] THEN
                        SUFF_TAC `(2 / a * (a - a pow 2/ (2 * a))) = 1` THEN1
                                PROVE_TAC [ABS_1] THEN
                        RW_TAC real_ss [real_div, POW_2, REAL_SUB_LDISTRIB,
                        REAL_INV_MUL, REAL_POS_NZ, REAL_MUL_ASSOC] THEN
                        KNOW_TAC `((2 * inv a * a * a * inv 2 * inv a) =
                        ((inv 2 * 2) * (inv a * a) * (inv a * a))) /\
                        ((2 * inv a * a) = (2 * (inv a * a)))` THEN1
                                REAL_ARITH_TAC THEN
                        ASM_REWRITE_TAC [] THEN
                        RW_TAC real_ss [REAL_MUL_LINV, REAL_POS_NZ]],

                NTAC 3 (POP_ASSUM MP_TAC) THEN
                POP_ASSUM (MP_TAC o Q.SPECL [`x`, `a`]) THEN
                RW_TAC real_ss [contl, LIM] THEN
                NTAC 3 (POP_ASSUM MP_TAC) THEN
                POP_ASSUM (MP_TAC o Q.SPEC `e`) THEN
                RW_TAC real_ss [REAL_NEG_0] THEN
                FULL_SIMP_TAC std_ss [ABS_DIV, REAL_POS_NZ, REAL_LT_LDIV_EQ] THEN
                EXISTS_TAC (--`d:real`--) THEN
                RW_TAC real_ss [ABS_0, ABS_DIV, REAL_POS_NZ, REAL_LT_LDIV_EQ] THENL [
                        KNOW_TAC `0 < abs (~x)` THEN1 (
                                FULL_SIMP_TAC real_ss [GSYM ABS_NZ,
                                REAL_NE_LT_GT, REAL_NEG_LT0, REAL_NOT_LE]) THEN
                        KNOW_TAC `abs (~x) < d` THEN1 (
                                MATCH_MP_TAC REAL_LET_TRANS THEN
                                EXISTS_TAC (--`abs h`--) THEN
                                ASM_REWRITE_TAC [] THEN
                                KNOW_TAC `abs ~x = x` THEN1
                                        FULL_SIMP_TAC real_ss [abs, REAL_NEG_GE0] THEN
                                ASM_REWRITE_TAC [] THEN
                                MATCH_MP_TAC REAL_LE_TRANS THEN
                                EXISTS_TAC (--`~h:real`--) THEN
                                RW_TAC std_ss [REAL_LE_RNEG] THEN
                                RW_TAC std_ss [abs, REAL_LE_REFL, REAL_LE_NEGL]) THEN
                        KNOW_TAC `abs
                       (2 / a * (x + (~x) - (x + (~x)) pow 2 / (2 * a)) -
                        2 / a * (x - x pow 2 / (2 * a))) < e` THEN1 
                                PROVE_TAC [] THEN
                        SUFF_TAC `2 / a * (x + ~x - (x + ~x) pow 2 / (2 * a)) = 0` THEN1
                                PROVE_TAC [] THEN
                        RW_TAC real_ss [REAL_ADD_RINV], 


                        KNOW_TAC `0 < (a - x)` THEN1
                                RW_TAC std_ss [REAL_SUB_LT] THEN
                        KNOW_TAC `abs (a - x) = a - x` THEN1
                                PROVE_TAC [ABS_REFL, REAL_LT_IMP_LE] THEN

                        KNOW_TAC `abs (a - x) < d` THEN1 (
                                MATCH_MP_TAC REAL_LET_TRANS THEN
                                EXISTS_TAC (--`abs h`--) THEN
                                ASM_REWRITE_TAC [] THEN
                                MATCH_MP_TAC REAL_LE_TRANS THEN
                                EXISTS_TAC (--`h:real`--) THEN
                                PROVE_TAC [ABS_LE, REAL_LE_SUB_RADD,
                                        REAL_NOT_LT, REAL_ADD_SYM]) THEN
                        KNOW_TAC `abs
                        (2 / a * (x + (a - x) - (x + (a - x)) pow 2 / (2 * a)) -
                        2 / a * (x - x pow 2 / (2 * a))) < e` THEN1
                                PROVE_TAC [] THEN
                        SUFF_TAC `(2 / a *
                        (x + (a - x) - (x + (a - x)) pow 2 / (2 * a))) = 1` THEN1
                                PROVE_TAC [] THEN
                        RW_TAC real_ss [GSYM REAL_SUB_ASSOC2] THEN
                        RW_TAC real_ss [real_div, POW_2, REAL_SUB_LDISTRIB,
                        REAL_INV_MUL, REAL_POS_NZ, REAL_MUL_ASSOC] THEN
                        KNOW_TAC `((2 * inv a * a * a * inv 2 * inv a) =
                        ((inv 2 * 2) * (inv a * a) * (inv a * a))) /\
                        ((2 * inv a * a) = (2 * (inv a * a)))` THEN1
                                REAL_ARITH_TAC THEN
                        ASM_REWRITE_TAC [] THEN
                        RW_TAC real_ss [REAL_MUL_LINV, REAL_POS_NZ]], 

                NTAC 3 (POP_ASSUM MP_TAC) THEN
                POP_ASSUM (MP_TAC o Q.SPECL [`a`, `a`]) THEN
                RW_TAC real_ss [contl, LIM] THEN
                NTAC 3 (POP_ASSUM MP_TAC) THEN
                POP_ASSUM (MP_TAC o Q.SPEC `e`) THEN
                RW_TAC real_ss [REAL_NEG_0] THEN
                FULL_SIMP_TAC std_ss [ABS_DIV, REAL_POS_NZ, REAL_LT_LDIV_EQ] THEN
                EXISTS_TAC (--`d:real`--) THEN
                RW_TAC real_ss [ABS_0, ABS_DIV, REAL_POS_NZ, REAL_LT_LDIV_EQ] THENL [
                        KNOW_TAC `0 < abs (~a)` THEN1 (
                                FULL_SIMP_TAC real_ss [GSYM ABS_NZ,
                                REAL_NE_LT_GT, REAL_NEG_LT0, REAL_NOT_LE]) THEN
                        KNOW_TAC `abs (~a) < d` THEN1 (
                                MATCH_MP_TAC REAL_LET_TRANS THEN
                                EXISTS_TAC (--`abs h`--) THEN
                                ASM_REWRITE_TAC [] THEN
                                KNOW_TAC `abs ~a = a` THEN1 (
                                        RW_TAC real_ss [abs,
                                        REAL_NEG_GE0, GSYM REAL_NOT_LT]) THEN
                                ASM_REWRITE_TAC [] THEN
                                MATCH_MP_TAC REAL_LE_TRANS THEN
                                EXISTS_TAC (--`x:real`--) THEN
                                FULL_SIMP_TAC std_ss [REAL_NOT_LT] THEN
                                MATCH_MP_TAC REAL_LE_TRANS THEN
                                EXISTS_TAC (--`~h:real`--) THEN
                                RW_TAC real_ss [abs, REAL_LE_REFL,
                                        REAL_LE_RNEG,REAL_LE_NEGL]) THEN
                        KNOW_TAC `abs
                       (2 / a * (a + (~a) - (a + (~a)) pow 2 / (2 * a)) -
                        2 / a * (a - a pow 2 / (2 * a))) < e` THEN1
                                PROVE_TAC [] THEN
                        KNOW_TAC `2 / a * (a + ~a - (a + ~a) pow 2 / (2 * a)) = 0` THEN1
                                RW_TAC real_ss [REAL_ADD_RINV] THEN
                        KNOW_TAC `2 / a * (a - a pow 2 / (2 * a)) = 1` THEN1 (
                                RW_TAC real_ss [real_div, POW_2, REAL_SUB_LDISTRIB,
                                REAL_INV_MUL, REAL_POS_NZ, REAL_MUL_ASSOC] THEN
                                KNOW_TAC `((2 * inv a * a * a * inv 2 * inv a) =
                                ((inv 2 * 2) * (inv a * a) * (inv a * a))) /\
                                        ((2 * inv a * a) = (2 * (inv a * a)))` THEN1
                                        REAL_ARITH_TAC THEN
                                ASM_REWRITE_TAC [] THEN
                                RW_TAC real_ss [REAL_MUL_LINV, REAL_POS_NZ]) THEN
                        PROVE_TAC [], 


                        KNOW_TAC `0 < abs (x + h - a)` THEN1 (
                                FULL_SIMP_TAC real_ss [GSYM ABS_NZ, REAL_NE_LT_GT,
                                REAL_LT_SUB_RADD]) THEN
                        KNOW_TAC `abs (x + h -a) < d` THEN1 (
                                MATCH_MP_TAC REAL_LET_TRANS THEN
                                EXISTS_TAC (--`abs h`--) THEN
                                ASM_REWRITE_TAC [] THEN
                                KNOW_TAC `h < 0` THEN1 (
                                        KNOW_TAC `h < 0 = (x + h < x)` THEN1 (
                                                ONCE_REWRITE_TAC [REAL_ADD_SYM] THEN
                                                RW_TAC real_ss [REAL_LT_ADD_SUB]) THEN
                                        ASM_REWRITE_TAC [] THEN
                                        MATCH_MP_TAC REAL_LTE_TRANS THEN
                                        EXISTS_TAC (--`a:real`--) THEN
                                        FULL_SIMP_TAC std_ss [REAL_NOT_LT]) THEN
                                FULL_SIMP_TAC real_ss [abs, REAL_LT_IMP_LE,
                                REAL_NOT_LE, GSYM REAL_NOT_LT, REAL_SUB_LE] THEN
                                RW_TAC real_ss [REAL_LT_SUB_LADD] THEN
                                ONCE_REWRITE_TAC [REAL_ADD_SYM] THEN
                                ONCE_REWRITE_TAC [GSYM REAL_ADD_ASSOC] THEN
                                RW_TAC real_ss [REAL_ADD_RINV]) THEN
                        KNOW_TAC `abs
                       (2 / a * (a + (x + h - a) -
                                (a + (x + h - a)) pow 2 / (2 * a)) -
                        2 / a * (a - a pow 2 / (2 * a))) < e` THEN1
                                PROVE_TAC [] THEN
                        KNOW_TAC `(a + (x + h - a) - (a + (x + h - a)) pow 2 / (2 * a)) =
                        (x + h - (x + h) pow 2 / (2 * a))` THEN1
                                RW_TAC real_ss [GSYM REAL_SUB_ASSOC2] THEN

                        KNOW_TAC `2 / a * (a - a pow 2 / (2 * a)) = 1` THEN1 (
                                RW_TAC real_ss [GSYM REAL_SUB_ASSOC2] THEN
                                RW_TAC real_ss [real_div, POW_2, REAL_SUB_LDISTRIB,
                                REAL_INV_MUL, REAL_POS_NZ, REAL_MUL_ASSOC] THEN
                                KNOW_TAC `((2 * inv a * a * a * inv 2 * inv a) =
                                ((inv 2 * 2) * (inv a * a) * (inv a * a))) /\
                                ((2 * inv a * a) = (2 * (inv a * a)))` THEN1
                                REAL_ARITH_TAC THEN
                                ASM_REWRITE_TAC [] THEN
                                RW_TAC real_ss [REAL_MUL_LINV, REAL_POS_NZ]) THEN
                        PROVE_TAC []]]);

val TRIANGULAR_CDF_IS_CONT_CDF_FN = store_thm
   ("TRIANGULAR_CDF_IS_CONT_CDF_FN",
   (--`! a. (0 < a) ==>
        (IS_CONT_CDF_FN (\x. if (x <= 0) then 0 else
                        (if (x < a) then (2/a * (x - (x pow 2)/(2 * a))) else 1)))`--),
        REPEAT STRIP_TAC THEN
        REWRITE_TAC [IS_CONT_CDF_FN_def] THEN
        RW_TAC std_ss [IS_CONT_CDF_FN_def, TRIANGULAR_CDF_CONT,  
			TRIANGULAR_CDF_MONO, TRIANGULAR_CDF_AT_NEGINF, TRIANGULAR_CDF_AT_POSINF]);

val CDF_TRIANGULAR = store_thm
    ("CDF_TRIANGULAR",
    (--`!x a. (0 < a) ==>
        (prob bern {s | triangular_rv a s <= x} =
        (if (x <= 0) then 0 else
                        (if (x < a) then (2/a * (x - (x pow 2)/(2 * a))) else 1)))`--),
        REPEAT STRIP_TAC THEN
	REWRITE_TAC [triangular_rv_def] THEN
        KNOW_TAC `!s. a * (1 - sqrt (1 - std_unif_cont s)) =
                        (\x. a * (1 - sqrt (1 - x))) (std_unif_cont s)` THEN1
                RW_TAC std_ss [] THEN
        KNOW_TAC `(if (x <= 0) then 0 else
                        (if (x < a) then (2/a * (x - (x pow 2)/(2 * a))) else 1)) =
        (\x. (if (x <= 0) then 0 else
                        (if (x < a) then (2/a * (x - (x pow 2)/(2 * a))) else 1))) x` THEN1
                RW_TAC std_ss [] THEN
        ONCE_ASM_REWRITE_TAC [] THEN
        MATCH_MP_TAC ITM THEN
        RW_TAC std_ss [TRIANGULAR_CDF_IS_CONT_CDF_FN, INV_CDF_FN_TRIANGULAR]);

val TRIANGULAR_RV_LE_IN_EVENTS_BERN = store_thm
    ("TRIANGULAR_RV_LE_IN_EVENTS_BERN",
    (--`!x a. (0 < a) ==>
        ({s | triangular_rv a s  <= x} IN events bern)`--),
        REPEAT STRIP_TAC THEN
	REWRITE_TAC [triangular_rv_def] THEN
        KNOW_TAC `!s. a * (1 - sqrt (1 - std_unif_cont s)) =
                        (\x. a * (1 - sqrt (1 - x))) (std_unif_cont s)` THEN1
                RW_TAC std_ss [] THEN
        ONCE_ASM_REWRITE_TAC [] THEN
        MATCH_MP_TAC FN_STD_UNIF_CONT_EVENTS_BERN THEN
        EXISTS_TAC (--`(\x. (if (x <= 0) then 0 else
                        (if (x < a) then (2/a * (x - (x pow 2)/(2 * a))) else 1)))`--) THEN
        RW_TAC std_ss [TRIANGULAR_CDF_IS_CONT_CDF_FN, INV_CDF_FN_TRIANGULAR]);

val TRIANGULAR_RV_EQ_IN_EVENTS_BERN = store_thm
    ("TRIANGULAR_RV_EQ_IN_EVENTS_BERN",
    (--`!x a. (0 < a) ==>
        ({s | triangular_rv a s = x} IN events bern)`--),
        REPEAT STRIP_TAC THEN
	REWRITE_TAC [triangular_rv_def] THEN
        KNOW_TAC `!s. a * (1 - sqrt (1 - std_unif_cont s)) =
                        (\x. a * (1 - sqrt (1 - x))) (std_unif_cont s)` THEN1
                RW_TAC std_ss [] THEN
        ONCE_ASM_REWRITE_TAC [] THEN
        MATCH_MP_TAC FN_STD_UNIF_CONT_EQ_EVENTS_BERN THEN
        EXISTS_TAC (--`(\x. (if (x <= 0) then 0 else
                        (if (x < a) then (2/a * (x - (x pow 2)/(2 * a))) else 1)))`--) THEN
        RW_TAC std_ss [TRIANGULAR_CDF_IS_CONT_CDF_FN, INV_CDF_FN_TRIANGULAR]);

val PMF_TRIANGULAR = store_thm
   ("PMF_TRIANGULAR",
   (--`!x a. (0 < a) ==>
        (prob bern {s | triangular_rv a s = x} = 0)`--),
	REPEAT STRIP_TAC THEN
	REWRITE_TAC [triangular_rv_def] THEN
	KNOW_TAC `!s. a * (1 - sqrt (1 - std_unif_cont s)) =
                        (\x. a * (1 - sqrt (1 - std_unif_cont x))) s` THEN1
                RW_TAC std_ss [] THEN
	ONCE_ASM_REWRITE_TAC [] THEN
	MATCH_MP_TAC CONT_PROB_BERN_EQ_0 THEN
	POP_ASSUM K_TAC THEN
	RW_TAC std_ss [CDF_def, CDF_in_events_bern_def, TRIANGULAR_RV_EQ_IN_EVENTS_BERN,
		TRIANGULAR_RV_LE_IN_EVENTS_BERN, CDF_TRIANGULAR, TRIANGULAR_CDF_CONT, GSYM triangular_rv_def]);

val _ = export_theory();


