open HolKernel Parse boolLib bossLib arithmeticTheory pred_setTheory
     listTheory sequenceTheory state_transformerTheory
     probabilityTheory formalizeUseful extra_numTheory combinTheory
     pairTheory realTheory realLib extra_boolTheory jrhUtils
     extra_pred_setTheory prob_algebraTheory probTheory realSimps
     extra_realTheory extra_pred_setTools measureTheory numTheory
     simpLib prob_extraTheory probTools prob_extraTools pred_setLib
     seqTheory std_unifTheory subtypeTheory prob_uniformTheory 
     prob_bernoulliTheory prob_bernoulli_pTheory binomialTheory
     prob_uniformTheory prob_bernoulliTheory;
     
     



val _ = new_theory "disc_expec";

fun LEAST_ELIM_TAC (asl, w) = let
  val least_terms = find_terms numSyntax.is_least w
  val tbc = TRY_CONV BETA_CONV
  fun doit t =
    if free_in t w then
      CONV_TAC (UNBETA_CONV t) THEN
      MATCH_MP_TAC whileTheory.LEAST_ELIM THEN
      CONV_TAC
        (FORK_CONV
           (BINDER_CONV tbc, (* ?n. P n *)
            BINDER_CONV      (* !n. (!m. m < n ==> ~P m) /\ P n ==> Q n *)
              (FORK_CONV
                 (FORK_CONV
                    (BINDER_CONV (RAND_CONV (RAND_CONV tbc)), (* !m.... *)
                     tbc), (* P n *)
                    tbc))))
    else NO_TAC
in
  FIRST (map doit least_terms)
end (asl, w);


val expec_def = Define
        `expec (R: (num -> bool) -> num#(num -> bool)) = 
		suminf (\n. &n * prob bern {s | FST (R s) = n})`;

val INDEP_FN_SET_FST_EVENTS = store_thm
 ("INDEP_FN_SET_FST_EVENTS",
 (--`!R. (R IN indep_fn) ==> 
	(!m. {s | FST (R s) = m} IN events bern)`--),
		RW_TAC std_ss [] THEN
		KNOW_TAC `{s | FST (R s) = m} = ((\y. y = m) o FST o R)` THEN1
			PSET_TAC [o_DEF] THEN
		RW_TAC std_ss [INDEP_FN_FST_EVENTS]);

val SUM_TWO_GENERAL = store_thm
 ("SUM_TWO_GENERAL",
 (--`!f m n p. sum(m,n) f + sum(m+n,p) f = sum(m,n + p) f`--),
	REPEAT GEN_TAC THEN
	Induct_on `p` THEN1
		RW_TAC real_ss [sum] THEN
	
		KNOW_TAC `n + SUC p = SUC (n + p)` THEN1
			DECIDE_TAC THEN
		ASM_REWRITE_TAC [] THEN
		RW_TAC real_ss [sum] THEN
		PROVE_TAC [EQ_SYM_EQ, REAL_ADD_ASSOC]);
		
val CEILING_NEG = store_thm
 ("CEILING_NEG",
 (--`!x. (x <= 0) ==> (ceiling x = 0)`--),
	RW_TAC std_ss [ceiling_def] THEN
	LEAST_ELIM_TAC THEN
	CONJ_TAC THEN1 (
		EXISTS_TAC (--`ceiling x`--) THEN
		RW_TAC std_ss [LB_CEILING]) THEN

		RW_TAC std_ss [REAL_NOT_LE] THEN
		CCONTR_TAC THEN
		FULL_SIMP_TAC arith_ss [NOT_ZERO_LT_ZERO] THEN
		KNOW_TAC `&0 < x` THEN1
			RW_TAC std_ss [] THEN
		PROVE_TAC [REAL_NOT_LE]);

val CEIL_MONO2 = store_thm
  ("CEIL_MONO2",
  (--`!m n. (m <= n) ==> (ceiling m) <= (ceiling n)`--),
        RW_TAC std_ss [] THEN
	Cases_on `0 <= m` THEN1 (
		Cases_on `0 <= n` THEN1
			RW_TAC std_ss [CEIL_MONO] THEN
		FULL_SIMP_TAC std_ss [REAL_NOT_LE] THEN
		KNOW_TAC `n < m` THEN1
			PROVE_TAC [REAL_LTE_TRANS] THEN
		PROVE_TAC [real_lt]) THEN
	
	Cases_on `0 <= n` THEN1 (
		FULL_SIMP_TAC std_ss [REAL_NOT_LE, GSYM REAL_LE] THEN
		MATCH_MP_TAC REAL_LE_TRANS THEN
		EXISTS_TAC (--`0:real`--) THEN
		CONJ_TAC THENL [
			MATCH_MP_TAC REAL_EQ_IMP_LE THEN
			RW_TAC std_ss [REAL_INJ] THEN
			MATCH_MP_TAC CEILING_NEG THEN
			RW_TAC std_ss [REAL_LT_IMP_LE, GSYM REAL_LE, GSYM REAL_LT],
			
			MATCH_MP_TAC REAL_LE_TRANS THEN
			EXISTS_TAC (--`0:real`--) THEN
			RW_TAC std_ss [REAL_LE_REFL] THEN
			MATCH_MP_TAC REAL_LE_TRANS THEN
			EXISTS_TAC (--`n:real`--) THEN
			RW_TAC std_ss [LB_CEILING]]) THEN
			
		FULL_SIMP_TAC std_ss [REAL_NOT_LE, CEILING_NEG, REAL_LT_IMP_LE]);

val SUM_PMF_CDF = store_thm
 ("SUM_PMF_CDF",
 (--`!(Y: (num -> bool) -> num#(num -> bool)) n.
        (Y IN indep_fn) ==>
        (sum (0,n) (\j. prob bern {s | FST (Y s) = j}) =
                prob bern {s | FST (Y s) < n})`--),
	RW_TAC std_ss [] THEN
    	KNOW_TAC `(\j. prob bern {s | FST (Y s) = j}) =
                  (prob bern o (\j. {s | FST (Y s) = j }))` THEN1
       		RW_TAC std_ss [o_DEF] THEN
    	ONCE_ASM_REWRITE_TAC [] THEN
    	MATCH_MP_TAC PROB_FINITELY_ADDITIVE THEN
    	RW_TAC std_ss [PROB_SPACE_BERN, DISJOINT_DEF, INTER_DEF, SPECIFICATION, IN_FUNSET] THENL [
                	ONCE_REWRITE_TAC [GSYM SPECIFICATION] THEN
                	KNOW_TAC `{s | FST (Y s) = j} =
                                 ((\y. y = j) o FST o Y)` THEN1
                        	PSET_TAC [o_DEF] THEN
                	ASM_REWRITE_TAC [] THEN
                	MATCH_MP_TAC INDEP_FN_FST_EVENTS THEN
                	RW_TAC std_ss [],

                	REWRITE_TAC [EXTENSION] THEN
                	PSET_TAC [GSPECIFICATION] THEN
                	RW_TAC real_ss [],

                	REWRITE_TAC [EXTENSION] THEN
                	PSET_TAC [GSPECIFICATION, IN_BIGUNION_IMAGE, IN_COUNT]]);

val PMF_SUMS_1 = store_thm
  ("PMF_SUMS_1",
  (--`!R. (!n. {s | FST (R s) = n} IN events bern) ==>
		((\n. prob bern {s | FST (R s) = n}) sums 1)`--),
	REPEAT STRIP_TAC THEN
	REWRITE_TAC [GSYM PROB_BERN_BASIC] THEN
    	KNOW_TAC `!n. (\n. prob bern {s | FST (R s) = n}) =
                  (prob bern o (\n. {s | FST (R s) = n}))` THEN1 
       		RW_TAC std_ss [o_DEF] THEN
    	ONCE_ASM_REWRITE_TAC [] THEN
    	MATCH_MP_TAC PROB_COUNTABLY_ADDITIVE THEN
    	RW_TAC std_ss [PROB_SPACE_BERN, DISJOINT_DEF, 
		INTER_DEF, SPECIFICATION, IN_FUNSET] THENL [
                    PSET_TAC [GSPECIFICATION] THEN
                    DISJ2_TAC THEN
                    CCONTR_TAC THEN
                    PROVE_TAC [],

                    RW_TAC std_ss [EXTENSION, IN_UNIV, IN_BIGUNION_IMAGE,
                                      GSPECIFICATION] THEN
                    EXISTS_TAC (--`(x:(num -> bool))`--) THEN
                    PROVE_TAC []]);

val SUMINF_CMUL = store_thm
  ("SUMINF_CMUL",
  (--`!f x. (summable f)  ==>
                   (suminf (\n. x * f n) = x * suminf f)`--),
	RW_TAC std_ss [summable] THEN
	KNOW_TAC `(\n. x * f n) sums (x * s)` THEN1 (
    		MATCH_MP_TAC SER_CMUL THEN
    		RW_TAC std_ss []) THEN
	PROVE_TAC [SUM_UNIQ]);

val SUMINF_ADD = store_thm
 ("SUMINF_ADD",
 (--`!f g. (summable f) /\ (summable g) ==>
                   (suminf (\n. f n + g n) = suminf f + suminf g)`--),
	RW_TAC std_ss [summable] THEN
	KNOW_TAC `(\n. f n + g n) sums (s + s')` THEN1 (
    		MATCH_MP_TAC SER_ADD THEN
    		RW_TAC std_ss []) THEN
	PROVE_TAC [SUM_UNIQ]);

val EXPEC_POSITIVE = store_thm
 ("EXPEC_POSITIVE",
 (--`!R. (R IN indep_fn) /\ 
        summable (\n. & n * prob bern {s | FST (R s) = n}) ==>
	(0 <= expec R)`--),
	RW_TAC std_ss [expec_def, suminf, summable, sums] THEN
	SELECT_ELIM_TAC THEN
	RW_TAC std_ss [] THEN1
		PROVE_TAC [] THEN

		MATCH_MP_TAC LE_SEQ_IMP_LE_LIM THEN
		EXISTS_TAC (--`(\n. sum (0,n) (\n. & n * prob bern 
		{s | FST ((R: (num -> bool) -> num#(num -> bool)) s) = n}))`--) THEN
		RW_TAC std_ss [] THEN
		MATCH_MP_TAC SUM_POS THEN
		RW_TAC std_ss [] THEN
		MATCH_MP_TAC REAL_LE_MUL THEN
		RW_TAC real_ss [] THEN
		MATCH_MP_TAC PROB_POSITIVE THEN
		RW_TAC std_ss [PROB_SPACE_BERN, INDEP_FN_SET_FST_EVENTS]);


(* Some Common Random Variable Expectations *)

(*===============*)
(*BERNOULLI*)
(*===============*)

val bernoulli_num_def = Define
        `bernoulli_num p = BIND (prob_bernoulli p) 
		(\b. UNIT (if b then (1:num) else 0))`;

val PROB_BERNOULLI_NUM_EQ1 = store_thm
 ("PROB_BERNOULLI_NUM_EQ1",
 (--`!p. 0 <= p /\ p <= 1 ==> 
	(prob bern {s | FST (bernoulli_num p s) = 1} = p)`--),
	RW_TAC std_ss [bernoulli_num_def, BIND_DEF, UNIT_DEF, LAMBDA_PAIR] THEN
	KNOW_TAC `(prob bern {s | (if FST (prob_bernoulli p s) then (1:num) else 0) = 1} = p) =
	     (prob bern {s | (if FST (prob_bernoulli p s) then (1:num) else 0) = 1} = 
						prob bern {s | FST (prob_bernoulli p s)})` THEN1
		RW_TAC std_ss [PROB_BERNOULLI] THEN
	FULL_SIMP_TAC real_ss [] THEN
	MATCH_MP_TAC RAND_THM THEN
	REWRITE_TAC [EXTENSION] THEN
	RW_TAC real_ss [GSPECIFICATION]);

val EXPEC_BERNOULLI_NUM = store_thm
 ("EXPEC_BERNOULLI_NUM",
 (--`!p. 0 <= p /\ p <= 1  ==> (expec (\s. bernoulli_num p s) = p)`--),
	RW_TAC std_ss [expec_def] THEN
	ONCE_REWRITE_TAC [EQ_SYM_EQ] THEN
	MATCH_MP_TAC SUM_UNIQ THEN
	RW_TAC std_ss [sums, SEQ] THEN
	EXISTS_TAC (--`2:num`--) THEN
	RW_TAC std_ss [GREATER_EQ] THEN
	SUFF_TAC `sum (0,n) (\x. & x * prob bern {s | FST (bernoulli_num p s) = x}) = p` THEN1
		RW_TAC real_ss [ABS_0] THEN
	Induct_on `n` THENL [
		RW_TAC real_ss [],

	 	RW_TAC real_ss [sum] THEN
	 	FULL_SIMP_TAC real_ss [LESS_OR_EQ] THENL [
				FULL_SIMP_TAC arith_ss [GSYM LESS_OR_EQ] THEN
				KNOW_TAC `prob bern {s | FST (bernoulli_num p s) = n} = 0` THEN1 (
					RW_TAC std_ss [GSYM PROB_BERN_BASIC, GSYM GSPEC_F] THEN
					MATCH_MP_TAC RAND_THM THEN
					REWRITE_TAC [EXTENSION] THEN
					PSET_TAC [GSPECIFICATION] THEN
					DISJ2_TAC THEN
					CCONTR_TAC THEN
					KNOW_TAC `FST (bernoulli_num p s) <= 1` THEN1
						RW_TAC real_ss [bernoulli_num_def, o_DEF,
						BIND_DEF, UNIT_DEF, LAMBDA_PAIR] THEN
					FULL_SIMP_TAC arith_ss []) THEN		
				RW_TAC real_ss [],

				RW_TAC real_ss [SUM_1, PROB_BERNOULLI_NUM_EQ1]]]);

(*=========*)
(*Uniform*)
(*=========*)

val PROB_UNIFORM_ALWAYS_LT_N = store_thm
 ("PROB_UNIFORM_ALWAYS_LT_N",
 (--`!n. prob bern {s | FST (prob_uniform (SUC n) s) < SUC n} = 1`--),
	GEN_TAC THEN
	MP_TAC (Q.SPEC `n` PROB_UNIFORM_RANGE) THEN
	RW_TAC std_ss [probably_bern_def, probably_def]);


val PROB_UNIFORM_EQ_N = store_thm
 ("PROB_UNIFORM_EQ_N",
 (--`!n. (prob bern {s | FST (prob_uniform (SUC n) s) = SUC n} = 0)`--),
	STRIP_TAC THEN
	KNOW_TAC `{s | FST (prob_uniform (SUC n) s) <= SUC n} =
	{s | FST (prob_uniform (SUC n) s) < SUC n} UNION 
	{s | FST (prob_uniform (SUC n) s) = SUC n}` THEN1 (
		REWRITE_TAC [EXTENSION] THEN
		RW_TAC arith_ss [IN_UNION, GSPECIFICATION]) THEN
	KNOW_TAC `prob bern {s | FST (prob_uniform (SUC n) s) <= SUC n} =
	prob bern {s | FST (prob_uniform (SUC n) s) < SUC n} +
	prob bern {s | FST (prob_uniform (SUC n) s) = SUC n}` THEN1 (
		MATCH_MP_TAC PROB_ADDITIVE THEN
		RW_TAC std_ss [PROB_SPACE_BERN, DISJOINT_DEF] THENL [
		
			KNOW_TAC  `{s | FST (prob_uniform (SUC n) s) < SUC n} =
        		((\y. y < SUC n) o FST o prob_uniform (SUC n))` THEN1
                        PSET_TAC [o_DEF] THEN
     		ASM_REWRITE_TAC [] THEN
     		MATCH_MP_TAC INDEP_FN_FST_EVENTS THEN
     		RW_TAC std_ss [INDEP_FN_PROB_UNIFORM],	

			KNOW_TAC  `{s | FST (prob_uniform (SUC n) s) = SUC n} =
        		((\y. y = SUC n) o FST o prob_uniform (SUC n))` THEN1
                        PSET_TAC [o_DEF] THEN
     		ASM_REWRITE_TAC [] THEN
     		MATCH_MP_TAC INDEP_FN_FST_EVENTS THEN
     		RW_TAC std_ss [INDEP_FN_PROB_UNIFORM],	
		
			REWRITE_TAC [EXTENSION] THEN
			PSET_TAC [IN_INTER] THEN
			RW_TAC arith_ss []]) THEN
	KNOW_TAC `prob bern {s | FST (prob_uniform (SUC n) s) < SUC n} +
          prob bern {s | FST (prob_uniform (SUC n) s) = SUC n} <= 1` THEN1 (
		MATCH_MP_TAC REAL_LE_TRANS THEN
		EXISTS_TAC (--`prob bern {s | FST (prob_uniform (SUC n) s) <= SUC n}`--) THEN
		CONJ_TAC THEN1
			FULL_SIMP_TAC std_ss [EQ_SYM_EQ, REAL_EQ_IMP_LE] THEN

			MATCH_MP_TAC PROB_LE_1 THEN
			REWRITE_TAC [PROB_SPACE_BERN] THEN
			KNOW_TAC  `{s | FST (prob_uniform (SUC n) s) <= SUC n} =
        		((\y. y <= SUC n) o FST o prob_uniform (SUC n))` THEN1
                        	PSET_TAC [o_DEF] THEN
     		ASM_REWRITE_TAC [] THEN
     		MATCH_MP_TAC INDEP_FN_FST_EVENTS THEN
     		RW_TAC std_ss [INDEP_FN_PROB_UNIFORM]) THEN	
	POP_ASSUM MP_TAC THEN
	KILL_ALL_TAC THEN
	RW_TAC std_ss [PROB_UNIFORM_ALWAYS_LT_N] THEN
	KNOW_TAC `prob bern {s | FST (prob_uniform (SUC n) s) = SUC n} <= 0` THEN1 (
		KNOW_TAC `1 + prob bern {s | FST (prob_uniform (SUC n) s) = SUC n} <= 1 + 0` THEN1
			RW_TAC real_ss [] THEN
		FULL_SIMP_TAC std_ss [REAL_LE_LADD]) THEN
	RW_TAC std_ss [GSYM REAL_LE_ANTISYM] THEN
	MATCH_MP_TAC PROB_POSITIVE THEN
	RW_TAC std_ss [PROB_SPACE_BERN] THEN
	KNOW_TAC  `{s | FST (prob_uniform (SUC n) s) = SUC n} =
        		((\y. y = SUC n) o FST o prob_uniform (SUC n))` THEN1
                        PSET_TAC [o_DEF] THEN
     		ASM_REWRITE_TAC [] THEN
     		MATCH_MP_TAC INDEP_FN_FST_EVENTS THEN
     		RW_TAC std_ss [INDEP_FN_PROB_UNIFORM]);


val PROB_UNIFORM_GE_SUC = store_thm
 ("PROB_UNIFORM_GE_SUC",
 (--`!n k. (SUC n <= SUC k) ==>
	(prob bern {s | FST (prob_uniform (SUC n) s) = SUC k} = 0)`--),
	REPEAT STRIP_TAC THEN
	KNOW_TAC `{s | FST (prob_uniform (SUC n) s) <= SUC k} =
	{s | FST (prob_uniform (SUC n) s) < SUC k} UNION 
	{s | FST (prob_uniform (SUC n) s) = SUC k}` THEN1 (
		REWRITE_TAC [EXTENSION] THEN
		RW_TAC arith_ss [IN_UNION, GSPECIFICATION]) THEN
	KNOW_TAC `prob bern {s | FST (prob_uniform (SUC n) s) <= SUC k} =
	prob bern {s | FST (prob_uniform (SUC n) s) < SUC k} +
	prob bern {s | FST (prob_uniform (SUC n) s) = SUC k}` THEN1 (
		MATCH_MP_TAC PROB_ADDITIVE THEN
		RW_TAC std_ss [PROB_SPACE_BERN, DISJOINT_DEF] THENL [
		
			KNOW_TAC  `{s | FST (prob_uniform (SUC n) s) < SUC k} =
        		((\y. y < SUC k) o FST o prob_uniform (SUC n))` THEN1
                        	PSET_TAC [o_DEF] THEN
     			ASM_REWRITE_TAC [] THEN
     			MATCH_MP_TAC INDEP_FN_FST_EVENTS THEN
     			RW_TAC std_ss [INDEP_FN_PROB_UNIFORM],	

			KNOW_TAC  `{s | FST (prob_uniform (SUC n) s) = SUC k} =
        		((\y. y = SUC k) o FST o prob_uniform (SUC n))` THEN1
                        	PSET_TAC [o_DEF] THEN
     			ASM_REWRITE_TAC [] THEN
     			MATCH_MP_TAC INDEP_FN_FST_EVENTS THEN
     			RW_TAC std_ss [INDEP_FN_PROB_UNIFORM],	
		
			REWRITE_TAC [EXTENSION] THEN
			PSET_TAC [IN_INTER] THEN
			RW_TAC arith_ss []]) THEN
	KNOW_TAC `prob bern {s | FST (prob_uniform (SUC n) s) < SUC k} +
	  			prob bern {s | FST (prob_uniform (SUC n) s) = SUC k} <= 1` THEN1 (
				MATCH_MP_TAC REAL_LE_TRANS THEN
				EXISTS_TAC (--`prob bern {s | FST (prob_uniform (SUC n) s) <= SUC k}`--) THEN
				CONJ_TAC THEN1
					FULL_SIMP_TAC std_ss [EQ_SYM_EQ, REAL_EQ_IMP_LE] THEN

				MATCH_MP_TAC PROB_LE_1 THEN
				REWRITE_TAC [PROB_SPACE_BERN] THEN
				KNOW_TAC  `{s | FST (prob_uniform (SUC n) s) <= SUC k} =
        				((\y. y <= SUC k) o FST o prob_uniform (SUC n))` THEN1
                        		PSET_TAC [o_DEF] THEN
     				ASM_REWRITE_TAC [] THEN
     				MATCH_MP_TAC INDEP_FN_FST_EVENTS THEN
     				RW_TAC std_ss [INDEP_FN_PROB_UNIFORM]) THEN	
	POP_ASSUM MP_TAC THEN
	NTAC 2 (POP_ASSUM K_TAC) THEN
	DISCH_TAC THEN
	KNOW_TAC `{s | FST (prob_uniform (SUC n) s) < SUC k} =
	{s | FST (prob_uniform (SUC n) s) < SUC n} UNION 
	{s | (SUC n <= FST (prob_uniform (SUC n) s)) /\ 
			FST (prob_uniform (SUC n) s) < SUC k}` THEN1 (
		REWRITE_TAC [EXTENSION] THEN
		RW_TAC arith_ss [IN_UNION, GSPECIFICATION, LEFT_OR_OVER_AND]) THEN
	KNOW_TAC `prob bern {s | FST (prob_uniform (SUC n) s) < SUC k} =
	prob bern {s | FST (prob_uniform (SUC n) s) < SUC n} +
	prob bern {s | (SUC n <= FST (prob_uniform (SUC n) s)) /\ 
			FST (prob_uniform (SUC n) s) < SUC k}` THEN1 (
		MATCH_MP_TAC PROB_ADDITIVE THEN
		RW_TAC std_ss [PROB_SPACE_BERN, DISJOINT_DEF] THENL [
		
			KNOW_TAC  `{s | FST (prob_uniform (SUC n) s) < SUC n} =
        		((\y. y < SUC n) o FST o prob_uniform (SUC n))` THEN1
                        	PSET_TAC [o_DEF] THEN
     			ASM_REWRITE_TAC [] THEN
     			MATCH_MP_TAC INDEP_FN_FST_EVENTS THEN
     			RW_TAC std_ss [INDEP_FN_PROB_UNIFORM],	

			KNOW_TAC `{s |
     			SUC n <= FST (prob_uniform (SUC n) s) /\
     				FST (prob_uniform (SUC n) s) < SUC k} =
			     {s | SUC n <= FST (prob_uniform (SUC n) s)} INTER
				{s | FST (prob_uniform (SUC n) s) < SUC k}` THEN1 (
				REWRITE_TAC [EXTENSION] THEN
				PSET_TAC [IN_INTER, GSPECIFICATION] THEN
				RW_TAC arith_ss []) THEN
			ONCE_ASM_REWRITE_TAC [] THEN
			MATCH_MP_TAC EVENTS_INTER THEN
			REWRITE_TAC [PROB_SPACE_BERN] THEN
			CONJ_TAC THENL [
				KNOW_TAC  `{s | SUC n <= FST (prob_uniform (SUC n) s)} =
        				((\y. SUC n <= y) o FST o prob_uniform (SUC n))` THEN1
                        		PSET_TAC [o_DEF] THEN
     				ASM_REWRITE_TAC [] THEN
     				MATCH_MP_TAC INDEP_FN_FST_EVENTS THEN
     				RW_TAC std_ss [INDEP_FN_PROB_UNIFORM],	

				KILL_ALL_TAC THEN
				KNOW_TAC  `{s | FST (prob_uniform (SUC n) s) < SUC k} =
        				((\y. y < SUC k) o FST o prob_uniform (SUC n))` THEN1
                        		PSET_TAC [o_DEF] THEN
     				ASM_REWRITE_TAC [] THEN
     				MATCH_MP_TAC INDEP_FN_FST_EVENTS THEN
     				RW_TAC std_ss [INDEP_FN_PROB_UNIFORM]],
		
			REWRITE_TAC [EXTENSION] THEN
			PSET_TAC [IN_INTER] THEN
			RW_TAC arith_ss []]) THEN
	POP_ASSUM MP_TAC THEN
	POP_ASSUM K_TAC THEN
	DISCH_TAC THEN
	FULL_SIMP_TAC std_ss [PROB_UNIFORM_ALWAYS_LT_N] THEN
	POP_ASSUM K_TAC THEN
	KNOW_TAC `prob bern
            {s |
             SUC n <= FST (prob_uniform (SUC n) s) /\
             FST (prob_uniform (SUC n) s) < SUC k} +
          prob bern {s | FST (prob_uniform (SUC n) s) = SUC k} <= 0` THEN1 (
		KNOW_TAC `1 +
          		prob bern
            		{s |
             		SUC n <= FST (prob_uniform (SUC n) s) /\
             		FST (prob_uniform (SUC n) s) < SUC k} +
          		prob bern {s | FST (prob_uniform (SUC n) s) = SUC k} <= 1 + 0` THEN1 
				RW_TAC real_ss [] THEN
		FULL_SIMP_TAC std_ss [REAL_LE_LADD, GSYM REAL_ADD_ASSOC]) THEN
	KNOW_TAC `0 <= prob bern
            {s |
             SUC n <= FST (prob_uniform (SUC n) s) /\
             FST (prob_uniform (SUC n) s) < SUC k} +
          prob bern {s | FST (prob_uniform (SUC n) s) = SUC k}` THEN1 (
		MATCH_MP_TAC REAL_LE_ADD THEN
		CONJ_TAC THENL [
			MATCH_MP_TAC PROB_POSITIVE THEN
			RW_TAC std_ss [PROB_SPACE_BERN] THEN
			KNOW_TAC `{s |
     			SUC n <= FST (prob_uniform (SUC n) s) /\
     				FST (prob_uniform (SUC n) s) < SUC k} =
			     {s | SUC n <= FST (prob_uniform (SUC n) s)} INTER
				{s | FST (prob_uniform (SUC n) s) < SUC k}` THEN1 (
				REWRITE_TAC [EXTENSION] THEN
				PSET_TAC [IN_INTER, GSPECIFICATION] THEN
				RW_TAC arith_ss []) THEN
			ONCE_ASM_REWRITE_TAC [] THEN
			MATCH_MP_TAC EVENTS_INTER THEN
			REWRITE_TAC [PROB_SPACE_BERN] THEN
			CONJ_TAC THENL [
				KNOW_TAC  `{s | SUC n <= FST (prob_uniform (SUC n) s)} =
        				((\y. SUC n <= y) o FST o prob_uniform (SUC n))` THEN1
                        	PSET_TAC [o_DEF] THEN
     			ASM_REWRITE_TAC [] THEN
     			MATCH_MP_TAC INDEP_FN_FST_EVENTS THEN
     			RW_TAC std_ss [INDEP_FN_PROB_UNIFORM],	

				KILL_ALL_TAC THEN
				KNOW_TAC  `{s | FST (prob_uniform (SUC n) s) < SUC k} =
        				((\y. y < SUC k) o FST o prob_uniform (SUC n))` THEN1
                        		PSET_TAC [o_DEF] THEN
     			ASM_REWRITE_TAC [] THEN
     			MATCH_MP_TAC INDEP_FN_FST_EVENTS THEN
     			RW_TAC std_ss [INDEP_FN_PROB_UNIFORM]], 


			MATCH_MP_TAC PROB_POSITIVE THEN
			RW_TAC std_ss [PROB_SPACE_BERN] THEN
			KNOW_TAC  `{s | FST (prob_uniform (SUC n) s) = SUC k} =
        			((\y. y = SUC k) o FST o prob_uniform (SUC n))` THEN1
                        	PSET_TAC [o_DEF] THEN
     			ASM_REWRITE_TAC [] THEN
     			MATCH_MP_TAC INDEP_FN_FST_EVENTS THEN
     			RW_TAC std_ss [INDEP_FN_PROB_UNIFORM]]) THEN	
	KNOW_TAC `prob bern
            {s |
             SUC n <= FST (prob_uniform (SUC n) s) /\
             FST (prob_uniform (SUC n) s) < SUC k} +
          prob bern {s | FST (prob_uniform (SUC n) s) = SUC k} = 0` THEN1
		RW_TAC std_ss [GSYM REAL_LE_ANTISYM] THEN
	POP_ASSUM MP_TAC THEN
	KILL_ALL_TAC THEN
	DISCH_TAC THEN
	CCONTR_TAC THEN
	KNOW_TAC `prob bern
            {s |
             SUC n <= FST (prob_uniform (SUC n) s) /\
             FST (prob_uniform (SUC n) s) < SUC k} < 0` THEN1 (
		FULL_SIMP_TAC std_ss [REAL_LNEG_UNIQ, REAL_NEG_LT0] THEN
		KNOW_TAC `0 <= prob bern {s | FST (prob_uniform (SUC n) s) = SUC k}` THEN1 (
			MATCH_MP_TAC PROB_POSITIVE THEN
			RW_TAC std_ss [PROB_SPACE_BERN] THEN
			KNOW_TAC  `{s | FST (prob_uniform (SUC n) s) = SUC k} =
        			((\y. y = SUC k) o FST o prob_uniform (SUC n))` THEN1
                        	PSET_TAC [o_DEF] THEN
     			ASM_REWRITE_TAC [] THEN
     			MATCH_MP_TAC INDEP_FN_FST_EVENTS THEN
     			RW_TAC std_ss [INDEP_FN_PROB_UNIFORM]) THEN	
		PROVE_TAC [REAL_LE_LT]) THEN
	POP_ASSUM MP_TAC THEN
	RW_TAC std_ss [REAL_NOT_LT] THEN
	MATCH_MP_TAC PROB_POSITIVE THEN
	RW_TAC std_ss [PROB_SPACE_BERN] THEN
			KNOW_TAC `{s |
     			SUC n <= FST (prob_uniform (SUC n) s) /\
     				FST (prob_uniform (SUC n) s) < SUC k} =
			     {s | SUC n <= FST (prob_uniform (SUC n) s)} INTER
				{s | FST (prob_uniform (SUC n) s) < SUC k}` THEN1 (
				REWRITE_TAC [EXTENSION] THEN
				PSET_TAC [IN_INTER, GSPECIFICATION] THEN
				RW_TAC arith_ss []) THEN
			ONCE_ASM_REWRITE_TAC [] THEN
			MATCH_MP_TAC EVENTS_INTER THEN
			REWRITE_TAC [PROB_SPACE_BERN] THEN
			CONJ_TAC THENL [
				KNOW_TAC  `{s | SUC n <= FST (prob_uniform (SUC n) s)} =
        				((\y. SUC n <= y) o FST o prob_uniform (SUC n))` THEN1
                        		PSET_TAC [o_DEF] THEN
     			ASM_REWRITE_TAC [] THEN
     			MATCH_MP_TAC INDEP_FN_FST_EVENTS THEN
     			RW_TAC std_ss [INDEP_FN_PROB_UNIFORM],	

			KILL_ALL_TAC THEN
			KNOW_TAC  `{s | FST (prob_uniform (SUC n) s) < SUC k} =
        				((\y. y < SUC k) o FST o prob_uniform (SUC n))` THEN1
                        	PSET_TAC [o_DEF] THEN
     			ASM_REWRITE_TAC [] THEN
     			MATCH_MP_TAC INDEP_FN_FST_EVENTS THEN
     			RW_TAC std_ss [INDEP_FN_PROB_UNIFORM]]);

val PROB_UNIFORM_GE = store_thm
 ("PROB_UNIFORM_GE",
 (--`!n k. SUC n <= k ==>
         (prob bern {s | FST (prob_uniform (SUC n) s) = k} = 0)`--),
	GEN_TAC THEN
	Induct THEN1
		RW_TAC real_ss [] THEN
	RW_TAC real_ss [PROB_UNIFORM_GE_SUC]);

val ALL_NUM_SER = store_thm
  ("ALL_NUM_SER",
  (--`!n. sum (0,n) & = &n * (&n - 1)/2`--),
	Induct THEN1
     		RW_TAC real_ss [sum] THEN
	RW_TAC arith_ss [sum, ADD1, GSYM REAL_ADD] THEN
	RW_TAC real_ss [] THEN
	KNOW_TAC `& n * (& n - 1) / 2 + & n =
        	(& n * (& n - 1) + 2 * &n)/ 2` THEN1 (
        	RW_TAC std_ss [real_div, REAL_ADD_RDISTRIB,REAL_EQ_LADD] THEN
        	ONCE_REWRITE_TAC [REAL_MUL_SYM] THEN
        	ONCE_REWRITE_TAC [REAL_MUL_ASSOC] THEN
        	FULL_SIMP_TAC real_ss [REAL_MUL_ASSOC,REAL_MUL_LINV, REAL_POS_NZ]) THEN
	ASM_REWRITE_TAC [] THEN
	RW_TAC std_ss [GSYM REAL_MUL, GSYM REAL_ADD,
		real_div, REAL_EQ_RMUL] THEN
	DISJ2_TAC THEN
	REAL_ARITH_TAC);


val EXPEC_UNIFORM = store_thm
 ("EXPEC_UNIFORM",
 (--`!n. (expec (\s. prob_uniform (SUC n) s) = &n/2)`--),
	RW_TAC std_ss [expec_def] THEN
	ONCE_REWRITE_TAC [EQ_SYM_EQ] THEN
	MATCH_MP_TAC SUM_UNIQ THEN
	RW_TAC std_ss [sums, SEQ] THEN
	EXISTS_TAC (--`SUC (SUC n: num)`--) THEN
	Induct THEN1
		RW_TAC real_ss [] THEN

		FULL_SIMP_TAC std_ss [GREATER_EQ, sum] THEN
		RW_TAC std_ss [LESS_OR_EQ] THENL [	
				FULL_SIMP_TAC arith_ss [] THEN
				SUFF_TAC `& n' * prob bern {s | 
				FST (prob_uniform (SUC n) s) = n'} = 0` THEN1
					RW_TAC real_ss [ABS_0] THEN

					POP_ASSUM MP_TAC THEN
					KILL_ALL_TAC THEN
					Cases_on `n'` THEN1
						RW_TAC real_ss [] THEN
			
						RW_TAC real_ss [REAL_ENTIRE, PROB_UNIFORM_GE],
					
				RW_TAC real_ss [PROB_UNIFORM_EQ_N] THEN			
				POP_ASSUM K_TAC THEN
				KNOW_TAC `!n. sum (0,SUC n)
         				(\x. & x * prob bern 
					{s | FST (prob_uniform (SUC n) s) = x}) = 
						sum (0,SUC n) (\x. & x * (1/(&(SUC n))))` THEN1 (	
					GEN_TAC THEN
					MATCH_MP_TAC SUM_EQ THEN
					RW_TAC real_ss [PROB_BERN_UNIFORM]) THEN
				ASM_REWRITE_TAC [] THEN
				SUFF_TAC `sum (0,SUC n) (\x. & x * (1 / & (SUC n))) = & n / 2` THEN1
					RW_TAC real_ss [ABS_0] THEN
				ONCE_REWRITE_TAC [REAL_MUL_SYM] THEN
				RW_TAC real_ss [SUM_CMUL, ALL_NUM_SER, real_div, 
				REAL_MUL_LINV, REAL_MUL_ASSOC] THEN
				RW_TAC arith_ss [ADD1, GSYM REAL_ADD] THEN
				REAL_ARITH_TAC]);

						
(*=====================================*)
(*geometric random variable*)
(*=====================================*)

val prob_geometric_p_iter_def = Define
  `prob_geometric_p_iter p s = BIND (prob_bernoulli (1-p)) (\b. UNIT (b, SUC (SND s)))`;

val prob_geometric_p_loop_def = Define
  `prob_geometric_p_loop p = prob_while FST (prob_geometric_p_iter p)`;

val prob_geometric_p_def = Define
  `prob_geometric_p p =
   BIND (BIND (UNIT (T, SUC 0)) (prob_geometric_p_loop p)) (\s. UNIT (SND s - 1))`;

val INDEP_FN_PROB_GEOMETRIC_P_ITER = store_thm
  ("INDEP_FN_PROB_GEOMETRIC_P_ITER",
  (--`!a p. prob_geometric_p_iter p a IN indep_fn`--),
	RW_TAC std_ss [prob_geometric_p_iter_def, INDEP_FN_PROB_BERNOULLI, INDEP_FN_BIND,
                  INDEP_FN_UNIT]);

val PROB_GEOMETRIC_P_LOOP_TERMINATES = store_thm
 ("PROB_GEOMETRIC_P_LOOP_TERMINATES",
 (--`!p. 0 < p /\ p <= 1 ==> (prob_while_terminates FST (prob_geometric_p_iter p))`--),
	REPEAT STRIP_TAC THEN
	MATCH_MP_TAC PROB_WHILE_TERMINATES_SUFFICIENT THEN
	RW_TAC std_ss [INDEP_FN_PROB_GEOMETRIC_P_ITER, possibly_bern_def,
                     possibly_def] THENL [
		RW_TAC std_ss [prob_geometric_p_iter_def, UNIT_DEF, BIND_DEF, o_DEF,
                     UNCURRY] THEN
		KNOW_TAC `{s | ~FST (prob_bernoulli (1-p) s)} = 
			(\m. m = F) o FST o (prob_bernoulli (1-p))` THEN1
			PSET_TAC [o_DEF] THEN
		ASM_REWRITE_TAC [] THEN
		MATCH_MP_TAC INDEP_FN_FST_EVENTS THEN
		RW_TAC std_ss [INDEP_FN_PROB_BERNOULLI],

		RW_TAC std_ss [prob_geometric_p_iter_def, UNIT_DEF, BIND_DEF, o_DEF,
                     UNCURRY] THEN
		KNOW_TAC `prob bern {s | ~FST (prob_bernoulli (1-p) s)} = 
		1 - prob bern {s | FST (prob_bernoulli (1-p) s)}` THEN1 (
			MATCH_MP_TAC INDEP_FN_PROB_FST_NOT THEN
			RW_TAC std_ss [INDEP_FN_PROB_BERNOULLI]) THEN
		KNOW_TAC `0 <= 1 - p /\ (1 - p <= 1)` THEN1 (
			RW_TAC std_ss [REAL_SUB_LE, REAL_LE_SUB_RADD,
			REAL_LE_ADDR, REAL_LT_IMP_LE]) THEN
		RW_TAC std_ss [PROB_BERNOULLI, REAL_LT_IMP_LE] THEN
		MATCH_MP_TAC REAL_POS_NZ THEN
		ASM_REWRITE_TAC [REAL_SUB_LT, REAL_LT_SUB_RADD, REAL_LT_ADDR]]);

val INDEP_FN_PROB_GEOMETRIC_P_LOOP = store_thm
 ("INDEP_FN_PROB_GEOMETRIC_P_LOOP",
 (--`!a p. 0 < p /\ p <= 1 ==> (prob_geometric_p_loop p a IN indep_fn)`--),
	RW_TAC std_ss [prob_geometric_p_loop_def, INDEP_FN_PROB_WHILE,
                  PROB_GEOMETRIC_P_LOOP_TERMINATES,
                  INDEP_FN_PROB_GEOMETRIC_P_ITER]);

val INDEP_FN_PROB_GEOMETRIC_P = store_thm
 ("INDEP_FN_PROB_GEOMETRIC_P",
 (--`!p. 0 < p /\ p <= 1 ==> (prob_geometric_p p IN indep_fn)`--),
	RW_TAC std_ss [prob_geometric_p_def, INDEP_FN_BIND, INDEP_FN_UNIT,
                  INDEP_FN_PROB_GEOMETRIC_P_LOOP]);

val PROB_GEOMETRIC_P_LOOP_F = store_thm
 ("PROB_GEOMETRIC_P_LOOP_F",
 (--`!n p. 0 < p /\ p <= 1 ==> 
	(!*s. FST (prob_geometric_p_loop p (F, n) s) = (F, n))`--),
	RW_TAC std_ss [prob_geometric_p_loop_def] THEN
	MP_TAC (Q.ISPEC `\w : bool # num. w = (F, n)` PROB_WHILE) THEN
	RW_TAC std_ss [] THEN
	POP_ASSUM MATCH_MP_TAC THEN
	RW_TAC std_ss [PROB_GEOMETRIC_P_LOOP_TERMINATES,
                     INDEP_FN_PROB_GEOMETRIC_P_ITER] THEN
	MATCH_MP_TAC DEFINITELY_PROBABLY THEN
	Cases_on `n'` THEN1
	RW_TAC std_ss [prob_while_cut_def, BIND_DEF, o_THM, UNCURRY,
                     prob_geometric_p_iter_def, UNIT_DEF, prob_bernoulli_def] THEN
	RW_TAC std_ss [prob_while_cut_def, BIND_DEF, o_THM, UNCURRY,
                     prob_geometric_p_iter_def, UNIT_DEF, prob_bernoulli_def]);

val PROB_GEOMETRIC_P_LOOP_T_RANGE = store_thm
 ("PROB_GEOMETRIC_P_LOOP_T_RANGE",
 (--`!n p. 0 < p /\ p <= 1 ==>
	(!*s. n < SND (FST (prob_geometric_p_loop p (T,n) s)))`--),
	RW_TAC std_ss [prob_geometric_p_loop_def] THEN
	MP_TAC (Q.ISPEC `\w : bool # num. n < SND w` PROB_WHILE) THEN
	RW_TAC std_ss [] THEN
	POP_ASSUM MATCH_MP_TAC THEN
	RW_TAC std_ss [PROB_GEOMETRIC_P_LOOP_TERMINATES,
                     INDEP_FN_PROB_GEOMETRIC_P_ITER] THEN
	MATCH_MP_TAC DEFINITELY_PROBABLY THEN
	Q.SPEC_TAC (`n`, `n`) THEN
	Induct_on `n'` THEN1
	 	RW_TAC std_ss [prob_while_cut_def, UNIT_DEF] THEN
	REPEAT GEN_TAC THEN
	REVERSE (Cases_on `(FST (prob_bernoulli (1-p) s))`) THENL [
		FULL_SIMP_TAC std_ss [prob_while_cut_def, BIND_DEF, o_THM, UNCURRY,
                         prob_geometric_p_iter_def, UNIT_DEF] THEN
		Cases_on `n'` THEN1
			RW_TAC arith_ss [prob_while_cut_def, UNIT_DEF] THEN
			RW_TAC arith_ss [prob_while_cut_def, UNIT_DEF],

		FULL_SIMP_TAC std_ss [prob_while_cut_def, BIND_DEF, o_THM, UNCURRY,
                         prob_geometric_p_iter_def, UNIT_DEF] THEN
		STRIP_TAC THEN
		MATCH_MP_TAC LESS_TRANS THEN
		Q.EXISTS_TAC `SUC n` THEN
        	CONJ_TAC THEN1
                	DECIDE_TAC THEN
		RW_TAC arith_ss []]);

val PROB_GEOMETRIC_P_LOOP_T_RANGE2 = store_thm
 ("PROB_GEOMETRIC_P_LOOP_T_RANGE2",
 (--`!n p. 0 < p /\ p <= 1 ==>
	(!*s. n < SND (FST (prob_geometric_p_loop p (T,(SUC n)) s)))`--),
	REPEAT STRIP_TAC THEN
	MP_TAC (Q.SPECL [`SUC n`, `p`] PROB_GEOMETRIC_P_LOOP_T_RANGE) THEN
	KNOW_TAC `!*s. n < SUC n` THEN1 (
		MATCH_MP_TAC DEFINITELY_PROBABLY THEN
		DECIDE_TAC) THEN
	STRIP_TAC THEN
	KNOW_TAC `!*s. n < SUC n /\ SUC n < 
		SND (FST (prob_geometric_p_loop p (T,SUC n) s))` THEN1 (
		RW_TAC std_ss [PROBABLY_CONJ]) THEN
	POP_ASSUM MP_TAC THEN
	NTAC 2 (POP_ASSUM K_TAC) THEN
	KNOW_TAC `{s | n < SND (FST (prob_geometric_p_loop p (T,SUC n) s))} IN events bern` THEN1 (
		SUFF_TAC `{s | n < SND (FST (prob_geometric_p_loop p (T,SUC n) s))} =
                   (\x. n < SND x) o FST o prob_geometric_p_loop p (T, SUC n)` THEN1 (
               				RW_TAC std_ss [] THEN
					MATCH_MP_TAC INDEP_FN_FST_EVENTS THEN
                   			PROVE_TAC [INDEP_FN_PROB_GEOMETRIC_P_LOOP]) THEN
		PSET_TAC [o_THM]) THEN
	RW_TAC std_ss [probably_bern_def, probably_def] THEN
	SUFF_TAC `prob bern {s | n < SND (FST (prob_geometric_p_loop p (T,SUC n) s))}  = 
		prob bern
            	{s |
            	 n < SUC n /\
             	SUC n < SND (FST (prob_geometric_p_loop p (T,SUC n) s))}` THEN1
		RW_TAC std_ss [] THEN
	REWRITE_TAC [GSYM REAL_LE_ANTISYM] THEN
		CONJ_TAC THENL [
			RW_TAC std_ss [] THEN
			MATCH_MP_TAC PROB_LE_1 THEN
			RW_TAC std_ss [PROB_SPACE_BERN],

			MATCH_MP_TAC PROB_INCREASING THEN
			RW_TAC std_ss [PROB_SPACE_BERN, SUBSET_DEF, 
			GSPECIFICATION, prob_geometric_p_loop_def] THEN
				MATCH_MP_TAC LESS_TRANS THEN
				EXISTS_TAC (--`SUC n`--) THEN
				RW_TAC std_ss []]);


val PROB_GEOMETRIC_P_LOOP_STOPS = store_thm
 ("PROB_GEOMETRIC_P_LOOP_STOPS",
 (--`!n p. 0 < p /\ p <= 1 ==>
	(!*s. (SND (FST (prob_geometric_p_loop p (b,n) s)) = n) = ~b)`--),
	RW_TAC std_ss [prob_geometric_p_loop_def] THEN
	MP_TAC (Q.ISPEC `\w : bool # num. (SND w = n) = ~b` PROB_WHILE) THEN
	RW_TAC std_ss [] THEN
	POP_ASSUM MATCH_MP_TAC THEN
	RW_TAC std_ss [PROB_GEOMETRIC_P_LOOP_TERMINATES,
                     INDEP_FN_PROB_GEOMETRIC_P_ITER] THEN
	Cases_on `n'` THEN1 (
        	RW_TAC std_ss [prob_while_cut_def, UNIT_DEF] THEN
                	MATCH_MP_TAC DEFINITELY_PROBABLY THEN
                	RW_TAC std_ss []) THEN

        	RW_TAC std_ss [prob_while_cut_def, UNIT_DEF, PROBABLY_TRUE, BIND_DEF,
                     UNCURRY, o_THM, prob_geometric_p_iter_def] THEN
        	MATCH_MP_TAC DEFINITELY_PROBABLY THEN
        	RW_TAC std_ss [] THEN
        	POP_ASSUM K_TAC THEN
        	KNOW_TAC `!a b : num. b < a ==> ~(a = b)` THEN1
                	DECIDE_TAC THEN
        	KNOW_TAC `n < SUC n` THEN1
                	DECIDE_TAC THEN
        	POP_ASSUM MP_TAC THEN
        	Q.SPEC_TAC (`SUC n`, `m`) THEN
		Q.SPEC_TAC (`((prob_bernoulli (1-p) s))`, `A`) THEN
        	Induct_on `n''` THEN1
                	RW_TAC arith_ss [prob_while_cut_def, UNIT_DEF] THEN

                	RW_TAC arith_ss [prob_while_cut_def, BIND_DEF, UNIT_DEF, UNCURRY, o_THM,
                       prob_geometric_p_iter_def, sdest_def]);

val PROB_BERN_GEOMETRIC_P_LOOP_LEMMA = store_thm
 ("PROB_BERN_GEOMETRIC_P_LOOP_LEMMA",
 (--`!d n p. 0 < p /\ p <= 1 ==>
       (prob bern
       ((\x. SND x = n + SUC d) o FST o
        BIND (prob_bernoulli (1-p)) (\a. BIND (UNIT (a,SUC n)) (prob_geometric_p_loop p))) =
       (1-p) *
       prob bern
       ((\x. SND x = n + SUC d) o FST o
        (\a. BIND (UNIT (a,SUC n)) (prob_geometric_p_loop p)) T) +
       p *
       prob bern
       ((\x. SND x = n + SUC d) o FST o
        (\a. BIND (UNIT (a,SUC n)) (prob_geometric_p_loop p)) F))`--),
	REPEAT STRIP_TAC THEN
	(MP_TAC o
       Q.SPECL
       [`(prob_bernoulli (1-p))`,
        `\a. BIND (UNIT (a,SUC n)) (prob_while FST (prob_geometric_p_iter p))`,
        `1-p`,
        `\x. SND x = n + SUC d`] o
       INST_TYPE [alpha |-> Type`:bool # num`])
      PROB_BERN_BIND_BOOL THEN
	KNOW_TAC `(!x. (\a. BIND (UNIT (a,SUC n)) (prob_while FST (prob_geometric_p_iter p)))
          			x IN indep_fn)` THEN1 (
		GEN_TAC THEN
		RW_TAC std_ss [GSYM prob_geometric_p_loop_def] THEN
		MATCH_MP_TAC INDEP_FN_BIND THEN
		RW_TAC std_ss [INDEP_FN_UNIT, INDEP_FN_PROB_GEOMETRIC_P_LOOP]) THEN
	KNOW_TAC `(prob bern (FST o prob_bernoulli (1-p)) = (1-p))` THEN1 (
		KNOW_TAC `(FST o prob_bernoulli (1-p)) = {s | FST (prob_bernoulli (1-p) s)}` THEN1
			PSET_TAC [o_THM] THEN
		ASM_REWRITE_TAC [] THEN
		MATCH_MP_TAC PROB_BERNOULLI THEN
		RW_TAC std_ss [REAL_SUB_LE, REAL_LE_SUB_RADD,
			REAL_LE_ADDR, REAL_LT_IMP_LE]) THEN
	RW_TAC std_ss [INDEP_FN_PROB_BERNOULLI] THEN
	ASM_REWRITE_TAC [prob_geometric_p_loop_def] THEN
	REAL_ARITH_TAC);

val PROB_BERN_GEOMETRIC_P_LOOP = store_thm
 ("PROB_BERN_GEOMETRIC_P_LOOP",
 (--`!n d p. 0 < p /\ p <= 1 ==>
       (prob bern {s | SND (FST (prob_geometric_p_loop p (T,n) s)) = n + SUC d} =
       p * (1-p) pow d)`--),
	RW_TAC std_ss [prob_geometric_p_loop_def] THEN
	RW_TAC std_ss [PROB_WHILE_ADVANCE, INDEP_FN_PROB_GEOMETRIC_P_ITER,
                     PROB_GEOMETRIC_P_LOOP_TERMINATES, prob_geometric_p_iter_def,
                     GSYM BIND_ASSOC] THEN
	KNOW_TAC `!f : (num -> bool) -> (bool # num) # (num -> bool).
         {s | SND (FST (f s)) = n + SUC d} = (\x. SND x = n + SUC d) o FST o f` THEN1 (
		SET_EQ_TAC THEN
        	PSET_TAC [] THEN
        	RW_TAC std_ss [SPECIFICATION, o_THM]) THEN
	ASM_REWRITE_TAC [] THEN
	POP_ASSUM K_TAC THEN
	RW_TAC std_ss [GSYM prob_geometric_p_loop_def] THEN
	Q.SPEC_TAC (`n`, `n`) THEN
	Induct_on `d` THENL [
   		RW_TAC std_ss [PROB_BERN_GEOMETRIC_P_LOOP_LEMMA] THEN
        	RW_TAC real_ss [ADD_CLAUSES, pow, o_DEF, BIND_DEF, UNCURRY, UNIT_DEF,
                          prob_geometric_p_loop_def] THEN
       	KNOW_TAC `prob bern
           (\x.
              SND (FST (prob_while FST (prob_geometric_p_iter p) (F,SUC n) x)) =
              SUC n) =
           prob bern UNIV` THEN1 (
       			MATCH_MP_TAC PROB_BERN_UNIVERSAL THEN
           		Q.EXISTS_TAC
              		`\s. (SND (FST (prob_geometric_p_loop p (F, SUC n) s)) = SUC n) = ~F` THEN
           		RW_TAC std_ss [PROB_GEOMETRIC_P_LOOP_STOPS, EVENTS_BERN_UNIV,
                             IN_UNIV] THENL [
	    			RW_TAC std_ss [GSYM prob_geometric_p_loop_def] THEN
				SUFF_TAC `(\x. SND (FST (prob_geometric_p_loop p (F,SUC n) x)) 
				= SUC n) =
                   		(\x. SND x = SUC n) o FST o prob_geometric_p_loop p (F, SUC n)` THEN1 (
               				RW_TAC std_ss [] THEN
					MATCH_MP_TAC INDEP_FN_FST_EVENTS THEN
                   			PROVE_TAC [INDEP_FN_PROB_GEOMETRIC_P_LOOP]) THEN
				FUN_EQ_TAC THEN
               			RW_TAC std_ss [o_DEF],
			RW_TAC std_ss [SPECIFICATION, GSYM prob_geometric_p_loop_def]]) THEN
		REWRITE_TAC [GSYM ADD1] THEN
		ASM_REWRITE_TAC [] THEN
		KNOW_TAC 
          	`prob bern
           	(\x.
              SND (FST (prob_while FST (prob_geometric_p_iter p) (T,SUC n) x)) =
              SUC n) =
           prob bern {}` THEN1 (
			MATCH_MP_TAC PROB_BERN_UNIVERSAL THEN
           		Q.EXISTS_TAC
              	`\s. (SND (FST (prob_geometric_p_loop p (T, SUC n) s)) = SUC n) = ~T` THEN
           		RW_TAC std_ss [PROB_GEOMETRIC_P_LOOP_STOPS, EVENTS_BERN_EMPTY,
                             NOT_IN_EMPTY] THENL [
           			RW_TAC std_ss [GSYM prob_geometric_p_loop_def] THEN
				SUFF_TAC 
                  		`(\x. SND (FST (prob_geometric_p_loop p (T,SUC n) x)) = SUC n) =
                   		(\x. SND x = SUC n) o FST o prob_geometric_p_loop p (T, SUC n)` THEN1 (
					RW_TAC std_ss [] THEN
					PROVE_TAC [INDEP_FN_FST_EVENTS,
                                 INDEP_FN_PROB_GEOMETRIC_P_LOOP]) THEN
               			FUN_EQ_TAC THEN
               			RW_TAC std_ss [o_DEF],
           		RW_TAC std_ss [SPECIFICATION, GSYM prob_geometric_p_loop_def]]) THEN
       	ASM_REWRITE_TAC [] THEN
       	RW_TAC real_ss [PROB_BERN_BASIC, POW_1],




   	RW_TAC bool_ss [PROB_BERN_GEOMETRIC_P_LOOP_LEMMA] THEN
   	RW_TAC bool_ss [prob_geometric_p_loop_def, BIND_DEF, o_DEF, UNCURRY,
                     UNIT_DEF] THEN
   	RW_TAC bool_ss [PROB_WHILE_ADVANCE, INDEP_FN_PROB_GEOMETRIC_P_ITER,
                     PROB_GEOMETRIC_P_LOOP_TERMINATES, prob_geometric_p_iter_def,
                     GSYM BIND_ASSOC, UNIT_DEF] THEN
   	KNOW_TAC `!n d. (SUC n = n + SUC (SUC d)) = F` THEN1
		DECIDE_TAC THEN
   	ASM_REWRITE_TAC [] THEN
   	RW_TAC real_ss [GSYM EMPTY_DEF, PROB_BERN_EMPTY,
                      GSYM prob_geometric_p_loop_def] THEN
   	ONCE_REWRITE_TAC [pow] THEN
   	KNOW_TAC `(\x.
         SND
           (FST
              (BIND (prob_bernoulli (1-p))
                 (\b. BIND (\s. ((b,SUC (SUC n)),s)) (prob_geometric_p_loop p)) x)) =
         n + SUC (SUC d)) =
         ((\x. SND x = SUC n + SUC d) o FST o
           BIND (prob_bernoulli (1-p)) (\b. BIND (UNIT (b,SUC (SUC n))) (prob_geometric_p_loop p)))` THEN1 (
   		FUN_EQ_TAC THEN
        	RW_TAC std_ss [BIND_DEF, UNIT_DEF, o_DEF, UNCURRY, ADD_CLAUSES]) THEN
   	ASM_REWRITE_TAC [] THEN
   	RW_TAC real_ss [real_div, POW_INV, POW_ONE] THEN
   	REAL_ARITH_TAC]);


val PROB_BERN_GEOMETRIC_P = store_thm
 ("PROB_BERN_GEOMETRIC_P",
 (--`!n p. 0 < p /\ p <= 1 ==>
	(prob bern {s | FST (prob_geometric_p p s) = SUC n} = p * (1 - p) pow n)`--),
	RW_TAC std_ss [prob_geometric_p_def, BIND_DEF, o_THM, UNCURRY, UNIT_DEF] THEN
	SUFF_TAC `prob bern {s | SND (FST (prob_geometric_p_loop p (T,1) s)) - 1 = SUC n} =
       	prob bern {s | SND (FST (prob_geometric_p_loop p (T,1) s)) = 1 + (SUC n)}` THEN1
		RW_TAC std_ss [GSYM ADD1, PROB_BERN_GEOMETRIC_P_LOOP] THEN

	MATCH_MP_TAC PROB_BERN_UNIVERSAL THEN
	Q.EXISTS_TAC `\s. 0 < SND (FST (prob_geometric_p_loop p (T,SUC 0) s))` THEN
	ONCE_REWRITE_TAC [ONE] THEN
	RW_TAC std_ss [PROB_GEOMETRIC_P_LOOP_T_RANGE2] THENL [
		SUFF_TAC `{s | SND (FST (prob_geometric_p_loop p (T,1) s)) - 1 = SUC n} =
     			(\x. SND x - 1 = SUC n) o FST o prob_geometric_p_loop p (T, 1)` THEN1 (
			RW_TAC std_ss [] THEN
        		PROVE_TAC [INDEP_FN_FST_EVENTS, INDEP_FN_PROB_GEOMETRIC_P_LOOP]) THEN
    		SET_EQ_TAC THEN
    		PSET_TAC [] THEN
    		RW_TAC std_ss [SPECIFICATION, o_DEF],

    	SUFF_TAC 
    	`{s | SND (FST (prob_geometric_p_loop p (T,1) s)) = 1 + (SUC n)} =
     	(\x. SND x = 1 + (SUC n)) o FST o prob_geometric_p_loop p (T, 1)` THEN1 (
		RW_TAC std_ss [] THEN
		PROVE_TAC [INDEP_FN_FST_EVENTS, INDEP_FN_PROB_GEOMETRIC_P_LOOP]) THEN
    	SET_EQ_TAC THEN
    	PSET_TAC [] THEN
    	RW_TAC std_ss [SPECIFICATION, o_DEF],

    	RW_TAC std_ss [GSPECIFICATION] THEN
    	DECIDE_TAC]);

val SUM_SUCN_POWN = store_thm
 ("SUM_SUCN_POWN",
 (--`!q n. sum (0,n) (\x. & (SUC x) * q pow x) = 
	 sum (0,n) (\x. sum (0,n) (\i. q pow i)) - 
	 sum (0,n) (\x. sum (0,x) (\i. q pow i))`--),
	GEN_TAC THEN
	Induct THEN1
		RW_TAC real_ss [sum] THEN

		REWRITE_TAC [sum] THEN
		KNOW_TAC `sum (0,n) (\x. sum (0,n) (\i. q pow i) + (\i. q pow i) (0 + n)) =
		sum (0,n) (\x. sum (0,n) (\i. q pow i)) + sum (0,n) (\x. q pow n)` THEN1 (
			KNOW_TAC `sum (0,n) 
			(\x. sum (0,n) (\i. q pow i) + (\i. q pow i) (0 + n)) = 
			sum (0,n) (\x. (\x. sum (0,n) (\i. q pow i)) x + 
				(\x. (\i. q pow i) (0 + n)) x)` THEN1
				RW_TAC std_ss [] THEN
			RW_TAC arith_ss [SUM_ADD]) THEN
	
		KNOW_TAC `sum (0,n) (\x. sum (0,n) (\i. q pow i) + (\i. q pow i) (0 + n)) +
    		    (sum (0,n) (\i. q pow i) + (\i. q pow i) (0 + n)) -
    			(sum (0,n) (\x. sum (0,x) (\i. q pow i)) +
     				(\x. sum (0,x) (\i. q pow i)) (0 + n)) =
		sum (0,n) (\x. sum (0,n) (\i. q pow i)) - 
		(sum (0,n) (\x. sum (0,x) (\i. q pow i))) + 
		sum (0,n) (\x. q pow n) + 
		(sum (0,n) (\i. q pow i) + (\i. q pow i) (0 + n)) - 
		(\x. sum (0,x) (\i. q pow i)) (0 + n)` THEN1 (
			ASM_REWRITE_TAC [] THEN
			REAL_ARITH_TAC) THEN
		ASM_REWRITE_TAC [] THEN
		NTAC 3 (POP_ASSUM K_TAC) THEN
		KNOW_TAC `sum (0,n) (\x. sum (0,n) (\i. q pow i)) -
    		sum (0,n) (\x. sum (0,x) (\i. q pow i)) + sum (0,n) (\x. q pow n) +
    		(sum (0,n) (\i. q pow i) + (\i. q pow i) (0 + n)) -
    		(\x. sum (0,x) (\i. q pow i)) (0 + n) = 
		(sum (0,n) (\x. sum (0,n) (\i. q pow i)) -
    		sum (0,n) (\x. sum (0,x) (\i. q pow i))) + (sum (0,n) (\x. q pow n) +
    		(sum (0,n) (\i. q pow i) + (\i. q pow i) (0 + n)) -
    		(\x. sum (0,x) (\i. q pow i)) (0 + n))` THEN1
			REAL_ARITH_TAC THEN
		ASM_REWRITE_TAC [REAL_EQ_LADD] THEN
		POP_ASSUM K_TAC THEN
		RW_TAC std_ss [REAL_EQ_SUB_LADD, REAL_ADD_ASSOC, ADD1, GSYM REAL_ADD] THEN
		RW_TAC real_ss [REAL_ADD_RDISTRIB] THEN
		KNOW_TAC `& n * q pow n + q pow n + sum (0,n) (\i. q pow i) = 
		& n * q pow n + sum (0,n) (\i. q pow i) + q pow n` THEN1
			REAL_ARITH_TAC THEN
		ASM_REWRITE_TAC [REAL_EQ_RADD] THEN
		POP_ASSUM K_TAC THEN
		Induct_on `n` THEN1
			RW_TAC real_ss [sum] THEN

			RW_TAC std_ss [sum, ADD1, GSYM REAL_ADD, REAL_ADD_RDISTRIB,
			pow, REAL_MUL_ASSOC, SUM_CMUL] THEN
			RW_TAC real_ss [REAL_EQ_RADD] THEN
			KNOW_TAC `q * sum (0,n) (\x. q pow n) = 
				q * (& n * q pow n)` THEN1
				FULL_SIMP_TAC std_ss [EQ_SYM_EQ] THEN
			ONCE_ASM_REWRITE_TAC [] THEN
			REAL_ARITH_TAC);

val NUM_NUM_POWN_ZERO = store_thm
 ("NUM_NUM_POWN_ZERO",
 (--`!x. 0 < x /\ x < 1 ==> ((\n. &n * (x pow n)) --> 0)`--),
	RW_TAC std_ss [] THEN
	MATCH_MP_TAC SER_ZERO THEN
	MATCH_MP_TAC SER_RATIO THEN
	RW_TAC std_ss [] THEN
	KNOW_TAC `!n. abs (&n * x pow n) = (&n * x pow n)` THEN1
		RW_TAC real_ss [ABS_REFL, POW_POS, REAL_LE_MUL, REAL_LT_IMP_LE] THEN
	ASM_REWRITE_TAC [] THEN
	MP_TAC (Q.SPECL [`x`, `1`] REAL_MEAN) THEN
	RW_TAC std_ss [] THEN
	EXISTS_TAC (--`z:real`--) THEN
	RW_TAC std_ss [pow, REAL_MUL_ASSOC, REAL_POW_LT, REAL_LE_RMUL] THEN
	RW_TAC std_ss [ADD1, GSYM REAL_ADD, REAL_RDISTRIB] THEN
	ONCE_REWRITE_TAC [REAL_ADD_SYM] THEN
	RW_TAC real_ss [GSYM REAL_LE_SUB_LADD] THEN
	KNOW_TAC `!n. z * & n - & n * x = &n * (z - x)` THEN1 (
		GEN_TAC THEN
		FULL_SIMP_TAC std_ss [REAL_SUB_LDISTRIB, REAL_MUL_SYM]) THEN
	ASM_REWRITE_TAC [] THEN
	RW_TAC std_ss [GSYM REAL_LE_LDIV_EQ, REAL_SUB_LT] THEN
	EXISTS_TAC (--`(ceiling (x / (z - x)) : num)`--) THEN
	RW_TAC std_ss [GREATER_EQ] THEN
	MATCH_MP_TAC REAL_LE_TRANS THEN
	EXISTS_TAC (--`&(ceiling (x / (z - x)):num)`--) THEN
	RW_TAC std_ss [LB_CEILING, REAL_LE]);


val SUCN_POWN_SUMS = store_thm
 ("SUCN_POWN_SUMS",
 (--`!q. 0 <= q /\ q < 1  ==> 
	((\x. &(SUC x) * q pow x) sums (1/((1-q) pow 2)))`--),
	RW_TAC std_ss [sums, SUM_SUCN_POWN, GP_FINITE, REAL_LT_IMP_NE,
		GSYM SUM_SUB, real_div, GSYM REAL_SUB_RDISTRIB] THEN
	MP_TAC NUM_NUM_POWN_ZERO THEN
 	STRIP_TAC THEN
	ONCE_REWRITE_TAC [REAL_MUL_SYM] THEN
	KNOW_TAC `!n. (\x. inv (q - 1) * (q pow n - 1 - (q pow x - 1))) =
		(\x. inv (q - 1) * (\x. q pow n - q pow x) x)` THEN1 (
		GEN_TAC THEN
		KNOW_TAC `!x. (q pow n - 1 - (q pow x - 1)) = (q pow n - q pow x)` THEN1 (
			GEN_TAC THEN
			REAL_ARITH_TAC) THEN
		RW_TAC std_ss []) THEN
	ASM_REWRITE_TAC [SUM_CMUL, real_sub] THEN
	KNOW_TAC `!n. (\x. q pow n + ~(q pow x)) = 
	(\x. (\x. q pow n) x + (\x. ~(q pow x)) x)` THEN1
		RW_TAC std_ss [] THEN
	ONCE_ASM_REWRITE_TAC [] THEN
	ONCE_REWRITE_TAC [SUM_ADD] THEN
	RW_TAC std_ss [REAL_ADD_LDISTRIB, GSYM real_sub, real_div, REAL_NEG_SUB,
			REAL_NEG_LMUL, GP_FINITE, REAL_LT_IMP_NE, SUM_NEG] THEN
	ONCE_REWRITE_TAC [REAL_MUL_SYM] THEN
	KNOW_TAC `(\n.
         sum (0,n) (\x. q pow n) * inv (q - 1) +
         (1 - q pow n) * inv (q - 1) * inv (q - 1)) = 
	(\n. (\n. sum (0,n) (\x. q pow n) * inv (q - 1)) n + 
	(\n. (1 - q pow n) * inv (q - 1) * inv (q - 1)) n)` THEN1
		RW_TAC std_ss [] THEN
	ASM_REWRITE_TAC [] THEN
	KNOW_TAC `(1 * inv ((1 - q) pow 2)) =
	(0 * inv (q - 1)) + (1-0) * (inv (q - 1) * inv (q - 1))` THEN1 (
		RW_TAC real_ss [] THEN
		ONCE_REWRITE_TAC [GSYM REAL_NEG_MUL2] THEN
		KNOW_TAC `~((q - 1) = 0)` THEN1 (
			REWRITE_TAC [REAL_SUB_0] THEN
			PROVE_TAC [REAL_NE_LT_GT]) THEN
		RW_TAC std_ss [GSYM POW_2, REAL_POW_INV, REAL_NEG_INV, REAL_NEG_SUB]) THEN
	ASM_REWRITE_TAC [] THEN
	POP_ASSUM K_TAC THEN
	SUFF_TAC `(\n. sum (0,n) (\x. q pow n) * inv (q - 1)) --> (0 * inv (q - 1))` THEN1 (
		SUFF_TAC `(\n. (1 - q pow n) * inv (q - 1) * inv (q - 1)) --> 
					(1-0) * (inv (q - 1) * inv (q - 1))` THEN1 (
			REPEAT STRIP_TAC THEN
			MATCH_MP_TAC SEQ_ADD THEN
			RW_TAC std_ss []) THEN
	
			KNOW_TAC `(\n. (1 - q pow n) * inv (q - 1) * inv (q - 1)) = 
			(\n. (\n. (1 - q pow n)) n * (\n. inv (q - 1) * inv (q - 1)) n)` THEN1 
				RW_TAC std_ss [REAL_MUL_ASSOC] THEN
			ONCE_ASM_REWRITE_TAC [] THEN
			SUFF_TAC `(\n. 1 - q pow n) --> (1-0)` THEN1 (
				SUFF_TAC `(\n. inv (q - 1) * inv (q - 1)) --> 
				inv (q - 1) * inv (q - 1)` THEN1 (
					REPEAT STRIP_TAC THEN
					MATCH_MP_TAC SEQ_MUL THEN
					RW_TAC std_ss []) THEN

					RW_TAC std_ss [SEQ_CONST]) THEN
				KNOW_TAC `(\n. 1 - q pow n) = 
				(\n. (\n. 1) n - (\n. q pow n) n)` THEN1
					RW_TAC std_ss [] THEN
				ONCE_ASM_REWRITE_TAC [] THEN
				MATCH_MP_TAC SEQ_SUB THEN
				RW_TAC std_ss [SEQ_CONST] THEN
				MATCH_MP_TAC SEQ_POWER THEN
				MATCH_MP_TAC REAL_LET_TRANS THEN
				EXISTS_TAC (--`q:real`--) THEN
				ASM_REWRITE_TAC [] THEN
				MATCH_MP_TAC REAL_EQ_IMP_LE THEN
				RW_TAC std_ss [ABS_REFL]) THEN
	
		KNOW_TAC `(\n. sum (0,n) (\x. q pow n) * inv (q - 1)) = 
		(\n. (\n. sum (0,n) (\x. q pow n)) n * 
			(\n. inv (q - 1)) n)` THEN1 
			RW_TAC std_ss [] THEN
		ONCE_ASM_REWRITE_TAC [] THEN
		MATCH_MP_TAC SEQ_MUL THEN
		RW_TAC real_ss [SEQ_CONST] THEN
		KNOW_TAC `!n. sum (0,n) (\x. q pow n) = &n * q pow n` THEN1 (
			REPEAT GEN_TAC THEN
			Induct_on `n` THEN1
				RW_TAC real_ss [sum] THEN

				RW_TAC std_ss [sum, ADD1, GSYM REAL_ADD, 
				REAL_ADD_RDISTRIB,
				pow, REAL_MUL_ASSOC, SUM_CMUL] THEN
				RW_TAC real_ss [REAL_EQ_RADD] THEN
				REAL_ARITH_TAC) THEN
		
		ASM_REWRITE_TAC [] THEN
		UNDISCH_TAC (--`0 <= (q:real)`--) THEN
		RW_TAC std_ss [REAL_LE_LT] THEN1
			FULL_SIMP_TAC std_ss [] THEN
		
			RW_TAC real_ss [SEQ] THEN
			EXISTS_TAC (--`SUC 0`--) THEN
			REPEAT STRIP_TAC THEN
			KNOW_TAC `abs (& n * 0 pow n) = & n * 0 pow n` THEN1 (
				REWRITE_TAC [ABS_REFL] THEN
				Induct_on `n` THEN1
					RW_TAC real_ss [REAL_LE_REFL] THEN
				
					RW_TAC real_ss [REAL_LE_REFL, POW_0]) THEN
			ASM_REWRITE_TAC [] THEN
			Induct_on `n` THEN1
				RW_TAC real_ss [] THEN
			
				RW_TAC real_ss [POW_0]);

val EXPEC_GEOMETRIC_SUMMABLE = store_thm
 ("EXPEC_GEOMETRIC_SUMMABLE",
 (--`!p. 0 < p /\ p <= 1   ==> 
		summable (\x. & x * prob bern {s | FST (prob_geometric_p p s) = x})`--),
	RW_TAC std_ss [sums, summable] THEN
	MP_TAC NUM_NUM_POWN_ZERO THEN
 	STRIP_TAC THEN
	EXISTS_TAC (--`1/p`--) THEN
	KNOW_TAC `!n. sum (0,n) 
	(\x. & x * prob bern {s | FST (prob_geometric_p p s) = x}) =
	sum (0,n-1) (\x. &(SUC x) * prob bern {s | FST (prob_geometric_p p s) = (SUC x)})` THEN1 (
		GEN_TAC THEN
		ONCE_REWRITE_TAC [ADD1] THEN
		KNOW_TAC `(\x. & (x + 1) * 
		prob bern {s | FST (prob_geometric_p p s) = x + 1}) = 
		(\x. (\x. & x * prob bern {s | FST (prob_geometric_p p s) = x}) (x + 1))` THEN1
			RW_TAC std_ss [] THEN
		ASM_REWRITE_TAC [SUM_OFFSET] THEN
		RW_TAC real_ss [SUM_1] THEN
		Induct_on `n` THEN1
			RW_TAC real_ss [sum, SUM_1] THEN

			KNOW_TAC `SUC n - 1 + 1 = SUC n` THEN1
				DECIDE_TAC THEN
			RW_TAC std_ss []) THEN
	ASM_REWRITE_TAC [] THEN
	KNOW_TAC `!s. ((\n. sum (0,n - 1) (\x. & (SUC x) *
               prob bern {s | FST (prob_geometric_p p s) = SUC x})) --> s) = 
		((\n. (\n. sum (0,n - 1) (\x. & (SUC x) *
               prob bern {s | FST (prob_geometric_p p s) = SUC x})) (SUC n)) --> s)` THEN1
		RW_TAC std_ss [GSYM SEQ_SUC] THEN
	RW_TAC arith_ss [GSYM sums, PROB_BERN_GEOMETRIC_P] THEN
	ONCE_REWRITE_TAC [REAL_MUL_SYM] THEN
	SUFF_TAC `(\x. &(SUC x) * (1 - p) pow x) sums (1/((1-(1-p))pow 2))` THEN1 (
		KNOW_TAC `(1-(1-p)) = p` THEN1
			REAL_ARITH_TAC THEN
		ASM_REWRITE_TAC [] THEN
		KNOW_TAC `!x. p * (1 - p) pow x * & (SUC x) =
		p * & (SUC x) * (1 - p) pow x` THEN1 (
			GEN_TAC THEN
			REAL_ARITH_TAC) THEN
		ASM_REWRITE_TAC [] THEN
		POP_ASSUM K_TAC THEN
		KNOW_TAC `(\x. p * & (SUC x) * (1 - p) pow x) = 
			(\x. p * (\x.& (SUC x) * (1 - p) pow x) x)` THEN1
				RW_TAC std_ss [REAL_MUL_ASSOC] THEN
		ASM_REWRITE_TAC [] THEN
		RW_TAC std_ss [sums, SUM_CMUL] THEN
		KNOW_TAC `(\n. p) --> p` THEN1
			RW_TAC std_ss [SEQ_CONST] THEN
		KNOW_TAC `(\n. p * sum (0,n) (\x. & (SUC x) * (1 - p) pow x)) =
		(\n. (\n. p) n * (\n. sum (0,n) (\x. & (SUC x) * (1 - p) pow x)) n)` THEN1
			RW_TAC std_ss [] THEN
		KNOW_TAC `(1 / p) = p * (1 / p pow 2)` THEN1
			RW_TAC real_ss [real_div, POW_2, GSYM REAL_POW_INV,
				REAL_MUL_ASSOC, REAL_MUL_RINV, REAL_POS_NZ] THEN
		ONCE_ASM_REWRITE_TAC [] THEN
		MATCH_MP_TAC SEQ_MUL THEN
		RW_TAC std_ss []) THEN

	
		MATCH_MP_TAC SUCN_POWN_SUMS THEN
		RW_TAC std_ss [REAL_SUB_LE, REAL_LT_SUB_RADD,
					REAL_LT_ADDR]);

val EXPEC_GEOMETRIC = store_thm
 ("EXPEC_GEOMETRIC",
 (--`!p. 0 < p /\ p <= 1  ==> 
			(expec (\s. prob_geometric_p p s) = 1/p)`--),
	RW_TAC std_ss [expec_def, suminf, sums] THEN
	MP_TAC NUM_NUM_POWN_ZERO THEN
 	STRIP_TAC THEN
	KNOW_TAC `!n. sum (0,n) 
	(\x. & x * prob bern {s | FST (prob_geometric_p p s) = x}) =
	sum (0,n-1) (\x. &(SUC x) * prob bern {s | FST (prob_geometric_p p s) = (SUC x)})` THEN1 (
		GEN_TAC THEN
		ONCE_REWRITE_TAC [ADD1] THEN
		KNOW_TAC `(\x. & (x + 1) * 
		prob bern {s | FST (prob_geometric_p p s) = x + 1}) = 
		(\x. (\x. & x * prob bern {s | FST (prob_geometric_p p s) = x}) (x + 1))` THEN1
			RW_TAC std_ss [] THEN
		ASM_REWRITE_TAC [SUM_OFFSET] THEN
		RW_TAC real_ss [SUM_1] THEN
		Induct_on `n` THEN1
			RW_TAC real_ss [sum, SUM_1] THEN

			KNOW_TAC `SUC n - 1 + 1 = SUC n` THEN1
				DECIDE_TAC THEN
			RW_TAC std_ss []) THEN
	ASM_REWRITE_TAC [] THEN
	KNOW_TAC `!s. ((\n. sum (0,n - 1) (\x. & (SUC x) *
               prob bern {s | FST (prob_geometric_p p s) = SUC x})) --> s) = 
		((\n. (\n. sum (0,n - 1) (\x. & (SUC x) *
               prob bern {s | FST (prob_geometric_p p s) = SUC x})) (SUC n)) --> s)` THEN1
		RW_TAC std_ss [GSYM SEQ_SUC] THEN
	RW_TAC arith_ss [GSYM sums, GSYM suminf, PROB_BERN_GEOMETRIC_P] THEN
	ONCE_REWRITE_TAC [REAL_MUL_SYM] THEN
	SUFF_TAC `suminf (\x. &(SUC x) * (1 - p) pow x) = 1/((1-(1-p))pow 2)` THEN1 (
		KNOW_TAC `(1-(1-p)) = p` THEN1
			REAL_ARITH_TAC THEN
		ASM_REWRITE_TAC [] THEN
		KNOW_TAC `!x. p * (1 - p) pow x * & (SUC x) =
		p * & (SUC x) * (1 - p) pow x` THEN1 (
			GEN_TAC THEN
			REAL_ARITH_TAC) THEN
		ASM_REWRITE_TAC [] THEN
		POP_ASSUM K_TAC THEN
		KNOW_TAC `(\x. p * & (SUC x) * (1 - p) pow x) = 
			(\x. p * (\x.& (SUC x) * (1 - p) pow x) x)` THEN1
				RW_TAC std_ss [REAL_MUL_ASSOC] THEN
		ASM_REWRITE_TAC [] THEN
		STRIP_TAC THEN
		KNOW_TAC `suminf (\x. p * (\x. & (SUC x) * (1 - p) pow x) x) = 
		p * suminf (\x.& (SUC x) * (1 - p) pow x)` THEN1 (
			MATCH_MP_TAC SUMINF_CMUL THEN
			MATCH_MP_TAC SUM_SUMMABLE THEN
			EXISTS_TAC (--`1 / (p pow 2)`--) THEN
			POP_ASSUM MP_TAC THEN
			REWRITE_TAC [suminf] THEN
			SELECT_ELIM_TAC THEN
			KNOW_TAC `0 <= (1 - p) /\ (1 - p) < 1` THEN1
				RW_TAC std_ss [REAL_SUB_LE, REAL_LT_SUB_RADD,
					REAL_LT_ADDR] THEN
			RW_TAC std_ss [] THENL [
				EXISTS_TAC (--`(1 / (1 - (1-p)) pow 2)`--) THEN
				MATCH_MP_TAC SUCN_POWN_SUMS THEN
		        	RW_TAC std_ss [],
			
				KNOW_TAC ` (\x. & (SUC x) * (1 - p) pow x) 
				sums (1 / (1 - (1 - p)) pow 2)` THEN1 (
					MATCH_MP_TAC SUCN_POWN_SUMS THEN
		        		RW_TAC std_ss []) THEN
				RW_TAC std_ss []]) THEN

		RW_TAC real_ss [real_div, POW_2, GSYM REAL_POW_INV,
			REAL_MUL_ASSOC, REAL_MUL_RINV, REAL_POS_NZ]) THEN
	
		ONCE_REWRITE_TAC [EQ_SYM_EQ] THEN
		MATCH_MP_TAC SUM_UNIQ THEN
		MATCH_MP_TAC SUCN_POWN_SUMS THEN
		RW_TAC std_ss [REAL_SUB_LE, REAL_LT_SUB_RADD,
					REAL_LT_ADDR]);


(*=====================================*)
(*two random variables*)
(*=====================================*)

val PROB_SND_RV_SEQ = store_thm
 ("PROB_SND_RV_SEQ",
 (--`!(Y: (num -> bool) -> num#(num -> bool)) (X: (num -> bool) -> num#(num -> bool)) m.
	 (X IN indep_fn) /\ (Y IN indep_fn) ==>
	 (prob bern {s | FST (Y (SND (X s))) = m} = prob bern {s | FST (Y s) = m})`--),
	RW_TAC std_ss [] THEN
	KNOW_TAC `{s | FST (Y s) = m} = ((\y. y = m) o FST o Y)` THEN1
		PSET_TAC [o_DEF] THEN
	KNOW_TAC `{s | FST (Y (SND (X s))) = m} = (((\y. y = m) o FST o Y) o SND o X)` THEN1
		PSET_TAC [o_DEF] THEN
	ASM_REWRITE_TAC [] THEN
	MP_TAC (Q.ISPEC `(X: (num -> bool) -> num#(num -> bool))` INDEP_FN_PROB_PRESERVING) THEN
	RW_TAC std_ss [PROB_PRESERVING, GSPECIFICATION, PREIMAGE_ALT] THEN
	POP_ASSUM (MP_TAC o Q.ISPEC `
		((\y. y = m) o FST o (Y: (num -> bool) -> num#(num -> bool)))`) THEN
	KNOW_TAC `((\y. y = m) o FST o Y) IN events bern` THEN1
		RW_TAC std_ss [INDEP_FN_FST_EVENTS] THEN
	RW_TAC std_ss []);

val RV_EQ_SND_RV_SEQ_EVENTS_BERN = store_thm
 ("RV_EQ_SND_RV_SEQ_EVENTS_BERN",
 (--`!(Y: (num -> bool) -> num#(num -> bool)) (X: (num -> bool) -> num#(num -> bool)) m.
	 (X IN indep_fn) /\ (Y IN indep_fn) ==>
	{s | FST (Y (SND (X s))) = m} IN events bern`--),
	RW_TAC std_ss [] THEN
	KNOW_TAC `{s | FST (Y (SND (X s))) = m} = ((\y. y = m) o FST o Y) o SND o X` THEN1
		PSET_TAC [o_DEF] THEN
	ASM_REWRITE_TAC [] THEN
	MATCH_MP_TAC INDEP_FN_SND_EVENTS THEN
	RW_TAC std_ss [] THEN
	MATCH_MP_TAC INDEP_FN_FST_EVENTS THEN
	RW_TAC std_ss []);

val BIGUNION_INDUCT_STEP = store_thm
  ("BIGUNION_INDUCT_STEP",
  (--`!(A: num -> ('a -> bool)) (k:num).
   (BIGUNION (IMAGE A {n | (n <= SUC k)}) =
    (A (SUC k)) UNION (BIGUNION (IMAGE A {n | (n <= k)})))`--),
        RW_TAC std_ss [BIGUNION, IMAGE_DEF] THEN
        RW_TAC std_ss [PROVE [LE]
                (--`n <= SUC k = (n <= k) \/ (n = SUC k)`--)] THEN
        CONV_TAC (ONCE_DEPTH_CONV SET_SPEC_CONV)  THEN
        CONV_TAC (ONCE_DEPTH_CONV SET_SPEC_CONV)  THEN
        RW_TAC std_ss [LEFT_AND_OVER_OR]  THEN
	RW_TAC std_ss [EXISTS_OR_THM, RIGHT_AND_OVER_OR, UNION_DEF, GSPECIFICATION] THEN
	REWRITE_TAC [EXTENSION] THEN
	PSET_TAC [GSPECIFICATION] THEN
	PROVE_TAC [DISJ_SYM]);

val SUM_RV_BIGUNION = store_thm
 ("SUM_RV_BIGUNION",
 (--`!(X: (num -> bool) -> num#(num->bool)) (Y: (num -> bool) -> num#(num->bool)) k. 
	{s | FST (X s) + FST (Y (SND (X s))) = k} = 
	 BIGUNION (IMAGE (\i. {s | (FST (X s) = i) /\ (FST (Y (SND (X s))) = k - i)}) {i | i <= k})`--),
	RW_TAC std_ss [] THEN
	Induct_on `k` THEN1 (
		RW_TAC arith_ss [] THEN
        	REWRITE_TAC [BIGUNION, IMAGE_DEF] THEN
        	CONV_TAC (ONCE_DEPTH_CONV SET_SPEC_CONV) THEN
        	CONV_TAC (ONCE_DEPTH_CONV SET_SPEC_CONV) THEN
        	FULL_SIMP_TAC std_ss [] THEN
        	REWRITE_TAC [GSPEC_ID]) THEN 

		RW_TAC std_ss [ BIGUNION_INDUCT_STEP] THEN
		FULL_SIMP_TAC std_ss [EXTENSION] THEN
		GEN_TAC THEN
		POP_ASSUM (MP_TAC o Q.SPEC `x`) THEN
		PSET_TAC [GSPECIFICATION, IN_BIGUNION_IMAGE] THEN
		FULL_SIMP_TAC arith_ss []);

val EVENT_EQ_INTER_BERN = store_thm
 ("EVENT_EQ_INTER_BERN",
 (--`! (X: (num -> bool) -> num#(num -> bool)) (Y: (num -> bool) -> num#(num -> bool)).
	(X IN indep_fn) /\ (Y IN indep_fn) ==>
	(!m n. {s | (FST (X s) = m) /\ (FST (Y (SND (X s))) = n)} IN events bern)`--),
	RW_TAC std_ss [] THEN
	KNOW_TAC `{s | (FST (X s) = m) /\ (FST (Y (SND (X s))) = n)} =
           {s | ( FST (X s) = m)} INTER {s | (FST (Y (SND (X s))) = n)} ` THEN1 (
    		REWRITE_TAC [EXTENSION] THEN
    		PSET_TAC [IN_INTER]) THEN
	ASM_REWRITE_TAC [] THEN
	MATCH_MP_TAC EVENTS_INTER THEN
	RW_TAC std_ss [PROB_SPACE_BERN, RV_EQ_SND_RV_SEQ_EVENTS_BERN] THEN
	KNOW_TAC `{s | FST (X s) = m} = ((\y. y = m) o FST o X)` THEN1
		PSET_TAC [o_DEF] THEN
	RW_TAC std_ss [INDEP_FN_FST_EVENTS]);

val EVENT_EQ_BIGUNION_BERN = store_thm
 ("EVENT_EQ_BIGUNION_BERN",
 (--`! (X: (num -> bool) -> num#(num -> bool)) (Y: (num -> bool) -> num#(num -> bool)).
        (X IN indep_fn) /\ (Y IN indep_fn) ==>
        (!m n. BIGUNION (IMAGE (\i. {s | (FST (X s) = i) /\ (FST (Y (SND (X s))) = m - i)})
                                {i | i <= n}) IN events bern)`--),
	RW_TAC std_ss [] THEN
	MATCH_MP_TAC EVENTS_COUNTABLE_UNION THEN
	RW_TAC std_ss [PROB_SPACE_BERN] THENL [
		RW_TAC std_ss [IMAGE_DEF, SUBSET_DEF, GSPECIFICATION] THEN
		RW_TAC std_ss [EVENT_EQ_INTER_BERN],

		MATCH_MP_TAC COUNTABLE_IMAGE THEN
		RW_TAC std_ss [countable_def, IN_IMAGE, GSPECIFICATION] THEN
		EXISTS_TAC (--`(\x. (x:num))`--) THEN
		RW_TAC std_ss []]);

val PROB_ADDITIVE_ON_SUM_RV = store_thm
 ("PROB_ADDITIVE_ON_SUM_RV",
 (--`!(k: num) (X: (num -> bool) -> num#(num -> bool)) (Y: (num -> bool) -> num#(num -> bool)).
        (X IN indep_fn) /\ (Y IN indep_fn) ==>
	(prob bern
           (BIGUNION
              (IMAGE (\i. {s | (FST (X s) = i) /\ (FST (Y (SND (X s))) = k - i)}) {i | i <= k})) =
	      sum (0,SUC k) (\i. prob bern {s | (FST (X s) = i) /\ (FST (Y (SND (X s))) = (k - i))}))`--),
	RW_TAC std_ss [] THEN
	KNOW_TAC `!(k:num) (X: (num -> bool) -> num#(num -> bool)) 
	 (Y: (num -> bool) -> num#(num -> bool)). 
	 sum (0, SUC k) (\i. prob bern {s | (FST (X s) = i) /\ (FST (Y (SND (X s))) = (k - i))})  =
                  sum (0, SUC k) 
		(prob bern o (\i. {s | (FST (X s) = i) /\ (FST (Y (SND (X s))) = (k - i))}))` THEN1 
       		RW_TAC std_ss [o_DEF] THEN
	ONCE_REWRITE_TAC [EQ_SYM_EQ] THEN
	ONCE_ASM_REWRITE_TAC [] THEN
	MATCH_MP_TAC PROB_FINITELY_ADDITIVE THEN
	RW_TAC std_ss [PROB_SPACE_BERN, DISJOINT_DEF, INTER_DEF, SPECIFICATION, IN_FUNSET] THENL [
		    	ONCE_REWRITE_TAC [GSYM SPECIFICATION] THEN
		    	RW_TAC std_ss [EVENT_EQ_INTER_BERN],
			
                    	REWRITE_TAC [EXTENSION] THEN
		    	PSET_TAC [GSPECIFICATION] THEN
                    	DISJ2_TAC THEN
		    	RW_TAC arith_ss [],
                    

                    	RW_TAC arith_ss [EXTENSION, IN_UNIV, IN_BIGUNION_IMAGE,
                                      GSPECIFICATION, IN_COUNT] THEN
		    	ONCE_REWRITE_TAC [GSYM NOT_SUC_LESS_EQ] THEN
		    	REWRITE_TAC [NOT_LESS_EQUAL]]);

val SUM_PROB_INTER_BIGUNION = store_thm
 ("SUM_PROB_INTER_BIGUNION",
 (--`! (X: (num -> bool) -> num#(num -> bool)) (Y: (num -> bool) -> num#(num -> bool)) i n.
        (X IN indep_fn) /\ (Y IN indep_fn) ==>
        (sum (0,n) (\j. prob bern {s | (FST (X s) = i) /\ (FST (Y (SND (X s))) = j)}) =
		prob bern (BIGUNION 
			(IMAGE (\j. {s | (FST (X s) = i) /\ (FST (Y (SND (X s))) = j)}) {j | j < n})))`--),
	RW_TAC std_ss [] THEN
	KNOW_TAC `!i (X: (num -> bool) -> num#(num -> bool)) (Y: (num -> bool) -> num#(num -> bool)).
	sum (0,n) (\j. prob bern {s | (FST (X s) = i) /\ (FST (Y (SND (X s))) = j)})  =
        sum (0,n) (prob bern o (\j. {s | (FST (X s) = i) /\ (FST (Y (SND (X s))) = j)}))` THEN1
       		RW_TAC std_ss [o_DEF] THEN
	ONCE_ASM_REWRITE_TAC [] THEN
	MATCH_MP_TAC PROB_FINITELY_ADDITIVE THEN
	RW_TAC std_ss [PROB_SPACE_BERN, DISJOINT_DEF, INTER_DEF, SPECIFICATION, IN_FUNSET] THENL [
		    	ONCE_REWRITE_TAC [GSYM SPECIFICATION] THEN
		    	RW_TAC std_ss [EVENT_EQ_INTER_BERN],
			
                    	REWRITE_TAC [EXTENSION] THEN
		    	PSET_TAC [GSPECIFICATION] THEN
                    	DISJ2_TAC THEN
		    	RW_TAC arith_ss [],
                    

                    	RW_TAC arith_ss [EXTENSION, IN_UNIV, IN_BIGUNION_IMAGE,
                                      GSPECIFICATION, IN_COUNT]]);

val SUM_PROB_INTER_BIGUNION2 = store_thm
 ("SUM_PROB_INTER_BIGUNION2",
 (--`! (X: (num -> bool) -> num#(num -> bool)) (Y: (num -> bool) -> num#(num -> bool)) i n.
        (X IN indep_fn) /\ (Y IN indep_fn) ==>
        (sum (0,n) (\j. prob bern {s | (FST (X s) = j) /\ (FST (Y (SND (X s))) = i)}) =
		prob bern (BIGUNION 
			(IMAGE (\j. {s | (FST (X s) = j) /\ (FST (Y (SND (X s))) = i)}) {j | j < n})))`--),
	RW_TAC std_ss [] THEN
	KNOW_TAC `!i (X: (num -> bool) -> num#(num -> bool)) (Y: (num -> bool) -> num#(num -> bool)).
		sum (0,n) (\j. prob bern {s | (FST (X s) = j) /\ (FST (Y (SND (X s))) = i)})  =
        	sum (0,n) (prob bern o (\j. {s | (FST (X s) = j) /\ (FST (Y (SND (X s))) = i)}))` THEN1
       		RW_TAC std_ss [o_DEF] THEN
	ONCE_ASM_REWRITE_TAC [] THEN
	MATCH_MP_TAC PROB_FINITELY_ADDITIVE THEN
	RW_TAC std_ss [PROB_SPACE_BERN, DISJOINT_DEF, INTER_DEF, SPECIFICATION, IN_FUNSET] THENL [
		    	ONCE_REWRITE_TAC [GSYM SPECIFICATION] THEN
		    	RW_TAC std_ss [EVENT_EQ_INTER_BERN],
			
                    	REWRITE_TAC [EXTENSION] THEN
		    	PSET_TAC [GSPECIFICATION] THEN
                    	DISJ2_TAC THEN
		    	RW_TAC arith_ss [],
                    

                    	RW_TAC arith_ss [EXTENSION, IN_UNIV, IN_BIGUNION_IMAGE,
                                      GSPECIFICATION, IN_COUNT]]);

val SUM_REL_TWO_RV_EXPEC = store_thm
 ("SUM_REL_TWO_RV_EXPEC",
 (--`!M n. sum (0,n) (\x. sum (0,SUC x) (\i. & x * M i (x - i))) =
        sum (0,n) (\x. sum (0,n - x) (\i. & (x + i) * M x i))`--),
	REPEAT GEN_TAC THEN
	Induct_on `n` THEN1
        	RW_TAC real_ss [sum] THEN

        	RW_TAC real_ss [sum] THEN
        	KNOW_TAC `!n. SUC n - n = 1` THEN1 (
                	GEN_TAC THEN
                	DECIDE_TAC) THEN
        	RW_TAC arith_ss [SUM_1, REAL_EQ_RADD, REAL_ADD_ASSOC] THEN
        	KNOW_TAC `sum (0,n) (\x. sum (0,n - x) (\i. & (x + i) * M x i)) =
                	sum (0,n) (\x. sum (0,SUC x) (\i. & x * M i (x - i)))` THEN1
                	PROVE_TAC [EQ_SYM_EQ] THEN
        	FULL_SIMP_TAC std_ss [ADD_SYM, GSYM SUM_ADD] THEN
        	MATCH_MP_TAC SUM_EQ THEN
        	RW_TAC arith_ss [SUB, sum]);

val SUM_REL_TWO_RV_EXPEC2 = store_thm
 ("SUM_REL_TWO_RV_EXPEC2",
 (--`!M n. sum (0,n) (\x. & x * sum (0,SUC x) (\i. M i (x - i))) =
        sum (0,n) (\x. & x * sum (0,n - x) (\i. M x i + M i x))`--),
	REPEAT GEN_TAC THEN
	Induct_on `n` THEN1
        	RW_TAC real_ss [sum] THEN

        	ONCE_REWRITE_TAC [sum] THEN
		RW_TAC arith_ss [] THEN
		KNOW_TAC `!n. SUC n - n = 1` THEN1 (
                	GEN_TAC THEN
                	DECIDE_TAC) THEN
        	RW_TAC arith_ss [SUM_1] THEN
		KNOW_TAC `sum (0,n) (\x. & x * sum (0,SUC n - x) (\i. M x i + M i x)) =
			sum (0,n) (\x. & x * (sum (0, n - x) (\i. M x i + M i x) +
			(M x (n - x) + M (n - x) x)))` THEN1 (
			MATCH_MP_TAC SUM_EQ THEN
			RW_TAC arith_ss [REAL_EQ_MUL_LCANCEL] THEN
			Cases_on `r = 0` THEN1
				RW_TAC real_ss [REAL_INJ] THEN

				RW_TAC real_ss [REAL_INJ] THEN
				KNOW_TAC `SUC n - r = SUC (n - r)` THEN1
					RW_TAC arith_ss [] THEN
				ASM_REWRITE_TAC [] THEN
				RW_TAC arith_ss [sum]) THEN
		ASM_REWRITE_TAC [REAL_ADD_LDISTRIB] THEN
		KNOW_TAC `(\x. & x * sum (0,n - x) (\i. M x i + M i x) + 
		(&x * M x (n - x) + & x * M (n - x) x)) =
		(\x. (\x. & x * sum (0,n - x) (\i. M x i + M i x)) x + 
		(\x. (&x * M x (n - x) + & x * M (n - x) x)) x)` THEN1
			RW_TAC std_ss [] THEN
		ASM_REWRITE_TAC [SUM_ADD] THEN
		POP_ASSUM K_TAC THEN
		RW_TAC std_ss [GSYM REAL_ADD_ASSOC, REAL_EQ_LADD] THEN
		KNOW_TAC `sum (0,n) (\x. & x * M x (n - x) + & x * M (n - x) x) +
		(& n * M n 0 + & n * M 0 n) = 
		sum (0, SUC n) (\x. & x * M x (n - x) + & x * M (n - x) x)` THEN1
			RW_TAC arith_ss [sum] THEN
		ASM_REWRITE_TAC [] THEN
		KILL_ALL_TAC THEN
		KNOW_TAC `(\x. & x * M x (n - x) + & x * M (n - x) x) =
                        (\x. (\x. & x * M x (n - x)) x +
                        (\x. & x * M (n - x) x) x)` THEN1
                	RW_TAC std_ss [] THEN
        	ASM_REWRITE_TAC [SUM_ADD] THEN
        	POP_ASSUM K_TAC THEN
		KNOW_TAC `sum (0, SUC n) (\x. (\x. & x * M (n - x) x) ((\x. n - x) x)) =
                	sum (0, SUC n) (\x. & x * M (n - x) x)` THEN1 (
                	MATCH_MP_TAC SUM_PERMUTE_0 THEN
                	RW_TAC std_ss [EXISTS_UNIQUE_THM] THENL [
				EXISTS_TAC (--`n - y:num`--) THEN
				RW_TAC arith_ss [],

				RW_TAC arith_ss []]) THEN
		FULL_SIMP_TAC std_ss [EQ_SYM_EQ] THEN
		POP_ASSUM K_TAC THEN
		KNOW_TAC `sum (0,SUC n) (\x. & (n - x) * M (n - (n - x)) (n - x)) =
        	sum (0,SUC n) (\x. & (n - x) * M x (n - x)) ` THEN1 (
			MATCH_MP_TAC SUM_EQ THEN
			RW_TAC arith_ss []) THEN
		ASM_REWRITE_TAC [GSYM SUM_ADD, GSYM SUM_CMUL] THEN
		POP_ASSUM K_TAC THEN
		MATCH_MP_TAC SUM_EQ THEN
		RW_TAC arith_ss [] THEN
		RW_TAC arith_ss [GSYM REAL_ADD_RDISTRIB, REAL_ADD]);

val SEQ_DOUBLE = store_thm
 ("SEQ_DOUBLE",
 (--`!(f: num -> real) (l:real) . f --> l ==> ((\n. f (2*n)) --> l)`--),
	REPEAT GEN_TAC THEN REWRITE_TAC[SEQ, GREATER_EQ] THEN
	DISCH_THEN(fn th => X_GEN_TAC (--`e:real`--) THEN
    	DISCH_THEN(MP_TAC o MATCH_MP th)) THEN
	BETA_TAC THEN
	DISCH_THEN(X_CHOOSE_TAC (--`N:num`--)) THEN
	EXISTS_TAC (--`N:num`--)  THEN
	X_GEN_TAC (--`n:num`--) THEN DISCH_TAC THEN
	FIRST_ASSUM MATCH_MP_TAC  THEN
	MATCH_MP_TAC LESS_EQ_TRANS THEN
	EXISTS_TAC (--`n:num`--) THEN
	RW_TAC arith_ss []);

val TWO_RV_EXPEC_HELPER = store_thm
 ("TWO_RV_EXPEC_HELPER",
 (--`!(X: (num -> bool) -> num#(num -> bool)) (Y: (num -> bool) -> num#(num -> bool)) m.
	summable (\x. sum (0,SUC x) (\i. & x * prob bern
                   {s | (FST (X s) = i) /\ (FST (Y (SND (X s))) = x - i)})) /\
    ((\n. sum (0,n) (\a. sum (0,n) (\b. & (a + b) * 
                 prob bern {s | (FST (X s) = a) /\ (FST (Y (SND (X s))) = b)}))) --> m) /\
	(X IN indep_fn) /\ (Y IN indep_fn) ==>
	((\n. sum (0,n) (\x. sum (0,SUC x) (\i. & x * prob bern
                   {s | (FST (X s) = i) /\ (FST (Y (SND (X s))) = x - i)}))) --> m)`--),
	RW_TAC std_ss [summable, sums] THEN
	KNOW_TAC `(!m. {s | (FST (X s) = m)} IN events bern)` THEN1 (
		GEN_TAC THEN
		KNOW_TAC `{s | FST (X s) = m'} = ((\y. y = m') o FST o X)` THEN1
			PSET_TAC [o_DEF] THEN
		RW_TAC std_ss [INDEP_FN_FST_EVENTS]) THEN		
	KNOW_TAC `s <= m` THEN1 (
		MATCH_MP_TAC SEQ_LE THEN
		EXISTS_TAC (--`(\n. sum (0,n) (\x. sum (0,SUC x) (\i. & x * prob bern
        	{s | (FST ((X: (num -> bool) -> num#(num -> bool)) s) = i) /\
        	( FST ((Y: (num -> bool) -> num#(num -> bool)) (SND (X s))) = x - i)})))`--) THEN
		EXISTS_TAC (--`(\n. sum (0,n) (\a. sum (0,n) (\b. & (a + b) *
			prob bern {s | ( FST (
			(X: (num -> bool) -> num#(num -> bool)) s) = a) /\
			( FST ((Y: (num -> bool) -> num#(num -> bool))(SND (X s))) = b)})))`--) THEN
		RW_TAC std_ss [] THEN
		EXISTS_TAC (--`0:num`--) THEN
		RW_TAC arith_ss [GREATER_EQ] THEN
		KNOW_TAC `sum (0,n) (\x. sum (0,SUC x) (\i. & x * prob bern
			{s | (FST (X s) = i) /\ (FST (Y (SND (X s))) = x - i)})) =
			sum (0,n) (\x. sum (0,SUC x) (\i. & x * (\l m. prob bern
		 	{s | (FST (X s) = l) /\ (FST (Y (SND (X s))) = m)}) i (x-i)))` THEN1
			RW_TAC std_ss [] THEN
		ASM_REWRITE_TAC [] THEN
		POP_ASSUM K_TAC THEN
		RW_TAC std_ss [SUM_REL_TWO_RV_EXPEC] THEN
		KNOW_TAC `!x. sum (0,n - x) (\i. & (x + i) *
        		prob bern {s | (FST (X s) = x) /\ (FST (Y (SND (X s))) = i)}) =
        		sum (0,(n-x) + x) (\i. & (x + i) * prob bern {s | (FST (X s) = x) /\ (FST (Y (SND (X s))) = i)}) -
        		sum (n - x, x) (\i. & (x + i) *
                	prob bern {s | (FST (X s) = x) /\ (FST (Y (SND (X s))) = i)})` THEN1
			RW_TAC std_ss [REAL_EQ_SUB_LADD, SUM_TWO] THEN
		ASM_REWRITE_TAC [] THEN
		POP_ASSUM K_TAC THEN
		KNOW_TAC `!n. (\x. sum (0,n - x + x) (\i. & (x + i) * 
		prob bern {s | (FST (X s) = x) /\ (FST (Y (SND (X s))) = i)}) - sum (n - x,x)
		(\i. & (x + i) * 
		prob bern {s | (FST (X s) = x) /\ (FST (Y (SND (X s))) = i)})) = 
		(\x. (\x. sum (0,n - x + x) (\i. & (x + i) *
                prob bern {s | (FST (X s) = x) /\ (FST (Y (SND (X s))) = i)})) x - (\x. sum (n - x,x)
		(\i. & (x + i) *
		prob bern {s | (FST (X s) = x) /\ (FST (Y (SND (X s))) = i)})) x)` THEN1
			RW_TAC std_ss [] THEN
		RW_TAC std_ss [SUM_SUB, REAL_LE_SUB_RADD] THEN
		KNOW_TAC `sum (0,n) (\x. sum (0, (n-x) + x) (\i. & (x + i) * 
		prob bern {s | (FST (X s) = x) /\ (FST (Y (SND (X s))) = i)})) = 
			sum(0,n) (\x. sum (0,n) (\i. & (x + i) *
		prob bern {s | (FST (X s) = x) /\ (FST (Y (SND (X s))) = i)}))` THEN1 (
			MATCH_MP_TAC SUM_EQ THEN
			RW_TAC arith_ss []) THEN
		ASM_REWRITE_TAC [REAL_LE_ADDR] THEN
		MATCH_MP_TAC SUM_POS THEN
		RW_TAC std_ss [] THEN
		MATCH_MP_TAC SUM_POS THEN
		RW_TAC std_ss [] THEN
		MATCH_MP_TAC REAL_LE_MUL THEN
		RW_TAC real_ss [GSYM REAL_ADD] THEN
		MATCH_MP_TAC PROB_POSITIVE THEN
		RW_TAC std_ss [PROB_SPACE_BERN, EVENT_EQ_INTER_BERN]) THEN


	KNOW_TAC `(\n. sum (0,n) (\x. sum (0,SUC x) (\i. & x * prob bern
                {s | (FST (X s) = i) /\ (FST (Y (SND (X s))) = x - i)}))) =
		(\n. sum (0,n) (\x. sum (0, n - x) (\i. &(x + i) * prob bern
                {s | (FST (X s) = x) /\ (FST (Y (SND (X s))) = i)})))` THEN1 (
			KNOW_TAC `!n. sum (0,n) (\x. sum (0,SUC x) (\i. & x * prob bern
                {s | (FST (X s) = i) /\ (FST (Y (SND (X s))) = x - i)})) =
                sum (0,n) (\x. sum (0,SUC x) (\i. & x * (\l m. prob bern
                 {s | (FST (X s) = l) /\ (FST (Y (SND (X s))) = m)}) i (x-i)))` THEN1
                		RW_TAC std_ss [] THEN
        	ASM_REWRITE_TAC [] THEN
    		POP_ASSUM K_TAC THEN
       		RW_TAC std_ss [SUM_REL_TWO_RV_EXPEC]) THEN
	FULL_SIMP_TAC std_ss [] THEN
	POP_ASSUM K_TAC THEN
	SUFF_TAC `m <= s` THEN1 (
		DISCH_TAC THEN
		PROVE_TAC [REAL_LE_ANTISYM]) THEN

		POP_ASSUM K_TAC THEN
		MATCH_MP_TAC SEQ_LE THEN
		EXISTS_TAC (--`(\n. sum (0,n) (\a. sum (0,n) (\b. & (a + b) * prob bern
			{s | ( FST ((X: (num -> bool) -> num#(num -> bool)) s) = a) /\ 
			( FST ((Y: (num -> bool) -> num#(num -> bool)) (SND (X s))) = b)})))`--) THEN
		EXISTS_TAC (--`(\n. (\n. sum (0,n) (\x. sum (0,n - x) 
			(\i. & (x + i) * prob bern
			{s | (FST ((X: (num -> bool) -> num#(num -> bool)) s)
				 = x) /\ (FST ((Y: (num -> bool) -> num#(num -> bool)) (SND (X s)))
			 = i)}))) (2* n))`--) THEN
		RW_TAC std_ss [SEQ_DOUBLE] THEN
		EXISTS_TAC (--`0:num`--) THEN
		RW_TAC arith_ss [] THEN
		KNOW_TAC `sum (0,2 * n) (\x. sum (0,2 * n - x) (\i. & (i + x) *
		prob bern {s | (FST (X s) = x) /\ (FST (Y (SND (X s))) = i)})) = 
		sum (0,n) (\x. sum (0,2 * n - x) (\i. & (i + x) *
		prob bern {s | (FST (X s) = x) /\ (FST (Y (SND (X s))) = i)})) +
		sum (n,n) (\x. sum (0,2 * n - x) (\i. & (i + x) *
		prob bern {s | (FST (X s) = x) /\ (FST (Y (SND (X s))) = i)}))` THEN1
			RW_TAC std_ss [SUM_TWO, TIMES2] THEN
		ASM_REWRITE_TAC [] THEN
		POP_ASSUM K_TAC THEN
		MATCH_MP_TAC REAL_LE_TRANS THEN
		EXISTS_TAC (--`sum (0,n) (\x. sum (0,2 * n - x) (\i. & (i + x) *
                		prob bern {s | (FST ((X: (num -> bool) -> num#(num -> bool)) s) = x) /\ 
				(FST ((Y: (num -> bool) -> num#(num -> bool)) (SND (X s))) = i)}))`--) THEN
		RW_TAC std_ss [REAL_LE_ADDR] THENL [
			MATCH_MP_TAC SUM_LE THEN
			RW_TAC arith_ss [] THEN
			KNOW_TAC `sum (0,2 * n - r)
        		(\i. & (i + r) * prob bern {s | (FST (X s) = r) /\ (FST (Y (SND (X s))) = i)}) =
                		sum (0,n)
        		(\i. & (i + r) * prob bern {s | (FST (X s) = r) /\ (FST (Y (SND (X s))) = i)}) +
                		sum (n, n - r)
        		(\i. & (i + r) * prob bern {s | (FST (X s) = r) /\ (FST (Y (SND (X s))) = i)})` THEN1 (
				RW_TAC std_ss [SUM_TWO,TIMES2] THEN
				RW_TAC arith_ss []) THEN
			ASM_REWRITE_TAC [REAL_LE_ADDR] THEN
			MATCH_MP_TAC SUM_POS THEN
	        	RW_TAC std_ss [] THEN
       			MATCH_MP_TAC REAL_LE_MUL THEN
        		RW_TAC real_ss [GSYM REAL_ADD] THEN
        		MATCH_MP_TAC PROB_POSITIVE THEN
        		RW_TAC std_ss [PROB_SPACE_BERN, EVENT_EQ_INTER_BERN],


			MATCH_MP_TAC SUM_POS THEN
                	RW_TAC std_ss [] THEN
			MATCH_MP_TAC SUM_POS THEN
			RW_TAC std_ss [] THEN
                	MATCH_MP_TAC REAL_LE_MUL THEN
                	RW_TAC real_ss [GSYM REAL_ADD] THEN
                	MATCH_MP_TAC PROB_POSITIVE THEN
                	RW_TAC std_ss [PROB_SPACE_BERN, EVENT_EQ_INTER_BERN]]);

val SUM_TWO_RV_SUMMABLE = store_thm
 ("SUM_TWO_RV_SUMMABLE",
 (--`!(X: (num -> bool) -> num#(num -> bool)) (Y: (num -> bool) -> num#(num -> bool)). 
	 X IN indep_fn /\ Y IN indep_fn /\
	 summable (\x. & x * prob bern {s | FST (X s) = x}) /\
         summable (\x. & x * prob bern {s | FST (Y s) = x}) ==>
		summable (\x. & x * prob bern {s | FST (X s) + FST (Y (SND (X s))) = x})`--),
	RW_TAC std_ss [SUM_RV_BIGUNION, summable] THEN
	KNOW_TAC `!x n. prob bern (BIGUNION (IMAGE 
		(\i. {s | (FST (X s) = i) /\ (FST (Y (SND (X s))) = x - i)}) {i | i <= x})) = 
				sum (0,SUC x) (\i. prob bern {s | (FST (X s) = i) /\ 
							(FST (Y (SND (X s))) = x - i)})` THEN1 (
		GEN_TAC THEN
		MATCH_MP_TAC PROB_ADDITIVE_ON_SUM_RV THEN
		RW_TAC std_ss []) THEN
	FULL_SIMP_TAC std_ss [GSYM SER_CMUL] THEN
	POP_ASSUM K_TAC THEN
	RW_TAC std_ss [sums, GSYM convergent] THEN
	MATCH_MP_TAC SEQ_BCONV THEN
	CONJ_TAC THENL [
		MATCH_MP_TAC SEQ_BOUNDED_2 THEN
		EXISTS_TAC (--`0:real`--) THEN
		EXISTS_TAC (--`s + s':real`--) THEN
		GEN_TAC THEN
		CONJ_TAC THENL [
			RW_TAC std_ss [] THEN
			MATCH_MP_TAC SUM_POS THEN
			RW_TAC std_ss [] THEN
			MATCH_MP_TAC REAL_LE_MUL THEN
                	RW_TAC real_ss [GSYM REAL_ADD] THEN
			MATCH_MP_TAC SUM_POS THEN
			RW_TAC std_ss [] THEN
                	MATCH_MP_TAC PROB_POSITIVE THEN
                	RW_TAC std_ss [PROB_SPACE_BERN, EVENT_EQ_INTER_BERN],

			RW_TAC std_ss [] THEN
			MATCH_MP_TAC REAL_LE_TRANS THEN
			EXISTS_TAC (--`sum (0,n) (\x. & x *  
			(prob bern {s | (FST ((X: (num -> bool) -> num#(num -> bool)) s) = x)} + 
			 prob bern {s | (FST ((Y: (num -> bool) -> num#(num -> bool)) (SND (X s))) = x)}))`--) THEN
			CONJ_TAC THENL [
				KNOW_TAC `sum (0,n) (\x. & x * 
				sum (0,SUC x) (\i. prob bern 
					{s | (FST (X s) = i) /\ (FST (Y (SND (X s))) = x - i)})) =
				sum (0,n) (\x. & x * sum (0,n - x) (\i. prob bern
					{s | (FST (X s) = x) /\ (FST (Y (SND (X s))) = i)} + prob bern
					{s | (FST (X s) = i) /\ (FST (Y (SND (X s))) = x)}))` THEN1 (
					KNOW_TAC `sum (0,n) (\x. & x *
                                	sum (0,SUC x) (\i. prob bern
                                        {s | (FST (X s) = i) /\ (FST (Y (SND (X s))) = x - i)})) = 
					sum (0,n) (\x. & x * sum (0,SUC x) (\i. 
						(\l m. prob bern 
						{s | (FST (X s) = l) /\ (FST (Y (SND (X s))) = m)}) i (x-i)))` THEN1
						RW_TAC std_ss [] THEN
					ASM_REWRITE_TAC [] THEN
					POP_ASSUM K_TAC THEN
					RW_TAC std_ss [SUM_REL_TWO_RV_EXPEC2]) THEN
				FULL_SIMP_TAC std_ss [] THEN
				POP_ASSUM K_TAC THEN
				KNOW_TAC `!x. sum (0,n - x) 
				(\i. prob bern {s | (FST (X s) = x) /\ (FST (Y (SND (X s))) = i)} +
                        	prob bern {s | (FST (X s) = i) /\ (FST (Y (SND (X s))) = x)}) =
                		sum (0,n - x) 
				(\i. (\i. prob bern {s | (FST (X s) = x) /\ (FST (Y (SND (X s))) = i)}) i +
                        	(\i. prob bern {s | (FST (X s) = i) /\ (FST (Y (SND (X s))) = x)}) i)` THEN1
					RW_TAC std_ss [] THEN
				ASM_REWRITE_TAC [SUM_ADD] THEN
				POP_ASSUM K_TAC THEN
				KNOW_TAC `!x. sum (0,n - x) 
				(\i. prob bern {s | (FST (X s) = x) /\ (FST (Y (SND (X s))) = i)}) =
				prob bern (BIGUNION (IMAGE 
					(\i. {s | (FST (X s) = x) /\ (FST (Y (SND (X s))) = i)}) {i | i < n-x}))` THEN1 (
					GEN_TAC THEN
					MATCH_MP_TAC SUM_PROB_INTER_BIGUNION THEN
					RW_TAC std_ss []) THEN
				KNOW_TAC `!x. sum (0,n - x)
				(\i. prob bern {s | (FST (X s) = i) /\ (FST (Y (SND (X s))) = x)}) =
					prob bern (BIGUNION (IMAGE 
					(\i. {s | (FST (X s) = i) /\ (FST (Y(SND (X s)) ) = x)}) {i | i < n-x}))` THEN1 (
					GEN_TAC THEN
					MATCH_MP_TAC SUM_PROB_INTER_BIGUNION2 THEN
					RW_TAC std_ss []) THEN
				FULL_SIMP_TAC std_ss [] THEN
				NTAC 2 (POP_ASSUM K_TAC) THEN
				MATCH_MP_TAC SUM_LE THEN
				RW_TAC std_ss [] THEN
				MATCH_MP_TAC REAL_LE_LMUL_IMP THEN
				RW_TAC real_ss [] THEN
				MATCH_MP_TAC REAL_LE_ADD2 THEN
				CONJ_TAC THENL [
					MATCH_MP_TAC PROB_INCREASING THEN
					KNOW_TAC `!i. {s | (FST (X s) = r) /\ (FST (Y (SND (X s))) = i)} =
					{s | (FST (X s) = r)} INTER {s | (FST (Y (SND (X s))) = i)}` THEN1 (
						GEN_TAC THEN
						REWRITE_TAC [EXTENSION] THEN
						RW_TAC std_ss [IN_INTER, GSPECIFICATION]) THEN
					ASM_REWRITE_TAC [] THEN
					RW_TAC std_ss [PROB_SPACE_BERN, SUBSET_DEF, IN_INTER,
							IN_BIGUNION_IMAGE, GSPECIFICATION] THENL [
						MATCH_MP_TAC EVENTS_COUNTABLE_UNION THEN
						RW_TAC std_ss [PROB_SPACE_BERN] THENL [
							RW_TAC std_ss [IMAGE_DEF, IN_INTER,
							SUBSET_DEF, GSPECIFICATION] THEN
							MATCH_MP_TAC EVENTS_INTER THEN
							RW_TAC std_ss [PROB_SPACE_BERN, RV_EQ_SND_RV_SEQ_EVENTS_BERN] THEN
							KNOW_TAC `{s | FST (X s) = r} = ((\y. y = r) o FST o X)` THEN1
								PSET_TAC [o_DEF] THEN
							RW_TAC std_ss [INDEP_FN_FST_EVENTS],
						

							MATCH_MP_TAC COUNTABLE_IMAGE THEN
							RW_TAC std_ss [countable_def, 
							IN_IMAGE, GSPECIFICATION] THEN
							EXISTS_TAC (--`(\x. (x:num))`--) THEN
							RW_TAC std_ss []],

						KNOW_TAC `{s | FST (X s) = r} = ((\y. y = r) o FST o X)` THEN1
							PSET_TAC [o_DEF] THEN
						RW_TAC std_ss [INDEP_FN_FST_EVENTS]], 

					MATCH_MP_TAC PROB_INCREASING THEN
					KNOW_TAC `!i. {s | (FST (X s) = i) /\ (FST (Y (SND (X s))) = r)} =
					{s | (FST (X s) = i)} INTER {s | (FST (Y (SND (X s))) = r)}` THEN1 (
						GEN_TAC THEN
						REWRITE_TAC [EXTENSION] THEN
						RW_TAC std_ss [IN_INTER, GSPECIFICATION]) THEN
					ASM_REWRITE_TAC [] THEN
					RW_TAC std_ss [PROB_SPACE_BERN, SUBSET_DEF, IN_INTER,
							IN_BIGUNION_IMAGE, GSPECIFICATION] THENL [
						MATCH_MP_TAC EVENTS_COUNTABLE_UNION THEN
						RW_TAC std_ss [PROB_SPACE_BERN] THENL [
							RW_TAC std_ss [IMAGE_DEF, IN_INTER,
							SUBSET_DEF, GSPECIFICATION] THEN
							MATCH_MP_TAC EVENTS_INTER THEN
							RW_TAC std_ss [PROB_SPACE_BERN, RV_EQ_SND_RV_SEQ_EVENTS_BERN] THEN
							KNOW_TAC `{s | FST (X s) = i} = ((\y. y = i) o FST o X)` THEN1
								PSET_TAC [o_DEF] THEN
							RW_TAC std_ss [INDEP_FN_FST_EVENTS],

							MATCH_MP_TAC COUNTABLE_IMAGE THEN
							RW_TAC std_ss [countable_def, 
							IN_IMAGE, GSPECIFICATION] THEN
							EXISTS_TAC (--`(\x. (x:num))`--) THEN
							RW_TAC std_ss []], 

						RW_TAC std_ss [RV_EQ_SND_RV_SEQ_EVENTS_BERN]]],
					

			RW_TAC std_ss [PROB_SND_RV_SEQ] THEN
			KNOW_TAC `sum (0,n) (\x. & x * 
			(prob bern {s | FST (X s) = x} + prob bern {s | FST (Y s) = x})) = 
			(\n. sum (0,n) (\x. & x * 
				(prob bern {s | FST (X s) = x} + prob bern {s | FST (Y s) = x}))) n ` THEN1
				RW_TAC std_ss [] THEN
			ONCE_ASM_REWRITE_TAC [] THEN
			POP_ASSUM K_TAC THEN
			MATCH_MP_TAC SEQ_MONO_LE THEN
			CONJ_TAC THENL [
				RW_TAC std_ss [GSYM ADD1, sum, REAL_LE_ADDR] THEN
				MATCH_MP_TAC REAL_LE_MUL THEN
				RW_TAC real_ss [GSYM REAL_ADD] THEN		
				MATCH_MP_TAC REAL_LE_ADD THEN
				CONJ_TAC THENL [
					MATCH_MP_TAC PROB_POSITIVE THEN	
					RW_TAC std_ss [PROB_SPACE_BERN] THEN
					KNOW_TAC `{s | FST (X s) = n} = ((\y. y = n) o FST o X)` THEN1
						PSET_TAC [o_DEF] THEN
					RW_TAC std_ss [INDEP_FN_FST_EVENTS],

					MATCH_MP_TAC PROB_POSITIVE THEN	
					RW_TAC std_ss [PROB_SPACE_BERN] THEN
					KNOW_TAC `{s | FST (Y s) = n} = ((\y. y = n) o FST o Y)` THEN1
						PSET_TAC [o_DEF] THEN
					RW_TAC std_ss [INDEP_FN_FST_EVENTS]],
				RW_TAC std_ss [GSYM sums, REAL_ADD_LDISTRIB] THEN
				KNOW_TAC `(\x. & x * prob bern {s | FST (X s) = x} + 
					& x * prob bern {s | FST (Y s) = x}) =
					(\x. (\x. & x * prob bern {s | FST (X s) = x}) x + 
						(\x. & x * prob bern {s | FST (Y s) = x}) x)` THEN1
					RW_TAC std_ss [] THEN
				ONCE_ASM_REWRITE_TAC [] THEN
				POP_ASSUM K_TAC THEN		
				MATCH_MP_TAC SER_ADD THEN
				RW_TAC std_ss []]]],

		RW_TAC std_ss [MONO_SUC] THEN
		DISJ1_TAC THEN
		RW_TAC arith_ss [real_ge, sum, REAL_LE_ADDR] THEN
		MATCH_MP_TAC REAL_LE_MUL THEN
		RW_TAC real_ss [GSYM REAL_ADD] THEN
        	MATCH_MP_TAC REAL_LE_ADD THEN
		CONJ_TAC THENL [
			MATCH_MP_TAC SUM_POS THEN
        		RW_TAC std_ss [] THEN
        		MATCH_MP_TAC PROB_POSITIVE THEN
        		RW_TAC std_ss [PROB_SPACE_BERN, EVENT_EQ_INTER_BERN],


			MATCH_MP_TAC PROB_POSITIVE THEN
			RW_TAC std_ss [PROB_SPACE_BERN, EVENT_EQ_INTER_BERN]]]);

val EXPEC_SUM_TWO_RV = store_thm
 ("EXPEC_SUM_TWO_RV",
 (--`!(X: (num -> bool) -> num#(num -> bool)) (Y: (num -> bool) -> num#(num -> bool)). 
	 X IN indep_fn /\ Y IN indep_fn /\
	 summable (\x. & x * prob bern {s | FST (X s) = x}) /\
         summable (\x. & x * prob bern {s | FST (Y s) = x})  ==>
                (expec (\x. (FST (X x) + FST (Y (SND (X x))),SND (Y (SND (X x))))) = expec X + expec Y)`--),
	RW_TAC std_ss [BIND_DEF, o_DEF, UNCURRY_DEF, UNIT_DEF, LAMBDA_PAIR,
			summable, expec_def, suminf, sums] THEN
	KNOW_TAC `(!m. {s | FST (X s) = m} IN events bern)` THEN1 (
		GEN_TAC THEN
		KNOW_TAC `{s | FST (X s) = m} = ((\y. y = m) o FST o X)` THEN1
			PSET_TAC [o_DEF] THEN
		RW_TAC std_ss [INDEP_FN_FST_EVENTS]) THEN
	KNOW_TAC `(!m. {s | FST (Y s) = m} IN events bern)` THEN1 (
		GEN_TAC THEN
		KNOW_TAC `{s | FST (Y s) = m} = ((\y. y = m) o FST o Y)` THEN1
			PSET_TAC [o_DEF] THEN
		RW_TAC std_ss [INDEP_FN_FST_EVENTS]) THEN
	SELECT_ELIM_TAC THEN
	CONJ_TAC THEN1 (
		FULL_SIMP_TAC std_ss [GSYM summable, GSYM sums] THEN

		MATCH_MP_TAC SUM_TWO_RV_SUMMABLE THEN
		RW_TAC std_ss [] THENL [
			MATCH_MP_TAC SUM_SUMMABLE THEN
			PROVE_TAC [],

			MATCH_MP_TAC SUM_SUMMABLE THEN
                	PROVE_TAC []]) THEN
	
		RW_TAC std_ss [] THEN
		SELECT_ELIM_TAC THEN
		CONJ_TAC THEN1
			PROVE_TAC [] THEN	
	
			RW_TAC std_ss [] THEN
			SELECT_ELIM_TAC THEN
			CONJ_TAC THEN1
				PROVE_TAC [] THEN	
				RW_TAC std_ss [] THEN
				MATCH_MP_TAC SEQ_UNIQ THEN
				FULL_SIMP_TAC std_ss [SUM_RV_BIGUNION] THEN
				KNOW_TAC `!x n. prob bern (BIGUNION (IMAGE 
				(\i. {s | (FST (X s) = i) /\ (FST (Y (SND (X s))) = x - i)}) 
						{i | i <= x})) = 
						sum (0,SUC x)
            					(\i. prob bern {s | (FST (X s) = i) /\ 
						(FST (Y (SND (X s))) = x - i)})` THEN1 (
						 	GEN_TAC THEN
						 	MATCH_MP_TAC PROB_ADDITIVE_ON_SUM_RV THEN
						 	RW_TAC std_ss []) THEN
				FULL_SIMP_TAC std_ss [GSYM SUM_CMUL] THEN
				POP_ASSUM K_TAC THEN			
				EXISTS_TAC (--`(\n.
             				sum (0,n)
               				(\x.
                  			sum (0,SUC x)
                    			(\i.
                       			& x *
                       			prob bern
                      		{s | (FST ((X: (num -> bool) -> num#(num -> bool)) s) = i) /\
			     (FST ((Y: (num -> bool) -> num#(num -> bool)) (SND (X s))) = x - i)}
					     )))`--) THEN
				RW_TAC std_ss [] THEN
				SUFF_TAC `(\n. sum (0, n) (\a. sum (0, n) (\b.
					&(a + b) * prob bern
                   		{s | (FST (X s) = a) /\ (FST (Y (SND (X s))) = b)}))) --> (x' + x'')` THEN1 (
					DISCH_TAC THEN
					MATCH_MP_TAC TWO_RV_EXPEC_HELPER THEN
					RW_TAC std_ss [summable, sums] THEN
					EXISTS_TAC (--`x:real`--) THEN
					ASM_REWRITE_TAC []) THEN
		
					REWRITE_TAC [REAL_ADD_RDISTRIB, GSYM REAL_ADD] THEN
					KNOW_TAC `!a n. (\b. & a * prob bern 
					{s | (FST (X s) = a) /\ (FST (Y (SND (X s))) = b)} +
					& b * prob bern 
					{s | (FST (X s) = a) /\ (FST (Y (SND (X s))) = b)}) =
					(\b. (\b. & a * prob bern
                                        {s | (FST (X s) = a) /\ (FST (Y (SND (X s))) = b)}) b + 
					(\b. & b * prob bern
					{s | (FST (X s) = a) /\ (FST (Y (SND (X s))) = b)}) b)` THEN1
						RW_TAC std_ss [] THEN
					ASM_REWRITE_TAC [] THEN
					REWRITE_TAC [SUM_ADD] THEN
					POP_ASSUM K_TAC THEN
					KNOW_TAC `!n. (\a. sum (0,n) (\b.  & a * prob bern 
					{s | (FST (X s) = a) /\ (FST (Y (SND (X s))) = b)}) +
					sum (0,n) (\b. & b * prob bern 
					{s | (FST (X s) = a) /\ (FST (Y (SND (X s))) = b)})) = 
					(\a. (\a. sum (0,n) (\b.  & a * prob bern
                                        {s | (FST (X s) = a) /\ (FST (Y (SND (X s))) = b)})) a +
					(\a. sum (0,n) (\b. & b * prob bern
                                        {s | (FST (X s) = a) /\ (FST (Y (SND (X s))) = b)})) a )` THEN1
						RW_TAC std_ss [] THEN
                                	ASM_REWRITE_TAC [] THEN
                                	REWRITE_TAC [SUM_ADD] THEN
                                	POP_ASSUM K_TAC THEN
					KNOW_TAC `(\n. sum (0,n) (\a. sum (0,n) (\b. & a *
					prob bern {s | (FST (X s) = a) /\ (FST (Y (SND (X s))) = b)})) +
					sum (0,n) (\a. sum (0,n) (\b. & b * 
					prob bern {s | (FST (X s) = a) /\ (FST (Y (SND (X s))) = b)}))) =
					(\n. (\n. sum (0,n) (\a. sum (0,n) (\b. & a *
                                        prob bern {s | (FST (X s) = a) /\ (FST (Y (SND (X s))) = b)})))n + 										(\n. sum (0,n) (\a. sum (0,n) (\b. & b *
                                      prob bern {s | (FST (X s) = a) /\ (FST (Y (SND (X s))) = b)}))) n)` THEN1
						RW_TAC std_ss [] THEN
                                	ASM_REWRITE_TAC [] THEN
					MATCH_MP_TAC SEQ_ADD THEN
					POP_ASSUM K_TAC THEN
					KNOW_TAC `!n. sum (0,n) (\a. sum (0,n) (\b. & b *
					prob bern {s | (FST (X s) = a) /\ (FST (Y (SND (X s))) = b)})) =
					 sum (0,n) (\b. sum (0,n) (\a. & b *
					prob bern {s | (FST (X s) = a) /\ (FST (Y (SND (X s))) = b)}))` THEN1 (
						Induct THEN1
							RW_TAC real_ss [sum] THEN

							RW_TAC real_ss [sum, REAL_ADD_ASSOC,
							REAL_EQ_RADD] THEN
							KNOW_TAC `!n'. (\a.sum (0,n')(\b. & b * 
								prob bern {s | (FST (X s) = a) /\ (FST (Y (SND (X s))) = b)}) +
								& n' * prob bern
					 			{s | (FST (X s) = a) /\ (FST (Y (SND (X s))) = n')}) =
								(\a. (\a.sum (0,n')(\b. & b *
                                        			prob bern {s | (FST (X s) = a) /\ (FST (Y (SND (X s))) = b)})) a +
								(\a. & n' * prob bern
                                         			{s | (FST (X s) = a) /\ (FST (Y (SND (X s))) = n')})a)` THEN1
								RW_TAC std_ss [] THEN
							KNOW_TAC `!n'.(\b. sum (0,n') (\a. & b * 
								prob bern {s | (FST (X s) = a) /\ (FST (Y (SND (X s))) = b)}) +
								& b * prob bern 
								{s | (FST (X s) = n') /\ (FST (Y (SND (X s))) = b)}) =
								(\b. (\b. sum (0,n') (\a. & b *
                                        			prob bern {s | (FST (X s) = a) /\ (FST (Y (SND (X s))) = b)}))b +
								(\b. & b * prob bern
                                        			{s | (FST (X s) = n') /\ (FST (Y (SND (X s))) = b)})b)` THEN1
								RW_TAC std_ss [] THEN
							ASM_REWRITE_TAC [] THEN
							REWRITE_TAC [SUM_ADD] THEN
							NTAC 2 (POP_ASSUM K_TAC) THEN
							ASM_REWRITE_TAC [] THEN
							REAL_ARITH_TAC) THEN
					ASM_REWRITE_TAC [] THEN		
					POP_ASSUM K_TAC THEN
					KNOW_TAC `!a b.{s | (FST (X s) = a) /\ (FST (Y (SND (X s))) = b)} = 
						{s | (FST (X s) = a)} INTER {s | (FST (Y (SND (X s))) = b)}` THEN1 (
						REPEAT GEN_TAC THEN
						ONCE_REWRITE_TAC [EXTENSION] THEN
						PSET_TAC [GSPECIFICATION, IN_INTER]) THEN
					ASM_REWRITE_TAC [] THEN
					POP_ASSUM K_TAC THEN
					KNOW_TAC `!a b. prob bern ({s | FST (X s) = a} INTER
                    					{s | FST (Y (SND (X s))) = b}) =
							prob bern {s | FST (X s) = a} * prob bern {s | FST (Y s) = b}` THEN1 (
						REPEAT GEN_TAC THEN
						KNOW_TAC `{s | FST (X  s) = a} = ((\y. y = a) o FST o X )` THEN1
							PSET_TAC [o_DEF] THEN
						KNOW_TAC `{s | FST (Y  (SND (X  s))) = b} = 	
							((\y. FST (Y  y) = b) o SND o X )` THEN1
							PSET_TAC [o_DEF] THEN		
						ONCE_ASM_REWRITE_TAC [] THEN
						KNOW_TAC `{s | FST (Y s) = b} = (\y. FST (Y y) = b)` THEN1 (
							REWRITE_TAC [EXTENSION] THEN
							PSET_TAC [GSPECIFICATION]) THEN
						ONCE_ASM_REWRITE_TAC [] THEN
						MATCH_MP_TAC INDEP_FN_PROB THEN
						KNOW_TAC `(\y. FST (Y y) = b) = ((\y. y = b) o FST o Y)` THEN1
							PSET_TAC [o_DEF] THEN
						ONCE_ASM_REWRITE_TAC [] THEN
						RW_TAC std_ss [INDEP_FN_FST_EVENTS]) THEN
					FULL_SIMP_TAC std_ss [SUM_CMUL, REAL_MUL_ASSOC] THEN
					ONCE_REWRITE_TAC [REAL_MUL_COMM] THEN
					RW_TAC std_ss [SUM_CMUL] THENL [
						KNOW_TAC `(\n. sum (0,n) (\b. prob bern {s | FST (Y s) = b}) *
						sum (0,n) (\a. & a * prob bern {s | FST (X s) = a})) =
						(\n. (\n. sum (0,n) (\b. prob bern {s | FST (Y s) = b})) n *
						(\n. sum (0,n) (\a. & a * prob bern {s | FST (X s) = a})) n)` THEN1
							RW_TAC std_ss [] THEN
						ASM_REWRITE_TAC [] THEN
						KNOW_TAC `x' = 1 * x'` THEN1
							RW_TAC real_ss [] THEN
						ONCE_ASM_REWRITE_TAC [] THEN
						NTAC 2 (POP_ASSUM K_TAC) THEN
						MATCH_MP_TAC SEQ_MUL THEN
						FULL_SIMP_TAC std_ss [GSYM sums, PMF_SUMS_1],

						ONCE_REWRITE_TAC [REAL_MUL_SYM] THEN
						KNOW_TAC `!a b. prob bern {s | FST (X s) = a} * 
								prob bern {s | FST (Y s) = b} = 
							prob bern {s | FST (Y s) = b} * 
								prob bern {s | FST (X s) = a}` THEN1
							PROVE_TAC [REAL_MUL_SYM] THEN
						ASM_REWRITE_TAC [] THEN
						RW_TAC std_ss [SUM_CMUL, REAL_MUL_ASSOC] THEN
						ONCE_REWRITE_TAC [REAL_MUL_SYM] THEN
						RW_TAC std_ss [SUM_CMUL] THEN
						KNOW_TAC `(\n. sum (0,n) (\a. prob bern {s | FST (X s) = a}) *
							sum (0,n) (\b. & b * prob bern {s | FST (Y s) = b})) =
							(\n. (\n.sum (0,n) (\a. prob bern {s | FST (X s) = a})) n *
							(\n. sum (0,n) (\b. & b * prob bern {s | FST (Y s) = b})) n)` THEN1
							RW_TAC std_ss [] THEN
						KNOW_TAC `x'' = 1 * x''` THEN1
							RW_TAC real_ss [] THEN
						ONCE_ASM_REWRITE_TAC [] THEN
						MATCH_MP_TAC SEQ_MUL THEN
						FULL_SIMP_TAC std_ss [GSYM sums, PMF_SUMS_1]]);

val INDEP_FN_SUM_TWO_RV = store_thm
 ("INDEP_FN_SUM_TWO_RV",
 (--`!(X: (num -> bool) -> num#(num -> bool)) (Y: (num -> bool) -> num#(num -> bool)). 
	 X IN indep_fn /\ Y IN indep_fn ==>
		((\x. (FST (X x) + FST (Y (SND (X x))),SND (Y (SND (X x))))) IN indep_fn)`--),
	REPEAT STRIP_TAC THEN
	KNOW_TAC `(\x. (FST (X x) + FST (Y (SND (X x))),SND (Y (SND (X x))))) = 
	BIND X (\a. BIND Y (\b. UNIT (a + b)))` THEN1
		RW_TAC std_ss [BIND_DEF, o_DEF, UNCURRY_DEF, UNIT_DEF, LAMBDA_PAIR] THEN
	ASM_REWRITE_TAC [] THEN
	RW_TAC std_ss [INDEP_FN_BIND, INDEP_FN_UNIT]);

(*=====================================================*)

val sum_rv_lst_def = Define
	`(sum_rv_lst [] s = (0,s)) /\
         (sum_rv_lst (h::L: ((num -> bool) -> num#(num->bool)) list) s = 
		(FST (h s) + (FST (sum_rv_lst L (SND (h s)))),
				(SND (sum_rv_lst L (SND (h s))))))`;

val SUM_RV_LIST_MONAD = store_thm
 ("SUM_RV_LIST_MONAD",
 (--`(sum_rv_lst [] = UNIT 0) /\
        (!L h. sum_rv_lst (h::L)  =
		BIND h (\a. BIND (sum_rv_lst L) (\b. UNIT (a + b))))`--),
	RW_TAC arith_ss [BIND_DEF, UNIT_DEF] THENL [
		RW_TAC arith_ss [sum_rv_lst_def, GSYM EQ_EXT_EQ],

        	RW_TAC arith_ss [o_DEF] THEN
        	RW_TAC arith_ss [sum_rv_lst_def, GSYM EQ_EXT_EQ] THEN
        	RW_TAC std_ss [LAMBDA_PAIR]]);

val INDEP_FN_SUM_RV_LIST = store_thm
 ("INDEP_FN_SUM_RV_LIST",
 (--`!L. (!h. MEM h L ==> h IN indep_fn) ==>
			sum_rv_lst L IN indep_fn`--),
	REPEAT STRIP_TAC THEN
	Induct_on `L` THEN1
		RW_TAC std_ss [SUM_RV_LIST_MONAD, INDEP_FN_UNIT] THEN
	
		RW_TAC std_ss [SUM_RV_LIST_MONAD, INDEP_FN_UNIT] THEN
        	MATCH_MP_TAC INDEP_FN_BIND THEN
        	RW_TAC std_ss [] THENL [
			POP_ASSUM (MP_TAC o Q.SPEC `h`) THEN
			RW_TAC std_ss [MEM],
		
        		MATCH_MP_TAC INDEP_FN_BIND THEN
        		RW_TAC std_ss [INDEP_FN_UNIT] THEN
			KNOW_TAC `(!h. IS_EL h L ==> h IN indep_fn)` THEN1 (
				GEN_TAC THEN
				POP_ASSUM (MP_TAC o Q.SPEC `h'`) THEN	
				RW_TAC std_ss [MEM]) THEN
			FULL_SIMP_TAC std_ss []]);

val SUM_N_RV_SUMMABLE = store_thm
 ("SUM_N_RV_SUMMABLE",
 (--`!L. (!h. MEM h L ==> h IN indep_fn) /\
	(!a. MEM a L ==> (summable (\x. & x * prob bern {s | FST (a s) = x}))) ==>
		(summable (\x. & x * prob bern {s | FST ((sum_rv_lst L) s) = x}))`--),
	REPEAT STRIP_TAC THEN
	Induct_on `L` THENL [
		ONCE_REWRITE_TAC [MEM] THEN
		KNOW_TAC `!n. prob bern {s | FST (sum_rv_lst [] s) = n} =
                                (if (0 = n) then 1 else 0)` THEN1 (
                		RW_TAC std_ss [sum_rv_lst_def] THEN
                        	RW_TAC std_ss [GSPEC_T, PROB_BERN_BASIC] THEN
                        	RW_TAC std_ss [GSPEC_F, PROB_BERN_BASIC]) THEN
        	FULL_SIMP_TAC std_ss [] THEN
        	POP_ASSUM K_TAC THEN
		RW_TAC std_ss [summable, sums] THEN
		KNOW_TAC `!n. sum (0,n) (\n. & n * (if 0 = n then 1 else 0)) = 0` THEN1 (
                        	Induct THEN1
                                	RW_TAC arith_ss [sum] THEN

                                	RW_TAC real_ss [sum]) THEN
		ASM_REWRITE_TAC [] THEN
		EXISTS_TAC (--`0:real`--) THEN
		RW_TAC std_ss [SEQ_CONST],
	
		RW_TAC arith_ss [MEM, sum_rv_lst_def] THEN
		MATCH_MP_TAC SUM_TWO_RV_SUMMABLE THEN
		FULL_SIMP_TAC std_ss [AND_IMP_INTRO] THEN
		MATCH_MP_TAC INDEP_FN_SUM_RV_LIST THEN
		GEN_TAC THEN
		PROVE_TAC []]);

val EXPEC_SUM_N_RV = store_thm
 ("EXPEC_SUM_N_RV",
 (--`!L.  (!h. MEM h L ==> h IN indep_fn) /\
	(!a. MEM a L ==> (summable (\x. & x * prob bern {s | FST (a s) = x}))) ==>
	 (expec (sum_rv_lst L) = sum (0, LENGTH L) (\x. expec (EL (LENGTH L - (x+1)) L)))`--),
	REPEAT STRIP_TAC THEN
	Induct_on `L` THENL [
		ONCE_REWRITE_TAC [LENGTH] THEN
		ONCE_REWRITE_TAC [sum] THEN
		ONCE_REWRITE_TAC [expec_def] THEN
		ONCE_REWRITE_TAC [MEM] THEN
		REPEAT STRIP_TAC THEN
		KNOW_TAC `summable (\x. & x * prob bern {s | FST (sum_rv_lst [] s) = x})` THEN1 (
			MATCH_MP_TAC SUM_N_RV_SUMMABLE THEN
			RW_TAC std_ss [MEM]) THEN
		KNOW_TAC `!n. prob bern {s | FST (sum_rv_lst [] s) = n} = 
				(if (0 = n) then 1 else 0)` THEN1 (
				RW_TAC std_ss [sum_rv_lst_def] THEN
                		RW_TAC std_ss [GSPEC_T, PROB_BERN_BASIC] THEN
                		RW_TAC std_ss [GSPEC_F, PROB_BERN_BASIC]) THEN
		FULL_SIMP_TAC std_ss [] THEN
		POP_ASSUM K_TAC THEN
		RW_TAC std_ss [summable, suminf] THEN
		SELECT_ELIM_TAC THEN
		CONJ_TAC THEN1
			PROVE_TAC [summable] THEN
			
			RW_TAC std_ss [] THEN
			FULL_SIMP_TAC std_ss [sums] THEN
			MATCH_MP_TAC SEQ_UNIQ THEN
			EXISTS_TAC (--`(\n. sum (0,n) (\n. & n * (if 0 = n then 1 else 0)))`--) THEN
			ASM_REWRITE_TAC [] THEN
			KNOW_TAC `!n. sum (0,n) (\n. & n * (if 0 = n then 1 else 0)) = 0` THEN1 (
				Induct THEN1
					RW_TAC arith_ss [sum] THEN

					RW_TAC real_ss [sum]) THEN
					
			ASM_REWRITE_TAC [SEQ_CONST],

		RW_TAC arith_ss [LENGTH, MEM, sum, sum_rv_lst_def] THEN
		FULL_SIMP_TAC std_ss [AND_IMP_INTRO] THEN
		KNOW_TAC `expec (sum_rv_lst (h::L)) = 
		expec (\s. (sum_rv_lst (h::L)) s)` THEN1 (
			MATCH_MP_TAC RAND_THM THEN
			PROVE_TAC [ETA_THM]) THEN
		ASM_REWRITE_TAC [] THEN
		RW_TAC arith_ss [sum_rv_lst_def, ADD1, EL, HD] THEN
		KNOW_TAC `expec (\s. (FST (h s) + FST (sum_rv_lst L (SND (h s))),
          			SND (sum_rv_lst L (SND (h s))))) = 
				expec (sum_rv_lst L) + expec h` THEN1 (
			ONCE_REWRITE_TAC [REAL_ADD_SYM] THEN
			MATCH_MP_TAC EXPEC_SUM_TWO_RV THEN
			CONJ_TAC THEN1
				PROVE_TAC [MEM] THEN
				CONJ_TAC THENL [
					MATCH_MP_TAC INDEP_FN_SUM_RV_LIST THEN
					PROVE_TAC [MEM],
					CONJ_TAC THENL [
						POP_ASSUM K_TAC THEN
						POP_ASSUM (MP_TAC o Q.SPEC `h`) THEN
						RW_TAC std_ss [],
					
						MATCH_MP_TAC SUM_N_RV_SUMMABLE THEN
						CONJ_TAC THENL [
							GEN_TAC THEN
							NTAC 2 (POP_ASSUM K_TAC) THEN
							POP_ASSUM (MP_TAC o Q.SPEC `h'`) THEN
							RW_TAC std_ss [],

							GEN_TAC THEN	
							POP_ASSUM K_TAC THEN
							POP_ASSUM (MP_TAC o Q.SPEC `a`) THEN
							RW_TAC std_ss []]]]) THEN
		ASM_REWRITE_TAC [REAL_EQ_RADD] THEN
		POP_ASSUM K_TAC THEN
		MATCH_MP_TAC SUM_EQ THEN
		RW_TAC arith_ss [] THEN
		MATCH_MP_TAC RAND_THM THEN
		KNOW_TAC `(LENGTH L - r) = 
		SUC ((LENGTH L - (r + 1)))` THEN1
			DECIDE_TAC THEN
		ASM_REWRITE_TAC [EL, TL]]);


	


		

val _ = export_theory();
